#include <fstream>
#include <iomanip>
#include <string>
#include <string.h>
#include <stdio.h>

#include "NavierStokes.h"
#include "streamFunction2D.h"

#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Field.h"
#include "Integ.h"
#include "Constr.h"
#include "Mesh.h"
#include "param.h"
#include "paramList.h"
#include "Plot.h"
#include "Solver.h"
#include "TimeDep.h"

using namespace std ;
using namespace keyFE2 ;
using namespace keyFlow ;



int main(void) {
  char buf[256];

  Mesh MESH;
  
  // Reading and renumbering of the mesh:
  
  MESH.read("planeJet.msh","rsMesh");
  string sortType="frontal";
  if (sortType=="position")
  {
    Domain Dom=MESH.getDomain(0);
    Field coord=MESH.getCoordinateField(Dom);
    coord.setName("coordinates");
    Array ACoord((TYPE="arrayField",ORDER=1,DIM=2,NAME="Champ de coordonnes",
		  DERIV=false,COMPONENT(0)=coord.component(0),
		  COMPONENT(1)=coord.component(1)));
    Array ABrol((TYPE="array",ORDER=1,DIM=2,NAME="orient",COMPONENT(0)=1.0,
		 COMPONENT(1)=1.0e-4));
    Array ADist=accumulation(ACoord,ABrol,1);
    MESH=renumberDefDomainArray(MESH,ADist);
  }
  else if (sortType=="frontal") {
    Domain Bnd4=MESH.getDomain(4);
    Domain Bnd5=MESH.getDomain(5);
    Domain leftBnd=Bnd4+Bnd5;
    MESH=renumberDefDomainFrontal(MESH,leftBnd);
  }
  else if (sortType=="none") {
    // Nothing to do!
  }
  else {
    cerr << "Error in main :" << endl 
      << "   unknown sorting type :  " << sortType << "." << endl ;
    exit(0);
  }
  MESH.write("planeJet.msh2","rsMesh");
  
  // Domains:
  
  Domain Dom,Bnd1,Bnd2,Bnd3,Bnd4,Bnd5;
  Dom=MESH.getDomain(0);
  Dom.setName("main domain");
  Bnd1=MESH.getDomain(1);
  Bnd1.setName("lower boundary");
  Bnd2=MESH.getDomain(2);
  Bnd2.setName("right boundary");
  Bnd3=MESH.getDomain(3);
  Bnd3.setName("upper boundary");
  Bnd4=MESH.getDomain(4);
  Bnd4.setName("left wall boundary");
  Bnd5=MESH.getDomain(5);
  Bnd5.setName("inflow boundary");
  
  // Fields:

  Field coord=MESH.getCoordinateField(Dom);
  coord.setName("coordinates");
  Field pressure((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
		  COMPONENTNAME(0)="p",NAME="pressure"));
  Field velocities((DOM=Dom,INTERPOLATION="quadratic",NBRCOMPONENTS=2,
		    COMPONENTNAME(0)="u",COMPONENTNAME(1)="v",
		    NAME="velocities"));
  
  // TimeDep:
  
  TimeDep TD((NAME="transient",SCHEME="Galerkin",MAXSTEPSNBR=700,
	      DTMIN=0.000001,DTMAX=1.0,DTINIT=0.01,TINIT=0.0,EPSILON=0.001,
	      TFINAL=20.0,INFLU=velocities,ACTIVE=pressure));
  
  // Navier-Stokes equations:
  
  NavierStokes Prob1((param("type","2D2D - V P - F GPn"),
		        param("domain",Dom),
		        param("timeDep",TD),
		        param("coord",coord),
		        param("velo",velocities),
		        param("pressure",pressure),
		        param("integRule","order 5"),
		        param("name","2D Navier-Stokes equations"),
		        param("viscosity",0.5),
		        param("rho",500.0) ));
  
  Array ACoord=Prob1.coordinates;
  TensorCoordSys TCS=Prob1.tcs;
  GlobLoc GL=Prob1.gl;
  
  GlobLoc GL2((COORD=ACoord,GLTYPE="2D1D"));
  Integ INTG((GLOBLOC=GL,TENSORCOORDSYS=TCS,RULE=3));
  Integ INTG2((GLOBLOC=GL2,TENSORCOORDSYS=TCS,RULE=3));  
  
  GlobLoc::setDefaultGL(GL);
    
  // Everything for Navier-Stokes equations:
  
  Array Avelo=Prob1.velocities;  
  Array Apressure=Prob1.pressure;
  Array testMvt((TYPE="arrayTest",TEST=Avelo,
	         NAME="test for momentum equations")); 
    
  // Inflow is calculated explicitly:
  
  Array AY((TYPE="arrayComp",ORDER=0,NAME="Coord Y",
	    COMPONENT()=ACoord.component(1)));
  Array A0((TYPE="array",ORDER=0,NAME="0.03125",COMPONENT()=0.03125));
  Array A1((TYPE="array",ORDER=0,NAME="1.0",COMPONENT()=1.0));
  Array uIn;
  uIn=A1-(1.0/0.03125/0.03125)*pow(AY,2.0);
  Array AveloIn((TYPE="arrayComp",ORDER=1,DIM=2,NAME="vitesses_inflow",
		 TENSORCOORDSYS=TCS,COMPONENT(0)=uIn.component()));
  Constr Prob1BC1((NAME="inflow",TYPE="usual",DOM=Bnd5,TEST=testMvt,
		   EQUATION=Avelo-AveloIn,GLOBLOC=GL2,TENSORCOORDSYS=TCS));
  
  // Other boundary conditions:
  
  Array AveloY((TYPE="arrayField",ORDER=0,NAME="Vy",TENSORCOORDSYS=TCS,
	        COMPONENT(0)=velocities.component(1)));
  Array AveloY0((TYPE="array",ORDER=0,NAME="Vy=0",TENSORCOORDSYS=TCS,
		COMPONENT(0)=0.0));
  Array testMvtY((TYPE="arrayTest",TEST=AveloY,
		  NAME="Equations de mouvement Y"));  
  Constr Prob1BC2((NAME="Vy=0 lower boundary",TYPE="usual",DOM=Bnd1,
		   TEST=testMvtY,EQUATION=AveloY-AveloY0,GLOBLOC=GL2,
		   TENSORCOORDSYS=TCS));
  Array AveloZero((TYPE="array",ORDER=1,DIM=2,NAME="vitessesZero",
		   TENSORCOORDSYS=TCS));  
  Constr Prob1BC3((NAME="V=0 left boundary",TYPE="usual",DOM=Bnd4,
		   TEST=testMvt,EQUATION=Avelo-AveloZero,GLOBLOC=GL2,
		   TENSORCOORDSYS=TCS));

  // Equations for stream function:
  
  Field psi((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
	     COMPONENTNAME(0)="psi",NAME="psi"));
  streamFunction2D Prob2((param("type","2D2D - psi - V"),
		          param("psi",psi),
		          param("coord",coord),
		          param("velo",velocities),
		          param("domain",Dom),
		          param("integRule","order 5"),
		          param("name","stream function equations") ));
  
  Array Astream=Prob2.psi;
  Array testStream=test(Astream);
  Field streamBC((DOM=Bnd4,INTERPOLATION="constantOnDomain",
		   NBRCOMPONENTS=1,COMPONENTNAME(0)="psi",
		   NAME="value of stream function for BC"));
  Array AstreamBC((TYPE="arrayField",ORDER=0,NAME="streamBC",
		   TENSORCOORDSYS=TCS,COMPONENT()=streamBC.component(0)));
  
  Constr Prob2BC1((NAME="Psi=0 lower boundary",TYPE="usual",DOM=Bnd1,
		   TEST=testStream,EQUATION=Astream,GLOBLOC=GL2,
		   TENSORCOORDSYS=TCS));
  Constr Prob2BC2((NAME="Psi upper boundary condition",TYPE="usual",
		   DOM=Bnd4,TEST=testStream,EQUATION=Astream-AstreamBC,
		   GLOBLOC=GL2,TENSORCOORDSYS=TCS));
  
  // Resolution :
  
  Solver NS((PROBLEM=Prob1,CONSTR=Prob1BC1,CONSTR=Prob1BC2,
    CONSTR=Prob1BC3,CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=1));
  Solver STREAM((PROBLEM=Prob2,CONSTR=Prob2BC1,CONSTR=Prob2BC2,
		 CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=1));

  Plot output((DOM=Dom,COORD=ACoord,RESULT=Avelo,RESULT=Apressure,
	       RESULT=Astream));
  
  TD.addInflu(NS);   

  while(TD.isToBeContinued()) {
    NS.NewtonRaphson();
    if (TD.getCurrentStepIndex()==1
	||Int(TD.getCurrentTime()/0.5)!=Int(TD.getPreviousTime()/0.5)) {
      sprintf(buf,"%d",TD.getCurrentStepIndex());
      string filename="planeJet";
      filename+=buf;
      STREAM.NewtonRaphson();
      streamBC.writeValues(cerr);
      output.exec(filename);
    }
  }
}


