
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <string>
#include <string.h>
#include <stdio.h>

#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Field.h"
#include "Integ.h"
#include "Constr.h"
#include "Mesh.h"
#include "param.h"
#include "paramList.h"
#include "Plot.h"
#include "Solver.h"

using namespace std ;
using namespace keyFE2 ;



int main(void) {
  double Vx=1000.0;
  double Vy=500.0;
  double k=100.0;
  bool upwind=true;
  double dInterp=1.0;
  string interpolation="linear";
  
  // Reading and renumbering the mesh:
  
  Mesh MESH;
  MESH.read("square.msh","gmsh2dMesh");
  string sortType="frontal";
  if (sortType=="position")
  {
    Domain Dom=MESH.getDomain(0);
    Field coord=MESH.getCoordinateField(Dom);
    coord.setName("coordinates");
    Array ACoord((TYPE="arrayField",ORDER=1,DIM=2,NAME="Coordinates field",
                  DERIV=false,COMPONENT(0)=coord.component(0),
                  COMPONENT(1)=coord.component(1)));
    Array ABrol((TYPE="array",ORDER=1,DIM=2,NAME="orient",COMPONENT(0)=1.0,
                 COMPONENT(1)=1.0e-4));
    Array ADist=accumulation(ACoord,ABrol,1);
    MESH=renumberDefDomainArray(MESH,ADist);
  }
  else if (sortType=="frontal") {
    Domain Bnd1=MESH.getDomain(1);
    MESH=renumberDefDomainFrontal(MESH,-Bnd1);
  }
  else if (sortType=="none") {
    // Nothing to do!
  }
  else {
    cerr << "Error in main :" << endl 
      << "   unknown sorting type :  " << sortType << "." << endl ;
    exit(0);
  }
  MESH.write("square.msh2","rsMesh");
  
  // Various domains:
  
  Domain Dom,Bnd,Bnd1,Bnd2,Bnd3,Bnd4;
  Dom=MESH.getDomainFromIntId(11);
  Dom.setName("main domain");
  Bnd1=MESH.getDomainFromIntId(7);
  Bnd1.setName("lower boundary");
  Bnd2=MESH.getDomainFromIntId(8);
  Bnd2.setName("right boundary");
  Bnd3=MESH.getDomainFromIntId(9);
  Bnd3.setName("upper boundary");
  Bnd4=MESH.getDomainFromIntId(10);
  Bnd4.setName("left boundary");
  Bnd=Bnd1+Bnd2+Bnd3+Bnd4;
  Bnd.setName("boundary");

  // Coordinate systems and integration:
  
  Field coord=MESH.getCoordinateField(Dom);
  coord.setName("coordinates");
  
  Array ACoord((TYPE="arrayField",ORDER=1,DIM=2,NAME="Coordinates field",
                DERIV=false,COMPONENT(0)=coord.component(0),
                COMPONENT(1)=coord.component(1)));
  
  TensorCoordSys TCS((TCSTYPE="Cartesian2D",COORD=ACoord));
  
  GlobLoc GL((COORD=ACoord,GLTYPE="2D2D"));
  GlobLoc GL2((COORD=ACoord,GLTYPE="2D1D"));
  
  Integ INTG((GLOBLOC=GL,TENSORCOORDSYS=TCS,RULE=3));
  
  GlobLoc::setDefaultGL(GL);
  
  // Everything for the advection-diffusion equation:

  Field u((DOM=Dom,INTERPOLATION=interpolation,NBRCOMPONENTS=1,
           COMPONENTNAME(0)="u",NAME="u"));
  Array U((TYPE="arrayField",ORDER=0,NAME="U",
           TENSORCOORDSYS=TCS,COMPONENT()=u.component(0)));
	   
  Array Avelo((TYPE="array",ORDER=1,DIM=2,NAME="velocities",
               TENSORCOORDSYS=TCS,COMPONENT(0)=Vx,COMPONENT(1)=Vy));
  Array testU((TYPE="arrayTest",TEST=U,NAME="testU"));
  
  Array testConvect;
  if (upwind)
    testConvect=upwindF(Avelo,1.0,k,testU,0.0,GL,dInterp);
  else
    testConvect=testU;
  
  // Advection-diffusion equations:

  Array gradU=grad(U);
  gradU.setName("gradU");
  Array gradTestU=grad(testU);
  gradTestU.setName("gradTestU");
  
  Array Advect=testConvect*contractedMultiplication(Avelo,0,gradU,0);
  Advect.setName("Advective terms");
  Array Conduct=contractedMultiplication(gradTestU,0,k*gradU,0);  
  Conduct.setName("Conductive terms");

  Array AdvecDiff=Advect+Conduct;
  AdvecDiff.setName("Advection-diffusion equations");
  
  Array Prob1=integrate(AdvecDiff,Dom,INTG);
  
  // Boundary conditions on velocities and pressure:
  
  Array X((TYPE="arrayComp",ORDER=0,DIM=1,NAME="coord X",
	   COMPONENT()=ACoord.component(0),TENSORCOORDSYS=TCS));
  Array Y((TYPE="arrayComp",ORDER=0,DIM=1,NAME="coord Y",
	   COMPONENT()=ACoord.component(1),TENSORCOORDSYS=TCS));
	   
  Array uLeftBnd=pwConst2(0.0,0.2,1.0,0.5,0.0,Y);
  Array uOtherBnd=((TYPE="array",ORDER=0,NAME="0",
                    TENSORCOORDSYS=TCS,COMPONENT()=0.0));

  Constr Prob1BC1((NAME="U left boundary",TYPE="usual",DOM=Bnd4,
                   TEST=testU,EQUATION=U-uLeftBnd,GLOBLOC=GL2,
                   TENSORCOORDSYS=TCS));
  Constr Prob1BC2((NAME="U other boundaries",TYPE="usual",DOM=Bnd,
                   TEST=testU,EQUATION=U-uOtherBnd,GLOBLOC=GL2,
                   TENSORCOORDSYS=TCS,LAYER=-100));
  
  // Resolution:
  
  Solver SOLV((PROBLEM=Prob1,CONSTR=Prob1BC1,CONSTR=Prob1BC2,
               CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=6));
  SOLV.NewtonRaphson();
  Plot output((DOM=Dom,COORD=ACoord,TYPE="gmsh",RESULT=U));
  output.exec("brol");
  Plot output2((DOM=Dom,COORD=ACoord,TYPE="openDX",RESULT=U));
  output2.exec("brol");
  Plot output3((DOM=Dom,COORD=ACoord,RESULT=U));
  output3.exec("brol");
}


