
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <string>
#include <stdio.h>
#include <math.h>

#include "beamSection.h"

#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Mesh.h"
#include "Field.h"
#include "Solver.h"
#include "Constr.h"
#include "Object.h"
#include "Plot.h"
#include "Integ.h"
#include "param.h"
#include "paramList.h"
#include "err.h"

#define pi 3.14159265358979323846

using namespace std ;
using namespace keyFE2 ;



beamSection::beamSection(paramList PL) {
  try {
    
    bool silent=true;
    
    // Extraction of parameters:
    
    string meshFileName;
    Double localCoord[3]={0.0,0.0,0.0};
    
    meshFileName=PL.extractStringData("meshFileName",true);
    E=PL.extractDoubleData("YoungModulus",true);
    G=PL.extractDoubleData("ShearModulus",true);
    
    // Reading and renumbering of the Mesh:
    
    try {
//      MESH.read(meshFileName,"emc2Mesh");
      MESH.read(meshFileName,"gmsh2dMesh");
      Domain Dom=MESH.getDefDomain();
      Domain Point=Dom.getFirstNode();
      MESH=renumberDefDomainFrontal(MESH,Point);
    }
    catchAll("beamSection::beamSection","at the end of reading and renumbering of the Mesh.");
    
    // Domains:
    
    Dom=MESH.getDefDomain();
    Bnd=Dom.getBoundary();
    Point=Dom.getFirstNode();
    Object obj=Point.getObject(0);
    
    // Coordinates and related entities (CoG):
    
    {
      Field CoordField=MESH.getCoordinateField(Dom);
      Array ACoord((TYPE="arrayField",ORDER=1,DIM=2,
                    NAME="Coordinates Array",DERIV=false,
                    COMPONENT(0)=CoordField.component(0),
                    COMPONENT(1)=CoordField.component(1)));
      
      GlobLoc GL2D((COORD=ACoord,GLTYPE="2D2D"));
      GlobLoc GL1D((COORD=ACoord,GLTYPE="2D1D"));
      TensorCoordSys TCS((TCSTYPE="Cartesian2D",COORD=ACoord));
      Integ ITG((GLOBLOC=GL2D,TENSORCOORDSYS=TCS,RULENAME="order 5"));
      
      Field CoGField((DOM=Dom,INTERPOLATION="constantOnDomain",
                      NBRCOMPONENTS=2,NAME="CoGField",
                      COMPONENTNAME(0)="yCoG",
                      COMPONENTNAME(1)="zCoG"));
      Array CoGArray((TYPE="arrayField",ORDER=1,DIM=2,NAME="CoGArray",
                      DERIV=true,
                      COMPONENT(0)=CoGField.component(0),
                      COMPONENT(1)=CoGField.component(1)));
      Array testCoGArray=Array((TYPE="arrayTest",TEST=CoGArray,
                                NAME="testCoGArray"));
      Array CoGEquation=
        integrate(accumulation(testCoGArray,CoGArray-ACoord,1),Dom,ITG);
      
      Solver SLV((PROBLEM=CoGEquation,CONVERGENCE=1.0e-4,
                  DIVERGENCE=1.0e4,NITMAX=1,SILENT=silent));
      SLV.NewtonRaphson();
      
      CoGArray.makeAtPoint(obj,localCoord);
      dYc=CoGArray.component(0).value();
      dZc=CoGArray.component(1).value();
      
      CoG=Array((TYPE="array",ORDER=1,DIM=2,NAME="CoG",
                 COMPONENT(0)=dYc,COMPONENT(1)=dZc));
      ACoord1=ACoord;
      ACoord1.setName("ACoord1");
      ACoord2=ACoord1-CoG;
      ACoord2.setName("ACoord2");
    }
    
    // Area and inertias in centered axes + principal directions:
    
    {
      GlobLoc GL2D((COORD=ACoord2,GLTYPE="2D2D"));
      TensorCoordSys TCS((TCSTYPE="Cartesian2D",COORD=ACoord2));
      Integ ITG((GLOBLOC=GL2D,TENSORCOORDSYS=TCS,RULENAME="order 5"));
      
      // Components y' and z' of the centered coordinates:
      
      Array yp((TYPE="arrayComp",ORDER=0,NAME="yp",
                COMPONENT()=ACoord2.component(0)));
      Array zp((TYPE="arrayComp",ORDER=0,NAME="zp",
                COMPONENT()=ACoord2.component(1)));
      
      // Equations for the Area:
      
      Array Unit((TYPE="array",ORDER=0,NAME="Unit",COMPONENT()=1.0));       
      Field AField((DOM=Dom,INTERPOLATION="constantOnDomain",
                    NBRCOMPONENTS=1,
                    COMPONENTNAME(0)="A",NAME="AField"));
      Array AArray((TYPE="arrayField",ORDER=0,NAME="AArray",DERIV=true,
                    COMPONENT()=AField.component(0)));
      Array testAArray=Array((TYPE="arrayTest",TEST=AArray,
                              NAME="testAArray"));       
      Array AEquation=
        testAArray*AArray-integrate(testAArray*Unit,Dom,ITG);
      
      // Equations for Inertias:
      
      Field IyField((DOM=Dom,INTERPOLATION="constantOnDomain",
                     NBRCOMPONENTS=1,
                     COMPONENTNAME(0)="Iy",NAME="IyField"));
      Array IyArray((TYPE="arrayField",ORDER=0,NAME="IyArray",
                     DERIV=true,
                     COMPONENT()=IyField.component(0)));
      Array testIyArray=Array((TYPE="arrayTest",TEST=IyArray,
                               NAME="testIyArray"));       
      Array IyEquation
        =testIyArray*IyArray-integrate(testIyArray*zp*zp,Dom,ITG);
      
      Field IzField((DOM=Dom,INTERPOLATION="constantOnDomain",
                     NBRCOMPONENTS=1,
                     COMPONENTNAME(0)="Iz",NAME="IzField"));
      Array IzArray((TYPE="arrayField",ORDER=0,NAME="IzArray",
                     DERIV=true,
                     COMPONENT()=IzField.component(0)));
      Array testIzArray=Array((TYPE="arrayTest",TEST=IzArray,
                               NAME="testIzArray"));       
      Array IzEquation=
        testIzArray*IzArray-integrate(testIzArray*yp*yp,Dom,ITG);
      
      Field IyzField((DOM=Dom,INTERPOLATION="constantOnDomain",
                      NBRCOMPONENTS=1,
                      COMPONENTNAME(0)="Iyz",NAME="IyzField"));
      Array IyzArray((TYPE="arrayField",ORDER=0,NAME="IyzArray",
                      DERIV=true,
                      COMPONENT()=IyzField.component(0)));
      Array testIyzArray=Array((TYPE="arrayTest",TEST=IyzArray,
                                NAME="testIyzArray"));       
      Array IyzEquation=
        testIyzArray*IyzArray-integrate(testIyzArray*yp*zp,Dom,ITG);
      
      Solver SLV((PROBLEM=AEquation,PROBLEM=IyEquation,
                  PROBLEM=IzEquation,PROBLEM=IyzEquation,
                  CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=1,
                  SILENT=silent));
      SLV.NewtonRaphson();
      
      AArray.makeAtPoint(obj,localCoord);
      dA=AArray.component().value();
      IyArray.makeAtPoint(obj,localCoord);
      Iyp=IyArray.component().value();
      IzArray.makeAtPoint(obj,localCoord);
      Izp=IzArray.component().value();
      IyzArray.makeAtPoint(obj,localCoord);
      Iyzp=IyzArray.component().value();

      A=Array((TYPE="array",ORDER=0,NAME="A",COMPONENT()=dA));
      
      if (fabs(2.0*Iyzp)<1.0e-8*(fabs(Izp)+fabs(Iyp)))
        alfa=0.0;
      else if (fabs(2.0*Iyzp)<fabs(Izp-Iyp))
        alfa=atan((2.0*Iyzp)/(Izp-Iyp))/2.0;
      else
        alfa=0.5*(atan((Iyp-Izp)/(2.0*Iyzp))-pi/2.0);
      
      Array ATransfo((TYPE="array",ORDER=2,DIM=2,NAME="ATransfo",
                      COMPONENT(0,0)=cos(alfa),COMPONENT(0,1)=sin(alfa),
                      COMPONENT(1,0)=-sin(alfa),COMPONENT(1,1)=cos(alfa)));
      ACoord3=contractedMultiplication(ATransfo,1,ACoord2,0);
      ACoord3.setName("ACoord3");
    }
    
    GlobLoc GL2D((COORD=ACoord3,GLTYPE="2D2D"));
    GlobLoc GL1D((COORD=ACoord3,GLTYPE="2D1D"));
    GlobLoc GL0D((COORD=ACoord3,GLTYPE="0D"));
    
    TensorCoordSys TCS((TCSTYPE="Cartesian2D",COORD=ACoord3));
    Integ ITG2D((GLOBLOC=GL2D,TENSORCOORDSYS=TCS,RULENAME="order 5"));
    Integ ITG1D((GLOBLOC=GL1D,TENSORCOORDSYS=TCS,RULENAME="order 5"));
    
    Array ANormal=GL1D.getNormal(TCS);    
    ANormal.setName("ANormal");
    
    Array y((TYPE="arrayComp",ORDER=0,NAME="y",
             COMPONENT()=ACoord3.component(0)));
    Array z((TYPE="arrayComp",ORDER=0,NAME="z",
             COMPONENT()=ACoord3.component(1)));
    
    {
      TensorCoordSys TCS((TCSTYPE="Cartesian2D",COORD=ACoord3));
      Integ ITG((GLOBLOC=GL2D,TENSORCOORDSYS=TCS,RULENAME="order 5"));
      
      // Equations for Iy:
      
      Field IyField((DOM=Dom,INTERPOLATION="constantOnDomain",NBRCOMPONENTS=1,
                     COMPONENTNAME(0)="Iy",NAME="IyField"));
      Array IyArray((TYPE="arrayField",ORDER=0,NAME="IyArray",DERIV=true,
                     COMPONENT()=IyField.component(0)));
      Array testIyArray=Array((TYPE="arrayTest",TEST=IyArray,NAME="testIyArray"));       
      Array IyEquation=testIyArray*IyArray-integrate(testIyArray*z*z,Dom,ITG);
      
      Field IzField((DOM=Dom,INTERPOLATION="constantOnDomain",NBRCOMPONENTS=1,
                     COMPONENTNAME(0)="Iz",NAME="IzField"));
      Array IzArray((TYPE="arrayField",ORDER=0,NAME="IzArray",DERIV=true,
                     COMPONENT()=IzField.component(0)));
      Array testIzArray=Array((TYPE="arrayTest",TEST=IzArray,NAME="testIzArray"));       
      Array IzEquation=testIzArray*IzArray-integrate(testIzArray*y*y,Dom,ITG);
      
      Field IyzField((DOM=Dom,INTERPOLATION="constantOnDomain",NBRCOMPONENTS=1,
                      COMPONENTNAME(0)="Iyz",NAME="IyzField"));
      Array IyzArray((TYPE="arrayField",ORDER=0,NAME="IyzArray",DERIV=true,
                      COMPONENT()=IyzField.component(0)));
      Array testIyzArray=Array((TYPE="arrayTest",TEST=IyzArray,NAME="testIyzArray"));       
      Array IyzEquation=testIyzArray*IyzArray-integrate(testIyzArray*y*z,Dom,ITG);
      
      Solver SLV((PROBLEM=IyEquation,PROBLEM=IzEquation,PROBLEM=IyzEquation,
                  CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=1,SILENT=silent));
      SLV.NewtonRaphson();
      
      IyArray.makeAtPoint(obj,localCoord);
      dIy=IyArray.component().value();
      IzArray.makeAtPoint(obj,localCoord);
      dIz=IzArray.component().value();
      IyzArray.makeAtPoint(obj,localCoord);
      dIyz=IyzArray.component().value();
      Iy=Array((TYPE="array",ORDER=0,NAME="Iy",COMPONENT()=dIy));
      Iz=Array((TYPE="array",ORDER=0,NAME="Iz",COMPONENT()=dIz));
    }
    
    // Calculation of torsional properties:
    
    Array tmpVect;
    
    {
      GlobLoc::setDefaultGL(GL2D);
      Field PhiField((DOM=Dom,INTERPOLATION="quadratic",
                      NBRCOMPONENTS=1,
                      COMPONENTNAME(0)="Phi",NAME="PhiField"));
      APhi=Array((TYPE="arrayField",ORDER=0,DERIV=true,NAME="APhi",
                  COMPONENT()=PhiField.component(0),
                  TENSORCOORDSYS=TCS));
      Array testAPhi=Array((TYPE="arrayTest",TEST=APhi,
                	    NAME="testAPhi"));
      Array PhiEquation=
        integrate(-accumulation(grad(testAPhi),grad(APhi),1),Dom,ITG2D);
      
      Field PField((DOM=Dom,INTERPOLATION="constantOnDomain",
                    NBRCOMPONENTS=1,
                    COMPONENTNAME(0)="P",NAME="PField"));
      Array PArray((TYPE="arrayField",ORDER=0,NAME="PArray",
                    DERIV=true,
                    COMPONENT()=PField.component(0)));
      Array testPArray=Array((TYPE="arrayTest",TEST=PArray,
                	      NAME="testPArray"));       
      Array PEquation=integrate(testPArray*APhi,Dom,ITG2D);
      
      Constr CSTR1((NAME="APhi=0 on one point",TYPE="usual",
                    DOM=Point,TEST=testAPhi,EQUATION=APhi-PArray,
                    GLOBLOC=GL0D,TENSORCOORDSYS=TCS));
      
      Array ATransfo((TYPE="array",ORDER=2,DIM=2,NAME="ATransfo",
                      COMPONENT(0,0)=0.0,COMPONENT(0,1)=-1.0,
                      COMPONENT(1,0)=1.0,COMPONENT(1,1)=0.0));
      tmpVect=contractedMultiplication(ATransfo,1,ACoord3,0);
      GlobLoc::setDefaultGL(GL1D);
      Array PhiEquation2=
        integrate(testAPhi*accumulation(-tmpVect,ANormal,1),Bnd,ITG1D);
      
      GlobLoc::setDefaultGL(GL2D);
      Field JField((DOM=Dom,INTERPOLATION="constantOnDomain",
                    NBRCOMPONENTS=1,
                    COMPONENTNAME(0)="J",NAME="JField"));
      Array JArray((TYPE="arrayField",ORDER=0,NAME="JArray",
                    DERIV=true,
                    COMPONENT()=JField.component(0)));
      Array testJArray=Array((TYPE="arrayTest",TEST=JArray,
                	      NAME="testJArray"));       
      Array JEquation=testJArray*JArray
        -integrate(testJArray*(y*y+z*z
           +accumulation(tmpVect,grad(APhi))),Dom,ITG2D);
	   
      Field CwField((DOM=Dom,INTERPOLATION="constantOnDomain",
                     NBRCOMPONENTS=1,
                     COMPONENTNAME(0)="Cw",NAME="CwField"));
      Array CwArray((TYPE="arrayField",ORDER=0,NAME="CwArray",
                     DERIV=true,
                     COMPONENT()=CwField.component(0)));
      Array testCwArray=Array((TYPE="arrayTest",TEST=CwArray,
                	       NAME="testCwArray"));       
      Array CwEquation=testCwArray*CwArray
        -integrate(testCwArray*accumulation(APhi,APhi),Dom,ITG2D);
      
      Field ToCField((DOM=Dom,INTERPOLATION="constantOnDomain",
                      NBRCOMPONENTS=2,
                      COMPONENTNAME(0)="yToC",COMPONENTNAME(1)="zToC",
                      NAME="ToCField"));
      Array ToCArray((TYPE="arrayField",ORDER=1,DIM=2,NAME="ToCArray",
                      DERIV=true,
                      COMPONENT(0)=ToCField.component(0),
                      COMPONENT(1)=ToCField.component(1)));
      Array testToCArray=Array((TYPE="arrayTest",TEST=ToCArray,
                		NAME="testToCArray"));
      Array ToCEquation=accumulation(testToCArray,ToCArray)
        -integrate(accumulation(testToCArray,ACoord3*APhi,1),Dom,ITG2D);
      
      Solver SLV((PROBLEM=PhiEquation,PROBLEM=PEquation,
                  PROBLEM=PhiEquation2,
                  PROBLEM=JEquation,PROBLEM=CwEquation,PROBLEM=ToCEquation,
                  CONSTR=CSTR1,
                  CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=6,
                  SILENT=silent));
      SLV.NewtonRaphson();
      
      JArray.makeAtPoint(obj,localCoord);
      dJ=JArray.component().value();
      J=Array((TYPE="array",ORDER=0,NAME="J",COMPONENT()=dJ));
      
      CwArray.makeAtPoint(obj,localCoord);
      dCw=CwArray.component().value();
      
      ToCArray.makeAtPoint(obj,localCoord);
      ToCy=-ToCArray.component(1).value()/Iy.component().value();
      ToCz=ToCArray.component(0).value()/Iz.component().value();
      ToC=Array((TYPE="array",ORDER=1,DIM=2,NAME="ToC",
                 COMPONENT(0)=ToCy,COMPONENT(1)=ToCz));
		 
      dCwPat=dCw-ToCy*dIy*ToCy-ToCz*dIz*ToCz-2.0*ToCy*dIyz*ToCz;
    }
    
    // Calculation of Y shear properties:
    
    {
      GlobLoc::setDefaultGL(GL2D);
      Field fField((DOM=Dom,INTERPOLATION="quadratic",NBRCOMPONENTS=1,
                    COMPONENTNAME(0)="f",NAME="fField"));
      Af=Array((TYPE="arrayField",ORDER=0,DERIV=true,NAME="Af",
                COMPONENT()=fField.component(0),TENSORCOORDSYS=TCS));
      Array testAf=Array((TYPE="arrayTest",TEST=Af,NAME="testAf"));
      Array fEquation=integrate(-accumulation(grad(testAf),grad(Af),1)+testAf*(1.0/dIz)*y,
                                Dom,ITG2D);
      
      Field PField((DOM=Dom,INTERPOLATION="constantOnDomain",
                    NBRCOMPONENTS=1,COMPONENTNAME(0)="P",NAME="PField"));
      Array PArray((TYPE="arrayField",ORDER=0,NAME="PArray",DERIV=true,
                    COMPONENT()=PField.component(0)));
      Array testPArray=Array((TYPE="arrayTest",TEST=PArray,
                              NAME="testPArray"));       
      Array PEquation=integrate(testPArray*Af,Dom,ITG2D);
      Constr CSTR1((NAME="Af=0 on one point",TYPE="usual",DOM=Point,
                    TEST=testAf,EQUATION=Af-PArray,GLOBLOC=GL0D,
                    TENSORCOORDSYS=TCS));
      
      GlobLoc::setDefaultGL(GL2D);
      Field JField((DOM=Dom,INTERPOLATION="constantOnDomain",
                    NBRCOMPONENTS=1,COMPONENTNAME(0)="J",NAME="JField"));
      Array JArray((TYPE="arrayField",ORDER=0,NAME="JArray",DERIV=true,
                    COMPONENT()=JField.component(0)));
      Array testJArray=Array((TYPE="arrayTest",TEST=JArray,
                              NAME="testJArray"));       
      Array JEquation=
	testJArray*JArray-integrate(testJArray*(dA/dIz)*y*Af,Dom,ITG2D);
      
      Solver SLV((PROBLEM=fEquation,PROBLEM=PEquation,PROBLEM=JEquation,
                  CONSTR=CSTR1,CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=1,
                  SILENT=silent));
      SLV.NewtonRaphson();
      
      JArray.makeAtPoint(obj,localCoord);
      dky=1.0/JArray.component().value();
      ky=Array((TYPE="array",ORDER=0,NAME="ky",COMPONENT()=dky));
    }
    
    // Calculation of Z shear properties:
    
    {
      GlobLoc::setDefaultGL(GL2D);
      Field gField((DOM=Dom,INTERPOLATION="quadratic",NBRCOMPONENTS=1,
                    COMPONENTNAME(0)="g",NAME="gField"));
      Ag=Array((TYPE="arrayField",ORDER=0,DERIV=true,NAME="Ag",
                COMPONENT()=gField.component(0),TENSORCOORDSYS=TCS));
      Array testAg=Array((TYPE="arrayTest",TEST=Ag,NAME="testAg"));
      Array gEquation=integrate(-accumulation(grad(testAg),grad(Ag),1)+testAg*(1.0/dIy)*z,
                                Dom,ITG2D);
      
      Field PField((DOM=Dom,INTERPOLATION="constantOnDomain",NBRCOMPONENTS=1,
                    COMPONENTNAME(0)="P",NAME="PField"));
      Array PArray((TYPE="arrayField",ORDER=0,NAME="PArray",DERIV=true,
                    COMPONENT()=PField.component(0)));
      Array testPArray=Array((TYPE="arrayTest",TEST=PArray,NAME="testPArray"));       
      Array PEquation=integrate(testPArray*Ag,Dom,ITG2D);
      Constr CSTR1((NAME="Ag=0 on one point",TYPE="usual",DOM=Point,
                    TEST=testAg,EQUATION=Ag-PArray,GLOBLOC=GL0D,
                    TENSORCOORDSYS=TCS));
      
      GlobLoc::setDefaultGL(GL2D);
      Field JField((DOM=Dom,INTERPOLATION="constantOnDomain",NBRCOMPONENTS=1,
                    COMPONENTNAME(0)="J",NAME="JField"));
      Array JArray((TYPE="arrayField",ORDER=0,NAME="JArray",DERIV=true,
                    COMPONENT()=JField.component(0)));
      Array testJArray=Array((TYPE="arrayTest",TEST=JArray,NAME="testJArray"));       
      Array JEquation=testJArray*JArray-integrate(testJArray*(dA/dIy)*z*Ag,Dom,ITG2D);
      
      Solver SLV((PROBLEM=gEquation,PROBLEM=PEquation,PROBLEM=JEquation,
                  CONSTR=CSTR1,
                  CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=1,SILENT=silent));
      SLV.NewtonRaphson();
      
      JArray.makeAtPoint(obj,localCoord);
      dkz=1.0/JArray.component().value();
      kz=Array((TYPE="array",ORDER=0,NAME="kz",COMPONENT()=dkz));
    }
    
    cerr.setf(ios::scientific);
    cerr << setprecision(7);
    
    cerr << endl << endl ;
    cerr << "Geometric Center of the section (wrt Mesh Coordinates): " 
      << endl << endl ;
    cerr << "   yCoG = " << dYc << endl ;
    cerr << "   zCoG = " << dZc << endl << endl ;
    cerr << "Centered inertias of the section: " << endl << endl ;
    cerr << "   Iyp  = " << Iyp << endl ;
    cerr << "   Izp  = " << Izp << endl ;
    cerr << "   Iyzp = " << Iyzp << endl << endl ;
    cerr << "Rotation for principal axes: " << endl << endl ;
    cerr << "   alfa = " << alfa << " (" << alfa*180.0/pi << ")" 
      << endl << endl ;
     
    cerr << "Section properties in principal axes: " << endl << endl ;
    cerr << "   A = " << dA << endl << endl ;
    cerr << "   ky = " << dky << endl ;
    cerr << "   kz = " << dkz << endl << endl ;    
    cerr << "   Iy  = " << dIy << endl ;
    cerr << "   Iz  = " << dIz << endl ;
    cerr << "   Iyz = " << dIyz << endl << endl ;
    cerr << "   J = " << dJ << endl ;
    cerr << "   Cw = " << dCw << endl ;
    cerr << "   CwPat = " << dCwPat << endl ;
    cerr << "   ToCy = " << ToCy << endl ;
    cerr << "   ToCz = " << ToCz << endl << endl ;

    Nx=Array((TYPE="array",ORDER=0,NAME="Nx",COMPONENT()=0.0));
    Ty=Array((TYPE="array",ORDER=0,NAME="Ty",COMPONENT()=0.0));
    Tz=Array((TYPE="array",ORDER=0,NAME="Tz",COMPONENT()=0.0));
    Mt=Array((TYPE="array",ORDER=0,NAME="Mt",COMPONENT()=0.0));
    Mfy=Array((TYPE="array",ORDER=0,NAME="Mfy",COMPONENT()=0.0));
    Mfz=Array((TYPE="array",ORDER=0,NAME="Mfz",COMPONENT()=0.0));
      
    Sxx=(1.0/dA)*Nx+(1.0/dIy)*Mfy*z-(1.0/dIz)*Mfz*y;
    Sxx.setName("Sxx");
    Sshear=Ty*grad(Af)+Tz*grad(Ag)+(1.0/dJ)*Mt*(grad(APhi)+tmpVect);
    Sshear.setName("Sshear");
    Array Sxy((TYPE="arrayComp",ORDER=0,NAME="Sxy",COMPONENT()=Sshear.component(0)));
    Array Sxz((TYPE="arrayComp",ORDER=0,NAME="Sxz",COMPONENT()=Sshear.component(1)));
    SVM=sqrt(Sxx*Sxx+3.0*(Sxy*Sxy+Sxz*Sxz));
    SVM.setName("SVM");
  }
  catchAll("beamSection::beamSection","at the end of constructor.");
  
}



beamSection::~beamSection(void) {
  ACoord1.setData(NULL);
  CoG.setData(NULL);
  A.setData(NULL);
  ACoord2.setData(NULL);
  ACoord3.setData(NULL);
  Iy.setData(NULL);
  Iz.setData(NULL);
  APhi.setData(NULL);
  ToC.setData(NULL);
  J.setData(NULL);
  Af.setData(NULL);
  ky.setData(NULL);
  Ag.setData(NULL);
  kz.setData(NULL);
}



void beamSection::plotResults(string filename,
                              Double Fx,Double Fy,Double Fz,
                              Double Mx,Double My,Double Mz) {
  Nx.component()=Fx;
  Ty.component()=Fy;
  Tz.component()=Fz;
  Mt.component()=Mx-ToCy*Fz+ToCz*Fy;
  Mfy.component()=My;
  Mfz.component()=Mz;
  
  Plot output((DOM=Dom,COORD=ACoord3,RESULT=Sxx,RESULT=Sshear,
	       RESULT=SVM,RESULT=APhi));
  output.exec(filename);
}
