
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <string>
#include <string.h>
#include <stdio.h>


#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Field.h"
#include "Integ.h"
#include "Constr.h"
#include "Mesh.h"
#include "param.h"
#include "paramList.h"
#include "Plot.h"
#include "Solver.h"
#include "TimeDep.h"

using namespace std ;
using namespace keyFE2 ;

int main(void) {
  char buf[256];

  Mesh MESH;
  
  // Reading and renumbering of the mesh:
  
  MESH.read("planeJet.msh","rsMesh");
  string sortType="frontal";
  if (sortType=="position")
  {
    Domain Dom=MESH.getDomain(0);
    Field coord=MESH.getCoordinateField(Dom);
    coord.setName("coordinates");
    Array ACoord((TYPE="arrayField",ORDER=1,DIM=2,NAME="Coordinates field",
                  DERIV=false,COMPONENT(0)=coord.component(0),
                  COMPONENT(1)=coord.component(1)));
    Array ABrol((TYPE="array",ORDER=1,DIM=2,NAME="orient",COMPONENT(0)=1.0,
                 COMPONENT(1)=1.0e-4));
    Array ADist=accumulation(ACoord,ABrol,1);
    MESH=renumberDefDomainArray(MESH,ADist);
  }
  else if (sortType=="frontal") {
    Domain Bnd4=MESH.getDomain(4);
    Domain Bnd5=MESH.getDomain(5);
    Domain leftBnd=Bnd4+Bnd5;
    MESH=renumberDefDomainFrontal(MESH,leftBnd);
  }
  else if (sortType=="none") {
    // Nothing to do !
  }
  else {
    cerr << "Error in main:" << endl 
      << "   unknown sorting type:  " << sortType << "." << endl ;
    exit(0);
  }
  MESH.write("planeJet.msh2","rsMesh");
  
  // Various domains:
  
  Domain Dom,Bnd1,Bnd2,Bnd3,Bnd4,Bnd5;
  Dom=MESH.getDomain(0);
  Dom.setName("main domain");
  Bnd1=MESH.getDomain(1);
  Bnd1.setName("lower boundary");
  Bnd2=MESH.getDomain(2);
  Bnd2.setName("right boundary");
  Bnd3=MESH.getDomain(3);
  Bnd3.setName("upper boundary");
  Bnd4=MESH.getDomain(4);
  Bnd4.setName("left wall boundary");
  Bnd5=MESH.getDomain(5);
  Bnd5.setName("inflow boundary");
  
  // Everything for coordinates and integrations:
  
  Field coord=MESH.getCoordinateField(Dom);
  coord.setName("coordinates");
  Array ACoord((TYPE="arrayField",ORDER=1,DIM=2,NAME="Coordinates field",
                DERIV=false,COMPONENT(0)=coord.component(0),
                COMPONENT(1)=coord.component(1)));

  TensorCoordSys TCS((TCSTYPE="Cartesian2D",COORD=ACoord));

  GlobLoc GL((COORD=ACoord,GLTYPE="2D2D"));
  GlobLoc GL2((COORD=ACoord,GLTYPE="2D1D"));

  Integ INTG((GLOBLOC=GL,TENSORCOORDSYS=TCS,RULE=3));
  Integ INTG2((GLOBLOC=GL2,TENSORCOORDSYS=TCS,RULE=3));  
  
  GlobLoc::setDefaultGL(GL);

  // Everything for Navier-Stokes equations:

  Field pressure((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
                  COMPONENTNAME(0)="p",NAME="pressure"));
  Field velocities((DOM=Dom,INTERPOLATION="quadratic",NBRCOMPONENTS=2,
                    COMPONENTNAME(0)="u",COMPONENTNAME(1)="v",
                    NAME="velocities"));
  Array Avelo((TYPE="arrayField",ORDER=1,DIM=2,NAME="velocities",
               TENSORCOORDSYS=TCS,COMPONENT(0)=velocities.component(0),
               COMPONENT(1)=velocities.component(1)));
  Array Apressure((TYPE="arrayField",ORDER=0,NAME="pressure",
                   TENSORCOORDSYS=TCS,COMPONENT()=pressure.component(0)));
  
  Array testMvt((TYPE="arrayTest",TEST=Avelo,NAME="Test for momentum equations"));  
  Array testDiv((TYPE="arrayTest",TEST=Apressure,
                 NAME="Test for continuity equations"));
  
  Array Unit((TYPE="array",ORDER=2,DIM=2,COMPONENT(0,0)=1.0,
              COMPONENT(1,1)=1.0,NAME="Unit tensor",TENSORCOORDSYS=TCS));
  Array rho((TYPE="array",ORDER=0,NAME="density",TENSORCOORDSYS=TCS,
             COMPONENT()=500.0));
  
  // TimeDep:
  
  TimeDep TD((NAME="transient",SCHEME="Galerkin",MAXSTEPSNBR=700,
              DTMIN=0.000001,DTMAX=1.0,DTINIT=0.01,TINIT=0.0,EPSILON=0.001,
              TFINAL=20.0,INFLU=velocities,ACTIVE=pressure));
  
  // Navier-Stokes equations:

  Array gradVelo=grad(Avelo);
  gradVelo.setName("Gradient of velocities");
  Array Convect=rho*(timeDeriv(Avelo,TD)+
                     contractedMultiplication(Avelo,0,gradVelo,1));
  Convect.setName("Convective terms");
  Array gradMvt=grad(testMvt);
  gradMvt.setName("gradMvt");
  
  Array divV=contract(gradVelo);
  divV.setName("divergence of velocities");
  
  Array D=0.5*(gradVelo+transpose(gradVelo));
  D.setName("strain rate tensor");
  Array T=D-Apressure*Unit;
  T.setName("Cauchy stress tensor");
  Array NavierStokes=accumulation(gradMvt,T)+accumulation(testMvt,Convect)
                     +testDiv*divV;
  NavierStokes.setName("Navier-Stokes equations");
  
  Array Prob1=integrate(NavierStokes,Dom,INTG);
  
  // Inflow is explicitly calculated:
  
  Array AY((TYPE="arrayComp",ORDER=0,NAME="Coord Y",
            COMPONENT()=ACoord.component(1)));
  Array A0((TYPE="array",ORDER=0,NAME="0.03125",COMPONENT()=0.03125));
  Array A1((TYPE="array",ORDER=0,NAME="1.0",COMPONENT()=1.0));
  Array uIn;
  uIn=A1-(1.0/0.03125/0.03125)*pow(AY,2.0);
  Array AveloIn((TYPE="arrayComp",ORDER=1,DIM=2,NAME="vitesses_inflow",
                 TENSORCOORDSYS=TCS,COMPONENT(0)=uIn.component()));
  Constr Prob1CF1((NAME="inflow",TYPE="usual",DOM=Bnd5,TEST=testMvt,
                   EQUATION=Avelo-AveloIn,GLOBLOC=GL2,TENSORCOORDSYS=TCS));
  
  // Other boundary conditions:
  
  Array AveloY((TYPE="arrayField",ORDER=0,NAME="Vy",TENSORCOORDSYS=TCS,
                COMPONENT(0)=velocities.component(1)));
  Array AveloY0((TYPE="array",ORDER=0,NAME="Vy=0",TENSORCOORDSYS=TCS,
                COMPONENT(0)=0.0));
  Array testMvtY((TYPE="arrayTest",TEST=AveloY,
                  NAME="Y component of test for momentum equations"));  
  Constr Prob1CF2((NAME="Vy=0 lower boundary",TYPE="usual",DOM=Bnd1,
                   TEST=testMvtY,EQUATION=AveloY-AveloY0,GLOBLOC=GL2,
                   TENSORCOORDSYS=TCS));
  Array AveloZero((TYPE="array",ORDER=1,DIM=2,NAME="Zero velocities",
                   TENSORCOORDSYS=TCS));  
  Constr Prob1CF3((NAME="V=0 left boundary",TYPE="usual",DOM=Bnd4,
                   TEST=testMvt,EQUATION=Avelo-AveloZero,GLOBLOC=GL2,
                   TENSORCOORDSYS=TCS));

  // Stream functions calculation:
  
  Field stream((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
                 COMPONENTNAME(0)="psi",NAME="stream"));
  Field streamBC((DOM=Bnd4,INTERPOLATION="constantOnDomain",
                   NBRCOMPONENTS=1,COMPONENTNAME(0)="psi",
                   NAME="stream for BC"));
  Array Astream((TYPE="arrayField",ORDER=0,NAME="stream",TENSORCOORDSYS=TCS,
                  COMPONENT()=stream.component(0)));
  Array AstreamBC((TYPE="arrayField",ORDER=0,NAME="streamBC",
                    TENSORCOORDSYS=TCS,COMPONENT()=streamBC.component(0)));
  Array Avelo2((TYPE="arrayField",ORDER=1,DIM=2,NAME="velocities (2)",DERIV=false,
                TENSORCOORDSYS=TCS,COMPONENT(0)=velocities.component(0),
                COMPONENT(1)=velocities.component(1)));
     // (Avelo 2 is the same as Avelo, but no derivation wrt nodal coordinates.)
  Array U((TYPE="arrayComp",ORDER=0,NAME="u",
           COMPONENT()=Avelo2.component(0)));
  Array V((TYPE="arrayComp",ORDER=0,NAME="v",
           COMPONENT()=Avelo2.component(1)));
  Array mV=-1.0*V;
  Array tmpVelo((TYPE="arrayComp",ORDER=1,DIM=2,NAME="special",
                 TENSORCOORDSYS=TCS,COMPONENT(0)=mV.component(),
                 COMPONENT(1)=U.component()));
  Array gradPsi=grad(Astream);
  Array testCourant((TYPE="arrayTest",TEST=Astream,
                     NAME="Stream function equations"));  
  Array gradCourant=grad(testCourant);
  Array eqCourant=gradPsi-tmpVelo;
  Array equationCourant=accumulation(gradCourant,eqCourant);
  Array Prob2=integrate(equationCourant,Dom,INTG);
  
  Constr Prob2CF1((NAME="Psi=0 lower boundary",TYPE="usual",DOM=Bnd1,
                   TEST=testCourant,EQUATION=Astream,GLOBLOC=GL2,
                   TENSORCOORDSYS=TCS));
  Constr Prob2CF2((NAME="Psi upper boundary condition",TYPE="usual",
                   DOM=Bnd4,TEST=testCourant,EQUATION=Astream-AstreamBC,
                   GLOBLOC=GL2,TENSORCOORDSYS=TCS));
  
  // Solvers:
  
  Solver NS((PROBLEM=Prob1,CONSTR=Prob1CF1,CONSTR=Prob1CF2,
             CONSTR=Prob1CF3,CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=1));
  Solver STREAM((PROBLEM=Prob2,CONSTR=Prob2CF1,CONSTR=Prob2CF2,
                 CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=1));  
  Plot output((DOM=Dom,COORD=ACoord,RESULT=Avelo,RESULT=Apressure,
               RESULT=Astream));
  
  TD.addInflu(NS); 
  
  while(TD.isToBeContinued()) {
    NS.NewtonRaphson();
    if (TD.getCurrentStepIndex()==1
        ||Int(TD.getCurrentTime()/0.5)!=Int(TD.getPreviousTime()/0.5)) {
      sprintf(buf,"%d",TD.getCurrentStepIndex());
      string filename="planeJet";
      filename+=buf;
      STREAM.NewtonRaphson();
      streamBC.writeValues(cerr);
      output.exec(filename);
    }
  }
}


