
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <string>
#include <string.h>
#include <stdio.h>

#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Field.h"
#include "Integ.h"
#include "Constr.h"
#include "Mesh.h"
#include "param.h"
#include "paramList.h"
#include "Plot.h"
#include "Result.h"
#include "Solver.h"

using namespace std ;
using namespace keyFE2 ;



int main(void) {
  char buf[256];
  int i;
  
  bool upwind=true;
  bool quadV=false;
  double dInterp;
  
  // Reading and renumbering of the mesh :
  
  Mesh MESH;
  MESH.read("squareStep.msh","rsMesh");
  string sortType="position";
  if (sortType=="position")
  {
    Domain Dom=MESH.getDomain(0);
    Field coord=MESH.getCoordinateField(Dom);
    coord.setName("coordinates");
    Array ACoord((TYPE="arrayField",ORDER=1,DIM=2,NAME="Champ de coordonnes",
		  DERIV=false,COMPONENT(0)=coord.component(0),
		  COMPONENT(1)=coord.component(1)));
    Array ABrol((TYPE="array",ORDER=1,DIM=2,NAME="orient",COMPONENT(0)=1.0,
		 COMPONENT(1)=1.0e-4));
    Array ADist=accumulation(ACoord,ABrol,1);
    MESH=renumberDefDomainArray(MESH,ADist);
  }
  else if (sortType=="frontal") {
    Domain Bnd1=MESH.getDomain(1);
    MESH=renumberDefDomainFrontal(MESH,-Bnd1);
  }
  else if (sortType=="none") {
    // Nothing to do !
  }
  else {
    cerr << "Error in main :" << endl 
      << "   unknown sorting type :  " << sortType << "." << endl ;
    exit(0);
  }
  MESH.write("squareStep.msh2","rsMesh");
  
  // Domains :
  
  Domain Dom,Bnd,Bnd1,Bnd2,Bnd3,Bnd4;
  Dom=MESH.getDomain(0);
  Dom.setName("main domain");
  Bnd1=MESH.getDomain(1);
  Bnd1.setName("left boundary");
  Bnd2=MESH.getDomain(2);
  Bnd2.setName("lower boundary");
  Bnd3=MESH.getDomain(3);
  Bnd3.setName("right boundary");
  Bnd4=MESH.getDomain(4);
  Bnd4.setName("upper boundary");
  Bnd=Dom.getBoundary();
  Bnd.setName("external boundary");
  
  // Coordinates, Integration rules :
  
  Field coord=MESH.getCoordinateField(Dom);
  coord.setName("coordinates");
  Array ACoord((TYPE="arrayField",ORDER=1,DIM=2,NAME="Champ de coordonnes",
		DERIV=false,COMPONENT(0)=coord.component(0),
		COMPONENT(1)=coord.component(1)));

  TensorCoordSys TCS((TCSTYPE="Cartesian2D",COORD=ACoord));

  GlobLoc GL((COORD=ACoord,GLTYPE="2D2D"));
  GlobLoc GL2((COORD=ACoord,GLTYPE="2D1D"));

  Integ INTG((GLOBLOC=GL,TENSORCOORDSYS=TCS,RULE=3));
  Integ INTG2((GLOBLOC=GL2,TENSORCOORDSYS=TCS,RULE=3));  
  
  GlobLoc::setDefaultGL(GL);

  // Everything for Navier-Stokes equations :
  
  Field pressure,velocities;
  if (quadV) {
    pressure=Field((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
		    COMPONENTNAME(0)="p",NAME="pressure"));
    velocities=Field((DOM=Dom,INTERPOLATION="quadratic",NBRCOMPONENTS=2,
		      COMPONENTNAME(0)="u",COMPONENTNAME(1)="v",
		      NAME="velocities"));
    dInterp=2.0;
  }
  else {
    pressure=Field((DOM=Dom,INTERPOLATION="constantPerElement",NBRCOMPONENTS=1,
		    COMPONENTNAME(0)="p",NAME="pressure"));
    velocities=Field((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=2,
		      COMPONENTNAME(0)="u",COMPONENTNAME(1)="v",
		      NAME="velocities"));
    dInterp=1.0;
  }
  Array Avelo((TYPE="arrayField",ORDER=1,DIM=2,NAME="vitesses",
	       TENSORCOORDSYS=TCS,COMPONENT(0)=velocities.component(0),
	       COMPONENT(1)=velocities.component(1)));
  Array Apressure((TYPE="arrayField",ORDER=0,NAME="pression",
		   TENSORCOORDSYS=TCS,COMPONENT()=pressure.component(0)));
  
  Array testMvt((TYPE="arrayTest",TEST=Avelo,NAME="Equations de mouvement"));  
  Array testDiv((TYPE="arrayTest",TEST=Apressure,NAME="Equations de continuit"));
  
  Array Unit((TYPE="array",ORDER=2,DIM=2,COMPONENT(0,0)=1.0,
	      COMPONENT(1,1)=1.0,NAME="Unit tensor",TENSORCOORDSYS=TCS));
  Array rho((TYPE="array",ORDER=0,NAME="density",TENSORCOORDSYS=TCS));
  
  // Navier-Stokes equations :

  Array gradVelo=grad(Avelo);
  gradVelo.setName("Champ gradients de vitesses");
  Array Convect=rho*contractedMultiplication(Avelo,0,gradVelo,1);
  Convect.setName("Termes convectifs");
  Array gradMvt=grad(testMvt);
  gradMvt.setName("gradMvt");
  
  Array divV=contract(gradVelo);
  divV.setName("divergence des vitesses");
  
  Array testConvect;
  if (upwind)
    testConvect=upwindF(Avelo,rho,1.0,testMvt,0.0,GL,dInterp);
  else
    testConvect=testMvt;
  
  Array D=0.5*(gradVelo+transpose(gradVelo));
  D.setName("Taux de dformation");
  Array T=2.0*D-Apressure*Unit; // viscosity=1.0
  T.setName("Tenseur des tensions de Cauchy");
  Array NavierStokes=accumulation(gradMvt,T)+accumulation(testConvect,Convect)
    +testDiv*divV;
  NavierStokes.setName("Equations de Navier-Stokes");
  
  Array Prob1=integrate(NavierStokes,Dom,INTG);
  
  // Inflow is explicitly calculated :
  
  Array AY((TYPE="arrayComp",ORDER=0,NAME="Coord Y",
	    COMPONENT()=ACoord.component(1)));
  Array A05((TYPE="array",ORDER=0,NAME="0.5",COMPONENT()=0.5));
  Array A025((TYPE="array",ORDER=0,NAME="0.25",COMPONENT()=0.25));
  Array uIn;
  uIn=6.0*(A025-pow(AY-A05,2.0));
  Array AveloIn((TYPE="arrayComp",ORDER=1,DIM=2,NAME="vitesses_inflow",
		 TENSORCOORDSYS=TCS,COMPONENT(0)=uIn.component()));
  Constr Prob1CF1((NAME="inflow",TYPE="usual",DOM=Bnd1,TEST=testMvt,
		   EQUATION=Avelo-AveloIn,GLOBLOC=GL2,TENSORCOORDSYS=TCS));
  
  // Other boundary conditions :
  
  Array AveloZero((TYPE="array",ORDER=1,DIM=2,NAME="vitessesZero",
		   TENSORCOORDSYS=TCS));
  Constr Prob1CF2((NAME="V=0 lower boundary",TYPE="usual",DOM=Bnd2,
		   TEST=testMvt,EQUATION=Avelo-AveloZero,GLOBLOC=GL2,
		   TENSORCOORDSYS=TCS));
  Constr Prob1CF3((NAME="V=0 upper boundary",TYPE="usual",DOM=Bnd4,
		   TEST=testMvt,EQUATION=Avelo-AveloZero,GLOBLOC=GL2,
		   TENSORCOORDSYS=TCS));

  // Calculation of stream function :
  
  Field courant((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
		 COMPONENTNAME(0)="psi",NAME="courant"));
  Field courantBC((DOM=Bnd4,INTERPOLATION="constantOnDomain",
		   NBRCOMPONENTS=1,COMPONENTNAME(0)="psi",
		   NAME="courant for BC"));
  Array Acourant((TYPE="arrayField",ORDER=0,NAME="courant",TENSORCOORDSYS=TCS,
		  COMPONENT()=courant.component(0)));
  Array AcourantBC((TYPE="arrayField",ORDER=0,NAME="courantBC",
		    TENSORCOORDSYS=TCS,COMPONENT()=courantBC.component(0)));
  Array Avelo2((TYPE="arrayField",ORDER=1,DIM=2,NAME="vitesses",DERIV=false,
	        TENSORCOORDSYS=TCS,COMPONENT(0)=velocities.component(0),
	        COMPONENT(1)=velocities.component(1)));
     // (same as Avelo, but no derivation wrt nodal values)
  Array U((TYPE="arrayComp",ORDER=0,NAME="u",
	   COMPONENT()=Avelo2.component(0)));
  Array V((TYPE="arrayComp",ORDER=0,NAME="v",
	   COMPONENT()=Avelo2.component(1)));
  Array mV=-1.0*V;
  Array tmpVelo((TYPE="arrayComp",ORDER=1,DIM=2,NAME="special",
		 TENSORCOORDSYS=TCS,COMPONENT(0)=mV.component(),
		 COMPONENT(1)=U.component()));
  Array gradPsi=grad(Acourant);
  Array testCourant((TYPE="arrayTest",TEST=Acourant,
		     NAME="Equations de courant"));  
  Array gradCourant=grad(testCourant);
  Array eqCourant=gradPsi-tmpVelo;
  Array equationCourant=accumulation(gradCourant,eqCourant);
  Array Prob2=integrate(equationCourant,Dom,INTG);
  Prob2.setName("fonction de courant");
  
  Constr Prob2CF1((NAME="Psi=0 lower boundary",TYPE="usual",DOM=Bnd2,
		   TEST=testCourant,EQUATION=Acourant,GLOBLOC=GL2,
		   TENSORCOORDSYS=TCS));
  Constr Prob2CF2((NAME="Psi upper boundary condition",TYPE="usual",
		   DOM=Bnd4,TEST=testCourant,EQUATION=Acourant-AcourantBC,
		   GLOBLOC=GL2,TENSORCOORDSYS=TCS));
  
  // Rsolution :
  
  Plot output((DOM=Dom,COORD=ACoord,RESULT=Avelo,RESULT=Apressure,
	       RESULT=Acourant,TYPE="gmsh"));
    
  for (i=0;i<=20;i++) {
    rho.component()=20.0*i;

    cerr << "New continuation step with rho = " << 20.0*i << endl ;
    
    Solver NS((PROBLEM=Prob1,CONSTR=Prob1CF1,CONSTR=Prob1CF2,
      CONSTR=Prob1CF3,CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=6,
      DELAY=30));
    NS.NewtonRaphson();

    Solver STREAM((PROBLEM=Prob2,CONSTR=Prob2CF1,CONSTR=Prob2CF2,
		   CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=1));
    STREAM.NewtonRaphson();
    courantBC.writeValues(cerr);
    
    sprintf(buf,"%d",i);
    string filename="squareStep";
    filename+=buf;
    output.exec(filename);
  }
}


