
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <string>
#include <string.h>
#include <stdio.h>

#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Field.h"
#include "Integ.h"
#include "Constr.h"
#include "Mesh.h"
#include "param.h"
#include "paramList.h"
#include "Plot.h"
#include "Solver.h"

using namespace std ;
using namespace keyFE2 ;



int main(void) {
  bool permut=true;
  
  // Reading and renumbering the mesh:
  
  Mesh MESH;
  MESH.read("square.msh","rsMesh");
  string sortType="frontal";
  if (sortType=="position")
  {
    Domain Dom=MESH.getDomain(0);
    Field coord=MESH.getCoordinateField(Dom);
    coord.setName("coordinates");
    Array ACoord((TYPE="arrayField",ORDER=1,DIM=2,NAME="Coordinates field",
                  DERIV=false,COMPONENT(0)=coord.component(0),
                  COMPONENT(1)=coord.component(1)));
    Array ABrol((TYPE="array",ORDER=1,DIM=2,NAME="orient",COMPONENT(0)=1.0,
                 COMPONENT(1)=1.0e-4));
    Array ADist=accumulation(ACoord,ABrol,1);
    MESH=renumberDefDomainArray(MESH,ADist);
  }
  else if (sortType=="frontal") {
    Domain Bnd1=MESH.getDomain(1);
    MESH=renumberDefDomainFrontal(MESH,-Bnd1);
  }
  else if (sortType=="none") {
    // Nothing to do!
  }
  else {
    cerr << "Error in main :" << endl 
      << "   unknown sorting type :  " << sortType << "." << endl ;
    exit(0);
  }
  MESH.write("square.msh2","rsMesh");
  
  // Various domains:
  
  Domain Dom,Bnd;
  Dom=MESH.getDomain(0);
  Dom.setName("Dom");
  Bnd=Dom.getBoundary();
  Dom.setName("Bnd");

  // Coordinate systems and integration:
  
  Field coord=MESH.getCoordinateField(Dom);
  coord.setName("coordinates");
  
  Array ACoord((TYPE="arrayField",ORDER=1,DIM=2,NAME="Coordinates field",
                DERIV=false,COMPONENT(0)=coord.component(0),
                COMPONENT(1)=coord.component(1)));
  
  TensorCoordSys TCS((TCSTYPE="Cartesian2D",COORD=ACoord));
  
  GlobLoc GL2D((COORD=ACoord,GLTYPE="2D2D"));
  GlobLoc GL1D((COORD=ACoord,GLTYPE="2D1D"));
  
  Integ INTG((GLOBLOC=GL2D,TENSORCOORDSYS=TCS,RULE=3));
  
  GlobLoc::setDefaultGL(GL2D);
  
  // Everything for the Poisson equations:

  Field temp1((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
               COMPONENTNAME(0)="T1",NAME="temp1"));
  Field temp2((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
               COMPONENTNAME(0)="T2",NAME="temp2"));
  Array ATemp1((TYPE="arrayField",ORDER=0,NAME="ATemp1",
                TENSORCOORDSYS=TCS,COMPONENT()=temp1.component(0)));
  Array ATemp2((TYPE="arrayField",ORDER=0,NAME="ATemp2",
                TENSORCOORDSYS=TCS,COMPONENT()=temp2.component(0)));
  
  Array ATest1((TYPE="arrayTest",TEST=ATemp1,NAME="ATest1"));
  Array ATest2((TYPE="arrayTest",TEST=ATemp2,NAME="ATest2"));
  
  Array source1((TYPE="array",ORDER=0,NAME="source1",TENSORCOORDSYS=TCS,
                 COMPONENT()=1.0));
  Array source2((TYPE="array",ORDER=0,NAME="source2",TENSORCOORDSYS=TCS,
                 COMPONENT()=2.0));
  
  // Poisson equations:

  Array testA,testB;
  if (permut) {
    testA=ATest2;
    testB=ATest1;
  } else {
    testA=ATest1;
    testB=ATest2;
  }
  
  Array eq1=contractedMultiplication(grad(testA),0,grad(ATemp1),0)
              +testA*source1;
  Array eq2=contractedMultiplication(grad(testB),0,grad(ATemp2),0)
              +testB*source2;
  Array eq=eq1+eq2;
  Array Prob=integrate(eq,Dom,INTG);

  Constr ProbBC1((NAME="ProbBC1",TYPE="usual",DOM=Bnd,
                  TEST=ATest1,EQUATION=ATemp1,GLOBLOC=GL1D,
                  TENSORCOORDSYS=TCS));
  Constr ProbBC2((NAME="ProbBC2",TYPE="usual",DOM=Bnd,
                  TEST=ATest2,EQUATION=ATemp2,GLOBLOC=GL1D,
                  TENSORCOORDSYS=TCS));
  
  // Resolution:
  
  Solver S((PROBLEM=Prob,CONSTR=ProbBC1,CONSTR=ProbBC2,
            CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=6,
	    PERMUT=false));
  S.NewtonRaphson();
  Plot output((DOM=Dom,COORD=ACoord,RESULT=ATemp1,RESULT=ATemp2));
  output.exec("brol");
}


