
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <string>
#include <string.h>
#include <stdio.h>

#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Field.h"
#include "Integ.h"
#include "Constr.h"
#include "Mesh.h"
#include "param.h"
#include "paramList.h"
#include "Plot.h"
#include "Solver.h"

using namespace std ;
using namespace keyFE2 ;



int main(void) {
  char buf[256];
  Int i;
  
  // Reading and renumbering the mesh:
  
  Mesh MESH;
  MESH.read("square.msh","rsMesh");
  string sortType="frontal";
  if (sortType=="position")
  {
    Domain Dom=MESH.getDomain(0);
    Field coord=MESH.getCoordinateField(Dom);
    coord.setName("coordinates");
    Array ACoord((TYPE="arrayField",ORDER=1,DIM=2,NAME="Coordinates field",
                  DERIV=false,COMPONENT(0)=coord.component(0),
                  COMPONENT(1)=coord.component(1)));
    Array ABrol((TYPE="array",ORDER=1,DIM=2,NAME="orient",COMPONENT(0)=1.0,
                 COMPONENT(1)=1.0e-4));
    Array ADist=accumulation(ACoord,ABrol,1);
    MESH=renumberDefDomainArray(MESH,ADist);
  }
  else if (sortType=="frontal") {
    Domain Bnd1=MESH.getDomain(1);
    MESH=renumberDefDomainFrontal(MESH,-Bnd1);
  }
  else if (sortType=="none") {
    // Nothing to do!
  }
  else {
    cerr << "Error in main :" << endl 
      << "   unknown sorting type :  " << sortType << "." << endl ;
    exit(0);
  }
  MESH.write("square.msh2","rsMesh");
  
  // Various domains:
  
  Domain Dom,Bnd,Bnd1,Bnd2,Bnd3,Bnd4,Corners1,Corners2,Corner;
  Dom=MESH.getDomain(0);
  Dom.setName("main domain");
  Bnd1=MESH.getDomain(1);
  Bnd1.setName("left boundary");
  Bnd2=MESH.getDomain(2);
  Bnd2.setName("lower boundary");
  Bnd3=MESH.getDomain(3);
  Bnd3.setName("right boundary");
  Bnd4=MESH.getDomain(4);
  Bnd4.setName("upper boundary");
  Corners1=Bnd1.getBoundary();
  Corners2=Bnd4.getBoundary();
  Corner=Corners1*Corners2;
  Bnd=Bnd1+Bnd2+Bnd3+Bnd4;

  // Coordinate systems and integration:
  
  Field coord=MESH.getCoordinateField(Dom);
  coord.setName("coordinates");
  
  Array ACoord((TYPE="arrayField",ORDER=1,DIM=3,NAME="Coordinates field",
                DERIV=false,COMPONENT(0)=coord.component(0),
                COMPONENT(1)=coord.component(1)));
  Array ACoord2((TYPE="arrayField",ORDER=1,DIM=2,NAME="Coordinates field",
                 DERIV=false,COMPONENT(0)=coord.component(0),
                 COMPONENT(1)=coord.component(1)));
  
  TensorCoordSys TCS((TCSTYPE="CylindricalRZT",COORD=ACoord));
  
  GlobLoc GL((COORD=ACoord,GLTYPE="3D2D"));
  GlobLoc GL2((COORD=ACoord,GLTYPE="3D1D"));
  GlobLoc GL0((COORD=ACoord,GLTYPE="0D"));
  
  Integ INTG((GLOBLOC=GL,TENSORCOORDSYS=TCS,RULE=3));
  
  GlobLoc::setDefaultGL(GL);
  
  // Everything for the Navier-Stokes equations:

  Field pressure((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
                  COMPONENTNAME(0)="p",NAME="pressure"));
  Field velocities((DOM=Dom,INTERPOLATION="quadratic",NBRCOMPONENTS=3,
                    COMPONENTNAME(0)="u",COMPONENTNAME(1)="v",
                    COMPONENTNAME(2)="w",NAME="velocities"));
  Array Avelo((TYPE="arrayField",ORDER=1,DIM=3,NAME="velocities",
               TENSORCOORDSYS=TCS,COMPONENT(0)=velocities.component(0),
               COMPONENT(1)=velocities.component(1),
               COMPONENT(2)=velocities.component(2)));
  Array Apressure((TYPE="arrayField",ORDER=0,NAME="pressure",
                   TENSORCOORDSYS=TCS,COMPONENT()=pressure.component(0)));
  
  Array testMvt((TYPE="arrayTest",TEST=Avelo,NAME="Test function for momentum equations"));  
  Array testDiv((TYPE="arrayTest",TEST=Apressure,
                 NAME="Test function for incompressibility equations"));
  
  Array Unit((TYPE="array",ORDER=2,DIM=3,COMPONENT(0,0)=1.0,COMPONENT(1,1)=1.0,
              COMPONENT(2,2)=1.0,NAME="Unit tensor",TENSORCOORDSYS=TCS));
  Array rho((TYPE="array",ORDER=0,NAME="density",TENSORCOORDSYS=TCS));
  
  // Navier-Stokes equations:

  Array gradVelo=grad(Avelo);
  gradVelo.setName("Gradient of velocities");
  Array Convect=rho*contractedMultiplication(Avelo,0,gradVelo,1);
  Convect.setName("Convective terms");
  Array gradMvt=grad(testMvt);
  gradMvt.setName("gradMvt");
  
  Array divV=contract(gradVelo);
  divV.setName("Divergence of velocities");
  
  Array D=0.5*(gradVelo+transpose(gradVelo));
  D.setName("Strain rate tensor");
  Array T=D-Apressure*Unit;
  T.setName("Cauchy stress tensor");
  Array NavierStokes=accumulation(gradMvt,T)+accumulation(testMvt,Convect)
    +testDiv*divV;
  NavierStokes.setName("Navier-Stokes equations");
  
  Array Prob1=integrate(NavierStokes,Dom,INTG);
  
  // Boundary conditions on velocities and pressure:
  
  Array AveloOne((TYPE="array",ORDER=1,DIM=3,NAME="velocities_left",
                  TENSORCOORDSYS=TCS,COMPONENT(0)=0.0,COMPONENT(1)=0.0,
                  COMPONENT(2)=1.0));
  Constr Prob1CF1((NAME="V=1 on boundary 1",TYPE="usual",DOM=Bnd1,
                   TEST=testMvt,EQUATION=Avelo-AveloOne,GLOBLOC=GL2,
                   TENSORCOORDSYS=TCS,LAYER=-100));
  
  Array AveloZero((TYPE="array",ORDER=1,DIM=3,NAME="Zero velocity",
                   TENSORCOORDSYS=TCS));
  Constr Prob1CF2((NAME="V=0 on boundary 2",TYPE="usual",DOM=Bnd2,
                   TEST=testMvt,EQUATION=Avelo-AveloZero,GLOBLOC=GL2,
                   TENSORCOORDSYS=TCS));
  Constr Prob1CF3((NAME="V=0 on boundary 3",TYPE="usual",DOM=Bnd3,
                   TEST=testMvt,EQUATION=Avelo-AveloZero,GLOBLOC=GL2,
                   TENSORCOORDSYS=TCS));
  Constr Prob1CF4((NAME="V=0 on boundary 4",TYPE="usual",DOM=Bnd4,
                   TEST=testMvt,EQUATION=Avelo-AveloZero,GLOBLOC=GL2,
                   TENSORCOORDSYS=TCS));

  Array ApressureZero((TYPE="array",ORDER=0,NAME="pressureZero",
                       TENSORCOORDSYS=TCS));
  Constr Prob1CF5((NAME="P=0 on a corner",TYPE="usual",DOM=Corner,
                   TEST=testDiv,EQUATION=Apressure-ApressureZero,GLOBLOC=GL0,
                   TENSORCOORDSYS=TCS));

  // Stream function calculation:

  Field stream((DOM=Dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
                COMPONENTNAME(0)="psi",NAME="stream"));
  Array Astream((TYPE="arrayField",ORDER=0,NAME="stream",TENSORCOORDSYS=TCS,
                 COMPONENT()=stream.component(0)));
  Array Avelo2((TYPE="arrayField",ORDER=1,DIM=3,NAME="vitesses",DERIV=false,
                TENSORCOORDSYS=TCS,COMPONENT(0)=velocities.component(0),
                COMPONENT(1)=velocities.component(1)));
     // (Avelo2 is the same as Avelo, but without derivation wrt nodal values).
  Array U((TYPE="arrayComp",ORDER=0,NAME="u",COMPONENT()=Avelo2.component(0)));
  Array V((TYPE="arrayComp",ORDER=0,NAME="v",COMPONENT()=Avelo2.component(1)));
  Array mV=-1.0*V;
  Array tmpVelo((TYPE="arrayComp",ORDER=1,DIM=3,NAME="special",
                 TENSORCOORDSYS=TCS,COMPONENT(0)=mV.component(),
                 COMPONENT(1)=U.component()));
  Array gradPsi=grad(Astream);
  Array testStream((TYPE="arrayTest",TEST=Astream,
                     NAME="Strem function equations"));  
  Array gradStream=grad(testStream);  
  Array R((TYPE="arrayComp",ORDER=0,NAME="R",
	   COMPONENT()=ACoord.component(0),TENSORCOORDSYS=TCS));
  Array deuxPiR=(2.0*3.141592653589793)*R;
  deuxPiR.setName("deuxPiR");
  Array eqStream=gradPsi-deuxPiR*tmpVelo;  
  Array equationStream=accumulation(gradStream,eqStream);
  Array Prob2=integrate(equationStream,Dom,INTG);

  // Stream function boundary conditions:
  
  Array APsiZero((TYPE="array",ORDER=0,NAME="psi = 0",TENSORCOORDSYS=TCS));
  Constr Prob2CF1((NAME="Psi=0 on boundary",TYPE="usual",DOM=Bnd,
                   TEST=testStream,EQUATION=Astream,GLOBLOC=GL2,
                   TENSORCOORDSYS=TCS));
  
  // Resolution:
  
  for (i=0;i<=10;i++) {
    Double Rho=50.0*i;
    rho.component()=Rho;
    
    cerr << "New continuation step with rho = " << Rho << endl ;
    
    Solver NS((PROBLEM=Prob1,CONSTR=Prob1CF1,CONSTR=Prob1CF2,CONSTR=Prob1CF3,
               CONSTR=Prob1CF4,CONSTR=Prob1CF5,CONVERGENCE=1.0e-4,
               DIVERGENCE=1.0e4,NITMAX=6));
    NS.NewtonRaphson();
    
    Solver STREAM((PROBLEM=Prob2,CONSTR=Prob2CF1,CONVERGENCE=1.0e-4,
                   DIVERGENCE=1.0e4,NITMAX=1));
    STREAM.NewtonRaphson();
    Plot output((DOM=Dom,COORD=ACoord2,RESULT=Avelo,RESULT=Apressure,
                 RESULT=Astream,TYPE="openDX"));
    Plot output2((DOM=Dom,COORD=ACoord2,RESULT=Avelo,RESULT=Apressure,
                 RESULT=Astream));
    
    sprintf(buf,"%d",i);
    string filename="square";
    filename+=buf;
    output.exec(filename);
    output2.exec(filename);
  }
}


