
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <string>
#include <stdio.h>

#include "NavierStokes2DConvect.h"
#include "util/util.h"

#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Field.h"
#include "Integ.h"
#include "param.h"
#include "paramList.h"
#include "err.h"

using namespace std ;
using namespace keyFE2 ;
using namespace keyFlow ;



NavierStokes2DConvect::NavierStokes2DConvect(paramList PL) {
  try{
    
    string problemTypeName,globLocTypeName,fieldTypeNames,arrayTypeNames;
    int locDim=-1,globDim=-1,eqDim=-1;
    bool coordFieldNeeded=false,veloFieldNeeded=false,
	 pressureFieldNeeded=false,tempFieldNeeded=false;
    bool gravityArrayExpected=false,gravityArrayFound=false;
    
    // Default parameters:
    
    coordFieldNeeded=true;
    
    // Problem type:
    
    problemTypeName=PL.extractStringData("type",true);
    separateProblemType(problemTypeName,globLocTypeName,fieldTypeNames,
			arrayTypeNames);
    
    // GlobLoc type name:
    
    if (globLocTypeName=="2D2D"||globLocTypeName=="3D2D") {
      globDim=GlobLoc::getGlobDimFromName(globLocTypeName);
      locDim=GlobLoc::getLocDimFromName(globLocTypeName);
      if (locDim!=2)
	errThrow("NavierStokes2DConvect::NavierStokes2DConvect",
		 "invalide GlobLoc type name: "+globLocTypeName+".");	
    }
    else
      errThrow("NavierStokes2DConvect::NavierStokes2DConvect",
	       "invalide GlobLoc type name: "+globLocTypeName+".");
    
    // Time Dependence:
    
    bool boolTransient=false;
    try {
      if (PL.getNbrParamsNamed("timeDep")>0) {
	localTimeDep=PL.extractTimeDepData("timeDep");
        boolTransient=true;
      }
    }
    catchAll("NavierStokes2DConvect::NavierStokes2DConvect",
	     "at the end of TimeDep recovery.");
    
    // Field type names:
    
    if (fieldTypeNames=="V P T") {
      veloFieldNeeded=true;
      pressureFieldNeeded=true;
      tempFieldNeeded=true;
    }
    else
      errThrow("NavierStokes2DConvect::NavierStokes2DConvect",
	       "unknown problem type: "+problemTypeName+".");
    
    // Array type names:
    
    if (arrayTypeNames=="F") {
      gravityArrayExpected=true;
    }
    else
      errThrow("NavierStokes2DConvect::NavierStokes2DConvect",
	       "unknown problem type: "+problemTypeName+".");
    
    // Boolean parameters:
    
    bool boolCoordMobil=false;
    bool boolAxisym=false;
    try {    
      if (PL.getNbrParamsNamed("coordMobil")>0)
	boolCoordMobil=PL.extractBoolData("coordMobil");
      if (PL.getNbrParamsNamed("axisym")>0)
	boolAxisym=PL.extractBoolData("axisym");
    }
    catchAll("NavierStokes2DConvect::NavierStokes2DConvect",
	     "at the end bools recovery.");
    
    // String parameters:
    
    string integRuleName,objectName;
    try {    
      integRuleName=PL.extractStringData("integRule",true);
      objectName=PL.extractStringData("name");
    }
    catchAll("NavierStokes2DConvect::NavierStokes2DConvect",
	     "at the end of strings recovery.");
    
    // Material parameters:
    
    Double dRho=0.0;
    bool boolFRho=false;
    function1 *fRho=NULL;
    string sRhoDep="";
    
    Double dViscosity=0.0;
    
    Double dRhoCp=0.0,dConductivity=0.0;
    
    try {    
      param P;
      
      P=PL.extractParamNamed("rho",true);
      if (P.getTypeName()=="function1") {
	boolFRho=true;
	fRho=(P.getFunction1Data());
	sRhoDep=PL.extractStringData("rhoDep");
      }
      else if (P.getTypeName()=="Double")
	dRho=P.getDoubleData();
      else
	errThrow("NavierStokes2DConvect::NavierStokes2DConvect","invalid type name for rho: "
		 +P.getTypeName()+".");
      
      P=PL.extractParamNamed("viscosity",true);
      if (P.getTypeName()=="Double")
	dViscosity=P.getDoubleData();
      else
	errThrow("NavierStokes2DConvect::NavierStokes2DConvect","invalid type name for viscosity: "
		 +P.getTypeName()+".");
      
      if (tempFieldNeeded) {
	P=PL.extractParamNamed("rhoCp",true);
	if (P.getTypeName()=="Double")
	  dRhoCp=P.getDoubleData();
	else
	  errThrow("NavierStokes2DConvect::NavierStokes2DConvect","invalid type name for rhoCp: " 
	  +P.getTypeName()+".");
	
	P=PL.extractParamNamed("Conduct",true);
	if (P.getTypeName()=="Double")
	  dConductivity=P.getDoubleData();
	else
	  errThrow("NavierStokes2DConvect::NavierStokes2DConvect","invalid type name for Conduct: "
		   +P.getTypeName()+".");
      } 
    }
    catchAll("NavierStokes2DConvect::NavierStokes2DConvect",
	     "at the end of material parameters recovery.");
    
    // Domain:
    
    Domain dom;
    try {    
      dom=PL.extractDomainData("domain",true);
      if (dom.getDim()!=locDim)
        errThrow("NavierStokes2DConvect::NavierStokes2DConvect",
	         "invalid domain dimension.");
    }
    catchAll("NavierStokes2DConvect::NavierStokes2DConvect",
	     "at the end of domain recovery.");
    
    // Recovering of all possible Field's:
    
    Field coordField,veloField,pressureField,tempField;
    try {      
      if (coordFieldNeeded)
	coordField=PL.extractFieldData("coord",true);
      if (veloFieldNeeded)
	veloField=PL.extractFieldData("velo",true);
      if (pressureFieldNeeded)
	pressureField=PL.extractFieldData("pressure",true);
      if (tempFieldNeeded)
	tempField=PL.extractFieldData("temperature",true);
    }
    catchAll("NavierStokes2DConvect::NavierStokes2DConvect",
	     "at the end of Fields' recovery.");
    
    // Array's parameters:
    
    Array gravity;
    try {    
      if (gravityArrayExpected) {
	gravity=PL.extractArrayData("gravity");
	if (gravity.getData()!=NULL) {
	  gravityArrayFound=true;
	  if (gravity.getOrder()!=1)
	    errThrow("NavierStokes2DConvect::NavierStokes2DConvect","gravity order must be 1");
	  if (gravity.getDim(0)!=globDim)
	    errThrow("NavierStokes2DConvect::NavierStokes2DConvect",
		     "number of components of gravity must match globDim.");
	}
	else
	  errThrow("NavierStokes2DConvect::NavierStokes2DConvect",
		   "gravity has to be provided.");
      }
    }
    catchAll("NavierStokes2DConvect::NavierStokes2DConvect",
	     "at the end of Arrays' recovery.");

    // Illegal parameters:
    
    if (PL.getNbrParams()!=0) {
      cerr << "Illegal parameters were found:" << endl ;
      PL.print();
      errThrow("NavierStokes2DConvect::NavierStokes2DConvect","");
    }
        
    // Coordinates and associated entities:
    
    Int veloFieldNbrC=veloField.getNbrComponents();
    if (veloFieldNbrC<2||veloFieldNbrC>3)
      errThrow("NavierStokes2DConvect::NavierStokes2DConvect",
	       "veloField has an invalid number of components.");
    
    if (boolAxisym||veloFieldNbrC==3)
      eqDim=3;
    else
      eqDim=2;
    
    if (eqDim!=globDim)
      errThrow("NavierStokes2DConvect::NavierStokes2DConvect",
	       "invalid equation dimensions for "+globLocTypeName);
    
    try {
      if (coordField.getNbrComponents()!=2)
	errThrow("NavierStokes2DConvect::NavierStokes2DConvect",
		 "coordinate Field must have 2 components.");
      
      coordinates=Array((TYPE="arrayField",ORDER=1,DIM=eqDim,
			 NAME="Coordinates",DERIV=boolCoordMobil,
			 COMPONENT(0)=coordField.component(0),
			 COMPONENT(1)=coordField.component(1)));
      
      if (boolTransient&&boolCoordMobil) {
	coordinatesTimeDeriv=timeDeriv(coordinates,localTimeDep);
	coordinatesTimeDeriv.setName("time derivative of Coordinates");
      }
      
      if (eqDim==3)
	gl=GlobLoc((COORD=coordinates,GLTYPE="3D2D"));
      else
	gl=GlobLoc((COORD=coordinates,GLTYPE="2D2D"));
      
      if (boolAxisym)
	tcs=TensorCoordSys((TCSTYPE="CylindricalRZT",COORD=coordinates));
      else if (eqDim==2)
	tcs=TensorCoordSys((TCSTYPE="Cartesian2D",COORD=coordinates));
      else
	tcs=TensorCoordSys((TCSTYPE="Cartesian3D",COORD=coordinates));
      
      GlobLoc::setDefaultGL(gl);
      itg=Integ((GLOBLOC=gl,TENSORCOORDSYS=tcs,RULENAME=integRuleName));
    }
    catchAll("NavierStokes2DConvect::NavierStokes2DConvect",
	     "at the end of coordinates and associated entities.");
    
    // Quantities related to velocities:
    
    Array testMvt;
    try {
      if (veloFieldNeeded) {
	if (eqDim==2)
	  velocities=Array((TYPE="arrayField",ORDER=1,DIM=2,NAME="Velocities",
			    DERIV=true,TENSORCOORDSYS=tcs,
			    COMPONENT(0)=veloField.component(0),
			    COMPONENT(1)=veloField.component(1)));
	else {
	  if (veloField.getNbrComponents()==3)
	    velocities=Array((TYPE="arrayField",ORDER=1,DIM=3,
			      NAME="Velocities",
			      DERIV=true,TENSORCOORDSYS=tcs,
			      COMPONENT(0)=veloField.component(0),
			      COMPONENT(1)=veloField.component(1),
			      COMPONENT(2)=veloField.component(2)));
	  else if (veloField.getNbrComponents()==2)
	    velocities=Array((TYPE="arrayField",ORDER=1,DIM=3,
			      NAME="Velocities",
			      DERIV=true,TENSORCOORDSYS=tcs,
			      COMPONENT(0)=veloField.component(0),
			      COMPONENT(1)=veloField.component(1)));
	}
	testMvt=Array((TYPE="arrayTest",TEST=velocities,
		       NAME="Momentum equations"));
      }
    }
    catchAll("NavierStokes2DConvect::NavierStokes2DConvect",
	     "at the end of quantities related to velocities.");
    
    // Quantities related to pressure:
    
    Array testDiv;
    try {
      if (pressureFieldNeeded) {
	pressure=Array((TYPE="arrayField",ORDER=0,NAME="Pressure",
			DERIV=true,TENSORCOORDSYS=tcs,
			COMPONENT()=pressureField.component(0)));
	testDiv=Array((TYPE="arrayTest",TEST=pressure,
		       NAME="Incompressibility equation"));
      }
    }
    catchAll("NavierStokes2DConvect::NavierStokes2DConvect",
	     "at the end of quantities related to pressure.");
    
    // Quantities related to temperature:
    
    Array testEnergy;
    
    try {
      if (tempFieldNeeded) {
	if (tempField.getData()==NULL) {
	  errThrow("NavierStokes2DConvect::NavierStokes2DConvect",
		   "temperature Field is needed for energy equation.");
	}
	else
	  temperature=Array((TYPE="arrayField",ORDER=0,NAME="Temperature",
			     DERIV=true,TENSORCOORDSYS=tcs,
			     COMPONENT()=tempField.component(0)));
	testEnergy=Array((TYPE="arrayTest",TEST=temperature,
			  NAME="Energy equation"));
      }
    }
    catchAll("NavierStokes2DConvect::NavierStokes2DConvect",
	     "at the end of quantities related to temperature.");
    
    // Momentum equations:
    
    Array momentumExpr;
    
    try {
      if (boolFRho) {
	rho=Array(temperature,*fRho);
	rho.setName("density");
      }
      else 
	rho=Array((TYPE="array",ORDER=0,COMPONENT()=dRho,
		   TENSORCOORDSYS=tcs,NAME="density"));
      
      gradVeloTensor=grad(velocities);
      gradVeloTensor.setName("gradient of velocities");
      strainRateTensor=0.5*(gradVeloTensor+transpose(gradVeloTensor));
      strainRateTensor.setName("strain rate tensor");
      strainRate=sqrt(2.0*accumulation(strainRateTensor,strainRateTensor));
      strainRate.setName("strain rate");
      
      viscosity=Array((TYPE="array",ORDER=0,COMPONENT()=dViscosity,
		       TENSORCOORDSYS=tcs,NAME="viscosity"));
      
      Array unit;
      if (eqDim==2)
	unit=Array((TYPE="array",ORDER=2,DIM=2,
		    COMPONENT(0,0)=1.0,COMPONENT(1,1)=1.0,
		    NAME="unit tensor",TENSORCOORDSYS=tcs));
      else if (eqDim==3)
	unit=Array((TYPE="array",ORDER=2,DIM=3,
		    COMPONENT(0,0)=1.0,COMPONENT(1,1)=1.0,COMPONENT(2,2)=1.0,
		    NAME="unit tensor",TENSORCOORDSYS=tcs));
      
      CauchyStressTensor=2.0*viscosity*strainRateTensor-pressure*unit;	
      CauchyStressTensor.setName("Cauchy stress tensor");
      
      if (boolTransient) {
	Array velocitiesTimeDeriv=timeDeriv(velocities,localTimeDep);
	velocitiesTimeDeriv.setName("time derivative of Velocities");
	if (boolCoordMobil)
	  inertia=rho*(velocitiesTimeDeriv
		       +contractedMultiplication((velocities-coordinatesTimeDeriv),0,gradVeloTensor,1));
	else
	  inertia=rho*(velocitiesTimeDeriv
		       +contractedMultiplication(velocities,0,gradVeloTensor,1));
      }
      else
	inertia=rho*contractedMultiplication(velocities,0,gradVeloTensor,1);
      inertia.setName("inertia tensor");
      
      Array testMomentum((TYPE="arrayTest",TEST=velocities,
			  NAME="test for momentum equations"));  
      Array gradTestMomentum=grad(testMomentum);
      gradTestMomentum.setName("gradient of test for momentum equations.");
      
      if (gravityArrayFound) {
	force=rho*gravity;
	momentumExpr=accumulation(gradTestMomentum,CauchyStressTensor)
	  +accumulation(testMomentum,inertia-force);
      }
      else
	momentumExpr=accumulation(gradTestMomentum,CauchyStressTensor)
	  +accumulation(testMomentum,inertia);
      momentumExpr.setName("momentum expression");
    }
    catchAll("NavierStokes2DConvect::NavierStokes2DConvect",
	     "at the end of momentum equations.");
    
    // Incompressibility equation:
    
    Array incompressibilityExpr;
    try {
      Array divV=contract(gradVeloTensor);
      divV.setName("divergence of velocities");
      Array testIncompressibility((TYPE="arrayTest",TEST=pressure,
				   NAME="test for incompressibility equation"));
      incompressibilityExpr=testIncompressibility*divV;
      incompressibilityExpr.setName("incompressibility expression");
    }
    catchAll("NavierStokes2DConvect::NavierStokes2DConvect",
	     "at the end of incompressibility equation.");
    
    // Energy equation:
    
    Array energyExpr;
    
    try {
      if (tempFieldNeeded) {
	rhoCp=Array((TYPE="array",ORDER=0,COMPONENT()=dRhoCp,
		     TENSORCOORDSYS=tcs,NAME="rhoCp"));
	
	conductivity=Array((TYPE="array",ORDER=0,COMPONENT()=dConductivity,
			    TENSORCOORDSYS=tcs,NAME="Conductivity"));
	
	Array testEnergy((TYPE="arrayTest",TEST=temperature,
			  NAME="test for energy equation"));
	
	Array gradTestEnergy=grad(testEnergy);
	Array gradTemperature=grad(temperature);
	
	energyFlux=velocities*(rhoCp*temperature)
	  -conductivity*gradTemperature;
	
	Array convectiveTerms;	
	if (boolTransient) {
	  Array temperatureTimeDeriv=timeDeriv(temperature,localTimeDep);
	  temperatureTimeDeriv.setName("time derivative of temperature");
	  if (boolCoordMobil)
	    convectiveTerms=rhoCp*(temperatureTimeDeriv
				   +contractedMultiplication((velocities-coordinatesTimeDeriv),0,gradTemperature,0));
	  else
	    convectiveTerms=rhoCp*(temperatureTimeDeriv
				   +contractedMultiplication(velocities,0,gradTemperature,0));
	}
	else
	  convectiveTerms=rhoCp*contractedMultiplication(velocities,0,gradTemperature,0);
	convectiveTerms.setName("energy convective vector");
	
	energyExpr=testEnergy*convectiveTerms
	  +accumulation(gradTestEnergy,conductivity*gradTemperature);
	energyExpr.setName("energy expression");
      }
    }
    catchAll("NavierStokes2DConvect::NavierStokes2DConvect",
	     "at the end of energy equation.");
    
    // Integration of all the equations:
    
    try {
      Array totalExpr=momentumExpr;
      if (incompressibilityExpr.getData()!=NULL)
	totalExpr=totalExpr+incompressibilityExpr;
      if (energyExpr.getData()!=NULL)
	totalExpr=totalExpr+energyExpr;
      totalExpr.setName("addition of integrable expressions");
      
      Array::operator=(integrate(totalExpr,dom,itg));
      
      setName(objectName);
    }
    catchAll("NavierStokes2DConvect::NavierStokes2DConvect","at the end of integration.");
  }
  catchAll("NavierStokes2DConvect::NavierStokes2DConvect","at the very end.");
}
