
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <string>
#include <stdio.h>

#include "Poisson1D.h"
#include "util/util.h"

#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Field.h"
#include "Integ.h"
#include "param.h"
#include "paramList.h"
#include "err.h"

using namespace std ;
using namespace keyFE2 ;
using namespace keyFlow ;



Poisson1D::Poisson1D(paramList PL) {
  try{
    
    string problemTypeName,globLocTypeName,fieldTypeNames,arrayTypeNames;
    int locDim=-1,globDim=-1;
    bool coordFieldNeeded=false,XFieldNeeded=false;
    
    // Default parameters:
    
    coordFieldNeeded=true;
    
    // Problem type:
    
    problemTypeName=PL.extractStringData("type",true);
    separateProblemType(problemTypeName,globLocTypeName,fieldTypeNames,
			arrayTypeNames);
    
    // GlobLoc type name:
    
    if (globLocTypeName!="2D1D"&&globLocTypeName!="3D1D")
      errThrow("Poisson1D::Poisson1D",
	       "invalide GlobLoc type name: "+globLocTypeName+".");
    globDim=GlobLoc::getGlobDimFromName(globLocTypeName);
    locDim=1;
    
    // Field type names:
    
    if (fieldTypeNames=="X")
      XFieldNeeded=true;
    else
      errThrow("Poisson1D::Poisson1D",
	       "unknown problem type: "+problemTypeName+".");
    
    // Array type names:
    
    if (arrayTypeNames!="")
      errThrow("Poisson1D::Poisson1D",
	       "unknown problem type: "+problemTypeName+".");
      
    // Boolean parameters:
    
    bool boolAxisym=false,boolCoordMobil=false;
    try {    
      if (PL.getNbrParamsNamed("axisym")>0)
	boolAxisym=PL.extractBoolData("axisym");
      if (PL.getNbrParamsNamed("coordMobil")>0)
	boolCoordMobil=PL.extractBoolData("coordMobil");
    }
    catchAll("Poisson1D::Poisson1D",
	     "at the end bools recovery.");
    
    if (boolAxisym&&globDim!=3)
      errThrow("Poisson1D::Poisson1D",
	       "globDim must be 3 for axisymmetric problems.");
    
    if ((!boolAxisym)&&(globDim!=2))
      errThrow("Poisson1D::Poisson1D",
	       "globDim must be 2 for non-axisymmetric problems.");
    
    // String parameters:
    
    string integRuleName,objectName;
    try {    
      integRuleName=PL.extractStringData("integRule",true);
      objectName=PL.extractStringData("name");
    }
    catchAll("Poisson1D::Poisson1D",
	     "at the end of strings recovery.");
    
    // Domain:
    
    Domain dom;
    try {    
      dom=PL.extractDomainData("domain",true);
      if (dom.getDim()!=locDim)
        errThrow("Poisson1D::Poisson1D",
	         "invalid domain dimension.");
    }
    catchAll("Poisson1D::Poisson1D",
	     "at the end of domain recovery.");
    
    // Recovering of all possible Field's:
    
    Field coordField,XField;
    try {      
      if (coordFieldNeeded)
	coordField=PL.extractFieldData("coord",true);
      if (XFieldNeeded)
	XField=PL.extractFieldData("X",true);
    }
    catchAll("Poisson1D::Poisson1D",
	     "at the end of Fields' recovery.");
    
    // Illegal parameters:
    
    if (PL.getNbrParams()!=0) {
      cerr << "Illegal parameters were found:" << endl ;
      PL.print();
      errThrow("Poisson1D::Poisson1D","");
    }
    
    // Coordinates and associated entities:
    
    try {
      if (coordField.getNbrComponents()!=2)
	errThrow("Poisson1D::Poisson1D",
		 "coordinate Field must have 2 components.");
      
      coordinates=Array((TYPE="arrayField",ORDER=1,DIM=globDim,
			 NAME="Coordinates",DERIV=boolCoordMobil,
			 COMPONENT(0)=coordField.component(0),
			 COMPONENT(1)=coordField.component(1)));
      if (boolAxisym) {
	tcs=TensorCoordSys((TCSTYPE="CylindricalRZT",COORD=coordinates));
	gl=GlobLoc((COORD=coordinates,GLTYPE="3D1D"));
      }
      else {
	tcs=TensorCoordSys((TCSTYPE="Cartesian2D",COORD=coordinates));
	gl=GlobLoc((COORD=coordinates,GLTYPE="2D1D"));;
      }
      itg=Integ((GLOBLOC=gl,TENSORCOORDSYS=tcs,RULENAME=integRuleName));
      
    }
    catchAll("Poisson1D::Poisson1D",
	     "at the end of coordinates and associated entities.");
    
    // Quantities related to unknown Field:
    
    Int XFieldNbrC=XField.getNbrComponents();
    if (XFieldNbrC!=1)
      errThrow("Poisson1D::Poisson1D",
	       "XField has an invalid number of components.");
    
    try {
      X=Array((TYPE="arrayField",ORDER=0,NAME="X",DERIV=true,TENSORCOORDSYS=tcs,
	       COMPONENT()=XField.component(0)));
      testX=Array((TYPE="arrayTest",TEST=X,NAME="Test for Poisson equations"));
    }
    catchAll("Poisson1D::Poisson1D",
	     "at the end of quantities related to X.");
    
    // Poisson equation:
    
    Array Poisson1DExpr;
    try {
      Poisson1DExpr=accumulation(grad(testX,gl),grad(X,gl));
      Poisson1DExpr.setName("Tangential remeshing expression");
    }
    catchAll("Poisson1D::Poisson1D",
	     "at the end of tengential remeshing equations.");
    
    // Integration of all the equations:
    
    try {
      Array::operator=(integrate(Poisson1DExpr,dom,itg));
      setName(objectName);
    }
    catchAll("Poisson1D::Poisson1D","at the end of integration.");
  }
  catchAll("Poisson1D::Poisson1D","at the very end.");
}




