
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <string>
#include <stdio.h>

#include "kinCond1D.h"
#include "Poisson/Poisson1D.h"
#include "util/util.h"

#include "Array.h"
#include "ArrayOperators.h"
#include "Constr.h"
#include "Domain.h"
#include "Field.h"
#include "GlobLoc.h"
#include "Integ.h"
#include "Problem.h"
#include "Solver.h"
#include "TensorCoordSys.h"
#include "err.h"
#include "param.h"
#include "paramList.h"

using namespace std ;
using namespace keyFE2 ;
using namespace keyFlow ;



kinCond1D::kinCond1D(paramList PL) {
  try{
    
    string problemTypeName,globLocTypeName,fieldTypeNames,arrayTypeNames;
    int locDim=-1,globDim=-1,eqDim=-1;
    bool coordFieldNeeded=false,veloFieldNeeded=false;
    Domain Point1,Point2;
    Array Azero,Aone;
    GlobLoc pointGL;
    
    // Default parameters:
    
    coordFieldNeeded=true;
    
    // Problem type:
    
    problemTypeName=PL.extractStringData("type",true);
    separateProblemType(problemTypeName,globLocTypeName,fieldTypeNames,
			arrayTypeNames);
    
    // GlobLoc type name:
    
    if (globLocTypeName!="2D1D"&&globLocTypeName!="3D1D")
      errThrow("kinCond1D::kinCond1D",
	       "invalide GlobLoc type name: "+globLocTypeName+".");
    globDim=GlobLoc::getGlobDimFromName(globLocTypeName);
    locDim=1;
    eqDim=globDim;
    
    // Time Dependence:
    
    bool boolTransient=false;
    try {
      if (PL.getNbrParamsNamed("timeDep")>0) {
	localTimeDep=PL.extractTimeDepData("timeDep");
        boolTransient=true;
      }
    }
    catchAll("kinCond1D::kinCond1D",
	     "at the end of TimeDep recovery.");
    
    // Field type names:
    
    if (fieldTypeNames=="Coord V") {
      coordFieldNeeded=true;
      veloFieldNeeded=true;
    }
    else
      errThrow("kinCond1D::kinCond1D",
	       "unknown problem type: "+problemTypeName+".");
    
    // Array type names:
    
    if (arrayTypeNames!="")
      errThrow("kinCond1D::kinCond1D",
	       "unknown problem type: "+problemTypeName+".");
    
    // Boolean parameters:
    
    bool boolAxisym=false,boolFixBegin=false,boolFixEnd=false,
         boolCoupledNormals=true,boolUpwindKinem=false;
    try {    
      if (PL.getNbrParamsNamed("axisym")>0)
	boolAxisym=PL.extractBoolData("axisym",true);
      if (PL.getNbrParamsNamed("fixBegin")>0)
	boolFixBegin=PL.extractBoolData("fixBegin",true);
      if (PL.getNbrParamsNamed("fixEnd")>0)
	boolFixEnd=PL.extractBoolData("fixEnd",true);
      if (PL.getNbrParamsNamed("coupledNormals")>0)
	boolCoupledNormals=PL.extractBoolData("coupledNormals",true);
      if (PL.getNbrParamsNamed("upwindKinem")>0)
	boolUpwindKinem=PL.extractBoolData("upwindKinem");
    }
    catchAll("kinCond1D::kinCond1D",
	     "at the end bools recovery.");
    
    if (boolAxisym&&globDim!=3)
      errThrow("kinCond1D::kinCond1D",
	       "globDim must be 3 for axisymmetric problems.");
    
    if ((!boolAxisym)&&(globDim!=2))
      errThrow("kinCond1D::kinCond1D",
	       "globDim must be 2 for non-axisymmetric problems.");
    
    // String parameters:
    
    string integRuleName,objectName;
    try {    
      integRuleName=PL.extractStringData("integRule",true);
      objectName=PL.extractStringData("name");
    }
    catchAll("kinCond1D::kinCond1D",
	     "at the end of strings recovery.");
    
    // Material parameters:
    
    Double dSurfaceTension=0.0;
    try {    
      param P=PL.extractParamNamed("surfaceTension",true);
      if (P.getTypeName()=="Double")
	dSurfaceTension=P.getDoubleData();
      else
	errThrow("kinCond1D::kinCond1D","invalid type name for surface tension: "
		 +P.getTypeName()+".");
    }
    catchAll("kinCond1D::kinCond1D",
	     "at the end of material parameters recovery.");
    
    // Domain:
    
    try {    
      dom=PL.extractDomainData("domain",true);
      if (dom.getDim()!=locDim)
        errThrow("kinCond1D::kinCond1D",
	         "invalid domain dimension.");
    }
    catchAll("kinCond1D::kinCond1D",
	     "at the end of domain recovery.");
    
    // Recovering of all possible Field's:
    
    Field coordField,veloField;
    try {      
      if (coordFieldNeeded)
	coordField=PL.extractFieldData("coord",true);
      if (veloFieldNeeded)
	veloField=PL.extractFieldData("velo",true);
    }
    catchAll("kinCond1D::kinCond1D",
	     "at the end of Fields' recovery.");
    
    // Illegal parameters:
    
    if (PL.getNbrParams()!=0) {
      cerr << "Illegal parameters were found:" << endl ;
      PL.print();
      errThrow("kinCond1D::kinCond1D","");
    }
    
    // Coordinates and associated entities:
    
    Array testCartCoordinates,testCoordinates,CartCoordNoDeriv,
      testCartCoordNoDeriv;
    TensorCoordSys testCartTCS,testTCS,CartTCS_ND;
    GlobLoc testCartGL,testGL,CartGL_ND,testCartGL_ND;
    Integ CartITG_ND;
    
    try {
      if (coordField.getNbrComponents()!=2)
	errThrow("kinCond1D::kinCond1D",
		 "coordinate Field must have 2 components.");
      
      CartCoordinates=Array((TYPE="arrayField",ORDER=1,DIM=2,
			     NAME="CartCoordinates",DERIV=true,
			     COMPONENT(0)=coordField.component(0),
			     COMPONENT(1)=coordField.component(1)));
      CartTCS=TensorCoordSys((TCSTYPE="Cartesian2D",COORD=CartCoordinates));
      CartGL=GlobLoc((COORD=CartCoordinates,GLTYPE="2D1D"));
      pointGL=GlobLoc((COORD=CartCoordinates,GLTYPE="0D"));
      CartITG=Integ((GLOBLOC=CartGL,TENSORCOORDSYS=CartTCS,RULENAME=integRuleName));
      
      CartCoordNoDeriv=Array((TYPE="arrayField",ORDER=1,DIM=2,
			     NAME="CartCoordNoDeriv",DERIV=false,
			     COMPONENT(0)=coordField.component(0),
			     COMPONENT(1)=coordField.component(1)));
      CartTCS_ND=TensorCoordSys((TCSTYPE="Cartesian2D",COORD=CartCoordNoDeriv));
      CartGL_ND=GlobLoc((COORD=CartCoordNoDeriv,GLTYPE="2D1D"));
      CartITG_ND=Integ((GLOBLOC=CartGL_ND,TENSORCOORDSYS=CartTCS_ND,RULENAME=integRuleName));

      testCartCoordinates=Array((TYPE="arrayTest",TEST=CartCoordinates,
                                 NAME="testCartCoordinates"));  
      testCartCoordNoDeriv=Array((TYPE="arrayTest",TEST=CartCoordNoDeriv,
                                  NAME="testCartCoordNoDeriv"));  
      testCartTCS=TensorCoordSys((TCSTYPE="Cartesian2D",
                                  COORD=testCartCoordinates));
      testCartGL=GlobLoc((COORD=testCartCoordinates,GLTYPE="2D1D"));
      testCartGL_ND=GlobLoc((COORD=testCartCoordNoDeriv,GLTYPE="2D1D"));
      
      if (boolAxisym) {
        coordinates=Array((TYPE="arrayField",ORDER=1,DIM=eqDim,
			   NAME="Coordinates",DERIV=true,
			   COMPONENT(0)=coordField.component(0),
			   COMPONENT(1)=coordField.component(1)));
	tcs=TensorCoordSys((TCSTYPE="CylindricalRZT",COORD=coordinates));
	gl=GlobLoc((COORD=coordinates,GLTYPE="3D1D"));
        itg=Integ((GLOBLOC=gl,TENSORCOORDSYS=tcs,RULENAME=integRuleName));
	
        testCoordinates=Array((TYPE="arrayTest",TEST=coordinates,
                               NAME="testCoordinates"));
	testTCS=TensorCoordSys((TCSTYPE="CylindricalRZT",COORD=testCoordinates));
	testGL=GlobLoc((COORD=testCoordinates,GLTYPE="3D1D"));
      }
      else {
	coordinates=CartCoordinates;
	tcs=CartTCS;
	gl=CartGL;
	itg=CartITG;
 
 	testCoordinates=testCartCoordinates;
	testTCS=CartTCS;
	testGL=CartGL;
      }
      
      if (boolTransient) {
	coordinatesTimeDeriv=timeDeriv(coordinates,localTimeDep);
	coordinatesTimeDeriv.setName("time derivative of Coordinates");
      }
      
      X=Array((TYPE="arrayField",ORDER=0,NAME="X",DERIV=true,
	       COMPONENT()=coordField.component(0),TENSORCOORDSYS=tcs));
      Y=Array((TYPE="arrayField",ORDER=0,NAME="Y",DERIV=true,
	       COMPONENT()=coordField.component(1),TENSORCOORDSYS=tcs));
    }
    catchAll("kinCond1D::kinCond1D",
	     "at the end of coordinates and associated entities.");
    
    // additional Field construction:
    
    Array ProbNormals;
    
    try {
      cerr << endl ;
      cerr << " ********************************************************** " << endl ; 
      cerr << " *                                                        * " << endl ; 
      cerr << " *  Beginning of reference coordinate Field construction  * " << endl ; 
      cerr << " *                                                        * " << endl ; 
      cerr << " ********************************************************** " << endl ; 
      cerr << endl ;
      
      // Two extremities of the Domain:
      
      dom.getExtremities(Point1,Point2);
      
      // Building of needed entities:
      
      refTCoordField=Field((DOM=dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
	                    COMPONENTNAME(0)="T",NAME="refTCoordField"));
      
      Poisson1D Prob((param("type","2D1D - X - "),
		      param("domain",dom),
		      param("coord",coordField),
		      param("X",refTCoordField),
		      param("axisym",false),
		      param("integRule","order 5"),
		      param("name","Prob for refTCoordField") ));
      
      Azero=Array((TYPE="array",ORDER=0,COMPONENT()=0.0,NAME="Azero",
		   TENSORCOORDSYS=CartTCS_ND));
      Aone=Array((TYPE="array",ORDER=0,COMPONENT()=1.0,NAME="Aone",
		  TENSORCOORDSYS=CartTCS_ND));
      Constr ProbBC1((NAME="refTCoord on boundary Point 1",TYPE="usual",DOM=Point1,
		      TEST=Prob.testX,EQUATION=Prob.X-Azero,GLOBLOC=pointGL,
		      TENSORCOORDSYS=CartTCS_ND));
      Constr ProbBC2((NAME="refTCoord on boundary Point 2",TYPE="usual",DOM=Point2,
		      TEST=Prob.testX,EQUATION=Prob.X-Aone,GLOBLOC=pointGL,
		      TENSORCOORDSYS=CartTCS_ND));
      
      // Resolution:
      
      Solver refTCoordSolver((PROBLEM=Prob,CONSTR=ProbBC1,CONSTR=ProbBC2,
		              CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=6));
      refTCoordSolver.NewtonRaphson();
      
      refTCoord=Array((TYPE="arrayField",ORDER=0,DERIV=false,NAME="refTCoord",
		       COMPONENT()=refTCoordField.component(0),
		       TENSORCOORDSYS=CartTCS));
      
      cerr << endl ;
      cerr << " **************************************************** " << endl ; 
      cerr << " *                                                  * " << endl ; 
      cerr << " *  End of reference coordinate Field construction  * " << endl ; 
      cerr << " *                                                  * " << endl ; 
      cerr << " **************************************************** " << endl ; 
      cerr << endl ;

      cerr << endl ;
      cerr << " ******************************************** " << endl ; 
      cerr << " *                                          * " << endl ; 
      cerr << " *  Beginning of normal Field construction  * " << endl ; 
      cerr << " *                                          * " << endl ; 
      cerr << " ******************************************** " << endl ; 
      cerr << endl ;

      Array normalTmp,testNormalTmp;

      normalField=Field((DOM=dom,
                         INTERPOLATION=coordField.getInterpolationTypeName(),
			 NBRCOMPONENTS=2,
	                 COMPONENTNAME(0)="Nx",
	                 COMPONENTNAME(1)="Ny",
			 NAME="normalField"));
      normal=Array((TYPE="arrayField",ORDER=1,DIM(0)=2,DERIV=false,
                    NAME="normal",
		    COMPONENT(0)=normalField.component(0),
		    COMPONENT(1)=normalField.component(1),
		    TENSORCOORDSYS=CartTCS));
      normalTmp=Array((TYPE="arrayField",ORDER=1,DIM(0)=2,DERIV=true,
                       NAME="normalTmp",
		       COMPONENT(0)=normalField.component(0),
		       COMPONENT(1)=normalField.component(1),
		       TENSORCOORDSYS=CartTCS));
      testNormalTmp=Array((TYPE="arrayTest",TEST=normalTmp,
                           NAME="testNormalTmp"));
			   
      Array equation2=
        accumulation(testNormalTmp,normalTmp-CartGL.getNormal(CartTCS));
      ProbNormals=integrate(equation2,dom,CartITG);
      ProbNormals.setName(objectName+" (Coupled calculation of normals)");
      
      Array equation2_ND=
        accumulation(testNormalTmp,normalTmp-Prob.gl.getNormal(Prob.tcs));
      Array Prob2_ND=integrate(equation2_ND,dom,Prob.itg);
      
      normalSolver=Solver((PROBLEM=Prob2_ND,CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,
                           NITMAX=6));
	
      cerr << endl ;
      cerr << " ************************************** " << endl ; 
      cerr << " *                                    * " << endl ; 
      cerr << " *  End of normal Field construction  * " << endl ; 
      cerr << " *                                    * " << endl ; 
      cerr << " ************************************** " << endl ; 
      cerr << endl ;
    }
    catchAll("kinCond1D::kinCond1D",
	     "at the end of reference coordinates and normal.");
	     
    // Quantities related to normal and tangential displacements:
    
    Array displNT_ND;
    
    displField=Field((DOM=dom,
                      INTERPOLATION=coordField.getInterpolationTypeName(),
	              NBRCOMPONENTS=2,MINREF=1.0,
	              COMPONENTNAME(0)="N",COMPONENTNAME(1)="T",
	              NAME="displField"));
		      
    displNT=Array((TYPE="arrayField",ORDER=1,DIM(0)=2,
                   DERIV=true,
                   NAME="displNT",
		   COMPONENT(0)=displField.component(0),
		   COMPONENT(1)=displField.component(1),
		   TENSORCOORDSYS=CartTCS));
    displNT_ND=Array((TYPE="arrayField",ORDER=1,DIM(0)=2,
                      DERIV=false,
                      NAME="displNT_ND",
		      COMPONENT(0)=displField.component(0),
		      COMPONENT(1)=displField.component(1),
		      TENSORCOORDSYS=CartTCS));
    displN=Array((TYPE="arrayField",ORDER=0,
                  DERIV=true,
                  NAME="displN",
		  COMPONENT()=displField.component(0),
		  TENSORCOORDSYS=CartTCS));
    displT=Array((TYPE="arrayField",ORDER=0,
                  DERIV=true,
                  NAME="displT",
		  COMPONENT()=displField.component(1),
		  TENSORCOORDSYS=CartTCS));
		  
    testDisplNT=Array((TYPE="arrayTest",TEST=displNT,NAME="testDisplNT"));
    testDisplN=Array((TYPE="arrayTest",TEST=displN,NAME="testDisplN"));
    testDisplT=Array((TYPE="arrayTest",TEST=displT,NAME="testDisplT"));
    
    Nx=Array((TYPE="arrayComp",ORDER=0,NAME="Nx",
    	      COMPONENT()=normal.component(0)));
    Ny=Array((TYPE="arrayComp",ORDER=0,NAME="Ny",
    	      COMPONENT()=normal.component(1)));
    Ty=Nx;
    Tx=-1.0*Ny;
    Nvect=Array((TYPE="arrayComp",ORDER=1,DIM(0)=2,
                 NAME="Nvect",
                 COMPONENT(0)=Nx.component(),
                 COMPONENT(1)=Ny.component()));
    Tvect=Array((TYPE="arrayComp",ORDER=1,DIM(0)=2,
                 NAME="Tvect",
                 COMPONENT(0)=Tx.component(),
                 COMPONENT(1)=Ty.component()));
    transfoNT=Array((TYPE="arrayComp",ORDER=2,DIM(0)=2,DIM(1)=2,
    	             NAME="transfoNT",
	             COMPONENT(0,0)=Nx.component(),
	             COMPONENT(0,1)=Ny.component(),
	             COMPONENT(1,0)=Tx.component(),
	             COMPONENT(1,1)=Ty.component()));
	       
    Array displXY=contractedMultiplication(transfoNT,0,displNT,0);
    Array displXY_ND=contractedMultiplication(transfoNT,0,displNT_ND,0);
    
    BCDom=Constr((NAME="BCDom",TYPE="usual",DOM=dom,LAYER=-10,
		  TEST=testCartCoordinates,
		  EQUATION=(CartCoordinates-CartCoordNoDeriv)-displXY,
		  GLOBLOC=CartGL,
		  TENSORCOORDSYS=CartTCS));
    BCPoint1=Constr((NAME="BCPoint1",TYPE="usual",DOM=Point1,
		     TEST=testDisplT,EQUATION=displT,
		     GLOBLOC=pointGL,TENSORCOORDSYS=CartTCS));
    BCPoint2=Constr((NAME="BCPoint2",TYPE="usual",DOM=Point2,
		     TEST=testDisplT,EQUATION=displT,
		     GLOBLOC=pointGL,TENSORCOORDSYS=CartTCS));
    addConstr(BCDom);
    addConstr(BCPoint1);
    addConstr(BCPoint2);
    
    if (boolFixBegin) {
      Constr BCPoint=Constr((NAME="fix N displacement on begin",TYPE="usual",
                             DOM=Point1,TEST=testDisplN,EQUATION=displN,
		             GLOBLOC=pointGL,TENSORCOORDSYS=CartTCS));
      addConstr(BCPoint);
    }
    
    if (boolFixEnd) {
      Constr BCPoint=Constr((NAME="fix N displacement on end",TYPE="usual",
                             DOM=Point2,TEST=testDisplN,EQUATION=displN,
		             GLOBLOC=pointGL,TENSORCOORDSYS=CartTCS));
      addConstr(BCPoint);
    }
    
    // Quantities related to velocities:
    
    Int veloFieldNbrC=veloField.getNbrComponents();
    if (veloFieldNbrC<2||veloFieldNbrC>3)
      errThrow("kinCond1D::kinCond1D",
	       "veloField has an invalid number of components.");
    
    Array testMvt;
    try {
      if (veloFieldNeeded) {
	if (eqDim==2)
	  velocities=Array((TYPE="arrayField",ORDER=1,DIM=2,NAME="Velocities",
			    DERIV=true,TENSORCOORDSYS=tcs,
			    COMPONENT(0)=veloField.component(0),
			    COMPONENT(1)=veloField.component(1)));
	else {
	  if (veloField.getNbrComponents()==3)
	    velocities=Array((TYPE="arrayField",ORDER=1,DIM=3,NAME="Velocities",
			      DERIV=true,TENSORCOORDSYS=tcs,
			      COMPONENT(0)=veloField.component(0),
			      COMPONENT(1)=veloField.component(1),
			      COMPONENT(2)=veloField.component(2)));
	  else if (veloField.getNbrComponents()==2)
	    velocities=Array((TYPE="arrayField",ORDER=1,DIM=3,NAME="Velocities",
			      DERIV=true,TENSORCOORDSYS=tcs,
			      COMPONENT(0)=veloField.component(0),
			      COMPONENT(1)=veloField.component(1)));
	}
	testMvt=Array((TYPE="arrayTest",TEST=velocities,
		       NAME="Momentum equations"));
      }
    }
    catchAll("kinCond1D::kinCond1D",
	     "at the end of quantities related to velocities.");
    
    updateNormalDispl();

    // Tangential remeshing:
    
    Array thompson1DExpr;
    try {
      thompson1DExpr=accumulation(grad(testDisplT,CartGL),grad(refTCoord,CartGL));
      thompson1DExpr.setName("Tangential remeshing expression");
    }
    catchAll("kinCond1D::kinCond1D",
	     "at the end of tengential remeshing equations.");    
    
    // Kinematic condition:
    
    Array kinCond1DExpr;
    try {
      Array testConvect;
      
      if (boolUpwindKinem) {
        double dInterp;
        if (coordField.getInterpolationTypeName()=="none"
            ||coordField.getInterpolationTypeName()=="constantOnDomain"
            ||coordField.getInterpolationTypeName()=="constantPerElement"
            ||coordField.getInterpolationTypeName()=="linear"
            ||coordField.getInterpolationTypeName()=="linearPerElement")
          dInterp=1.0;
        else
          dInterp=2.0;
	  
        if (boolTransient)
  	  testConvect=upwindF(velocities,1.0,0.0,testDisplN,
	                      coordinatesTimeDeriv,gl,dInterp);
  	else
	  testConvect=upwindF(velocities,1.0,0.0,testDisplN,0.0,gl,dInterp);
      }
      else
  	testConvect=testDisplN;
      testConvect.setName("testKinem");

      Array normale=gl.getNormal(tcs);
      if (boolTransient)
	kinCond1DExpr=testConvect*(accumulation(normale,velocities-coordinatesTimeDeriv));
      else
	kinCond1DExpr=testConvect*(accumulation(normale,velocities));
      kinCond1DExpr.setName("Kinematical condition expression");
    }
    catchAll("kinCond1D::kinCond1D",
	     "at the end of kinematical condition equations.");
    
    // Surface tension:
    
    Array capillarityExpr;
    try {
      Array coordinates2((TYPE="arrayField",ORDER=1,DIM=globDim,
			  NAME="coordinates with tcs",DERIV=true,
			  COMPONENT(0)=coordField.component(0),
		          COMPONENT(1)=coordField.component(1),
			  TENSORCOORDSYS=tcs));
      surfaceTension=Array((TYPE="array",ORDER=0,COMPONENT()=dSurfaceTension,
			    TENSORCOORDSYS=tcs,NAME="surface tension"));      
      capillarityExpr=surfaceTension*accumulation(grad(testMvt,gl),grad(coordinates2,gl));
      capillarityExpr=surfaceTension*accumulation(grad(testMvt,gl),grad(coordinates2,gl));
      capillarityExpr.setName("Capillarity expression");
    }
    catchAll("kinCond1D::kinCond1D",
	     "at the end of momentum equations.");
    
    // False coordinates expression:
    
    Array falseCoordExpr;
    try {
      Array Azero((TYPE="array",ORDER=0,COMPONENT()=1.0e-20,
                   TENSORCOORDSYS=tcs,NAME="Azero"));
      falseCoordExpr=Azero*accumulation(CartCoordinates,testCartCoordinates);
    }
    catchAll("kinCond1D::kinCond1D",
	     "at the end of momentum equations.");
    
    // Integration of all the equations:
    
    try {
      if (boolAxisym) {
	Array axisymExpr=kinCond1DExpr+capillarityExpr;
	axisymExpr.setName("addition of axisymmetric expressions");
	Array tmp1=integrate(axisymExpr,dom,itg);
	tmp1.setName(objectName+" (the Array, part 1)");
        addArray(tmp1);
	
	Array tmp2=integrate(thompson1DExpr+falseCoordExpr,dom,CartITG);
	tmp2.setName(objectName+" (the Array, part 2)");
        addArray(tmp2);
      }
      else {
	Array totalExpr=thompson1DExpr+kinCond1DExpr+capillarityExpr+falseCoordExpr;
	totalExpr.setName("addition of all expressions");
	
	Array tmp=integrate(totalExpr,dom,itg);
	tmp.setName(objectName+" (the Array)");
        addArray(tmp);
      }
      if (boolCoupledNormals)
        addArray(ProbNormals);
      
    }
    catchAll("kinCond1D::kinCond1D","at the end of integration.");
    
  }
  catchAll("kinCond1D::kinCond1D","at the very end.");
}




void kinCond1D::updateNormalDispl(void) {
  
  // Calculation of normalField:
  
  normalSolver.NewtonRaphson();
  {
     Array tmp1=sqrt(accumulation(normal,normal));
     Array tmp2=pow(tmp1,-1.0)*normal;
     normalField.interpolate(tmp2,dom,coordinates);
  }
  
}

