
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <string>
#include <stdio.h>

#include "kinCondXY.h"
#include "Poisson/Poisson1D.h"
#include "util/util.h"

#include "Array.h"
#include "ArrayOperators.h"
#include "Constr.h"
#include "Domain.h"
#include "Field.h"
#include "GlobLoc.h"
#include "Integ.h"
#include "Solver.h"
#include "TensorCoordSys.h"
#include "err.h"
#include "param.h"
#include "paramList.h"

using namespace std ;
using namespace keyFE2 ;
using namespace keyFlow ;



kinCondXY::kinCondXY(paramList PL) {
  try{
    
    string problemTypeName,globLocTypeName,fieldTypeNames,arrayTypeNames;
    int locDim=-1,globDim=-1,eqDim=-1;
    bool coordFieldNeeded=false,veloFieldNeeded=false;
    
    // Default parameters:
    
    coordFieldNeeded=true;
    
    // Problem type:
    
    problemTypeName=PL.extractStringData("type",true);
    separateProblemType(problemTypeName,globLocTypeName,fieldTypeNames,
			arrayTypeNames);
    
    // GlobLoc type name:
    
    if (globLocTypeName!="2D1D"&&globLocTypeName!="3D1D")
      errThrow("kinCondXY::kinCondXY",
	       "invalide GlobLoc type name: "+globLocTypeName+".");
    globDim=GlobLoc::getGlobDimFromName(globLocTypeName);
    locDim=1;
    eqDim=globDim;
    
    // Time Dependence:
    
    bool boolTransient=false;
    try {
      if (PL.getNbrParamsNamed("timeDep")>0) {
	localTimeDep=PL.extractTimeDepData("timeDep");
        boolTransient=true;
      }
    }
    catchAll("kinCondXY::kinCondXY",
	     "at the end of TimeDep recovery.");
    
    // Field type names:
    
    if (fieldTypeNames=="Coord V") {
      coordFieldNeeded=true;
      veloFieldNeeded=true;
    }
    else
      errThrow("kinCondXY::kinCondXY",
	       "unknown problem type: "+problemTypeName+".");
    
    // Array type names:
    
    if (arrayTypeNames!="")
      errThrow("kinCondXY::kinCondXY",
	       "unknown problem type: "+problemTypeName+".");
    
    // Boolean parameters:
    
    bool boolAxisym=false;
    bool boolUpwindKinem=false;
    try {    
      if (PL.getNbrParamsNamed("axisym")>0)
	boolAxisym=PL.extractBoolData("axisym");
      if (PL.getNbrParamsNamed("upwindKinem")>0)
	boolUpwindKinem=PL.extractBoolData("upwindKinem");
    }
    catchAll("kinCondXY::kinCondXY",
	     "at the end bools recovery.");
    
    if (boolAxisym&&globDim!=3)
      errThrow("kinCondXY::kinCondXY",
	       "globDim must be 3 for axisymmetric problems.");
    
    if ((!boolAxisym)&&(globDim!=2))
      errThrow("kinCondXY::kinCondXY",
	       "globDim must be 2 for non-axisymmetric problems.");
    
    // String parameters:
    
    string integRuleName,objectName;
    try {    
      integRuleName=PL.extractStringData("integRule",true);
      objectName=PL.extractStringData("name");
    }
    catchAll("kinCondXY::kinCondXY",
	     "at the end of strings recovery.");
    
    // Material parameters:
    
    Double dSurfaceTension=0.0;
    try {    
      param P=PL.extractParamNamed("surfaceTension",true);
      if (P.getTypeName()=="Double")
	dSurfaceTension=P.getDoubleData();
      else
	errThrow("kinCondXY::kinCondXY","invalid type name for surface tension: "
		 +P.getTypeName()+".");
    }
    catchAll("kinCondXY::kinCondXY",
	     "at the end of material parameters recovery.");
    
    // Domain:
    
    Domain dom;
    try {    
      dom=PL.extractDomainData("domain",true);
      if (dom.getDim()!=locDim)
        errThrow("kinCondXY::kinCondXY",
	         "invalid domain dimension.");
    }
    catchAll("kinCondXY::kinCondXY",
	     "at the end of domain recovery.");
    
    // Recovering of all possible Field's:
    
    Field coordField,veloField;
    try {      
      if (coordFieldNeeded)
	coordField=PL.extractFieldData("coord",true);
      if (veloFieldNeeded)
	veloField=PL.extractFieldData("velo",true);
    }
    catchAll("kinCondXY::kinCondXY",
	     "at the end of Fields' recovery.");
    
    // Illegal parameters:
    
    if (PL.getNbrParams()!=0) {
      cerr << "Illegal parameters were found:" << endl ;
      PL.print();
      errThrow("kinCondXY::kinCondXY","");
    }
    
    // Coordinates and associated entities:
    
    try {
      if (coordField.getNbrComponents()!=2)
	errThrow("kinCondXY::kinCondXY",
		 "coordinate Field must have 2 components.");
      
      CartCoordinates=Array((TYPE="arrayField",ORDER=1,DIM=2,
			     NAME="Cartesian coordinates",DERIV=true,
			     COMPONENT(0)=coordField.component(0),
			     COMPONENT(1)=coordField.component(1)));
      CartTCS=TensorCoordSys((TCSTYPE="Cartesian2D",COORD=CartCoordinates));
      CartGL=GlobLoc((COORD=CartCoordinates,GLTYPE="2D1D"));
      CartITG=Integ((GLOBLOC=CartGL,TENSORCOORDSYS=CartTCS,RULENAME=integRuleName));
      
      if (boolAxisym) {
        coordinates=Array((TYPE="arrayField",ORDER=1,DIM=eqDim,
			   NAME="Coordinates",DERIV=true,
			   COMPONENT(0)=coordField.component(0),
			   COMPONENT(1)=coordField.component(1)));
	tcs=TensorCoordSys((TCSTYPE="CylindricalRZT",COORD=coordinates));
	gl=GlobLoc((COORD=coordinates,GLTYPE="3D1D"));
        itg=Integ((GLOBLOC=gl,TENSORCOORDSYS=tcs,RULENAME=integRuleName));
      }
      else {
	coordinates=CartCoordinates;
	tcs=CartTCS;
	gl=CartGL;
	itg=CartITG;
      }
      
      if (boolTransient) {
	coordinatesTimeDeriv=timeDeriv(coordinates,localTimeDep);
	coordinatesTimeDeriv.setName("time derivative of Coordinates");
      }
      
      X=Array((TYPE="arrayField",ORDER=0,NAME="X",DERIV=true,
	       COMPONENT()=coordField.component(0),TENSORCOORDSYS=tcs));
      Y=Array((TYPE="arrayField",ORDER=0,NAME="Y",DERIV=true,
	       COMPONENT()=coordField.component(1),TENSORCOORDSYS=tcs));
      testX=Array((TYPE="arrayTest",TEST=X,NAME="testX"));  
      testY=Array((TYPE="arrayTest",TEST=Y,NAME="testY"));  
      
      if (boolAxisym) {
	CartX=Array((TYPE="arrayField",ORDER=0,NAME="CartX",DERIV=true,
		     COMPONENT()=coordField.component(0),TENSORCOORDSYS=CartTCS));
	CartY=Array((TYPE="arrayField",ORDER=0,NAME="CartY",DERIV=true,
		     COMPONENT()=coordField.component(1),TENSORCOORDSYS=CartTCS));
	testCartX=Array((TYPE="arrayTest",TEST=CartX,NAME="testCartX"));  
	testCartY=Array((TYPE="arrayTest",TEST=CartY,NAME="testCartY"));  
      }
      else {
	CartX=X;
	CartY=Y;
	testCartX=testX;
	testCartY=testY;
      }
    }
    catchAll("kinCondXY::kinCondXY",
	     "at the end of coordinates and associated entities.");
    
    // Reference coordinate Field construction:
    
    try {
      cerr << endl ;
      cerr << " ********************************************************** " << endl ; 
      cerr << " *                                                        * " << endl ; 
      cerr << " *  Beginning of reference coordinate Field construction  * " << endl ; 
      cerr << " *                                                        * " << endl ; 
      cerr << " ********************************************************** " << endl ; 
      cerr << endl ;
      
      // Two extremities of the Domain:
      
      Domain Point1,Point2;
      dom.getExtremities(Point1,Point2);
      
      GlobLoc pointGL((COORD=CartCoordinates,GLTYPE="0D"));;
      
      // Building of needed entities:
      
      refTCoordField=Field((DOM=dom,INTERPOLATION="linear",NBRCOMPONENTS=1,
	                    COMPONENTNAME(0)="T",NAME="refTCoordField"));
      
      Poisson1D Prob((param("type","2D1D - X - "),
		      param("domain",dom),
		      param("coord",coordField),
		      param("X",refTCoordField),
		      param("axisym",false),
		      param("integRule","order 5"),
		      param("name","Prob for refTCoordField") ));
      
      Array Azero((TYPE="array",ORDER=0,COMPONENT()=0.0,NAME="Azero",
		   TENSORCOORDSYS=CartTCS));
      Array Aone((TYPE="array",ORDER=0,COMPONENT()=1.0,NAME="Aone",
		  TENSORCOORDSYS=CartTCS));
      Constr ProbBC1((NAME="refTCoord on boundary Point 1",TYPE="usual",DOM=Point1,
		      TEST=Prob.testX,EQUATION=Prob.X-Azero,GLOBLOC=pointGL,
		      TENSORCOORDSYS=CartTCS));
      Constr ProbBC2((NAME="refTCoord on boundary Point 2",TYPE="usual",DOM=Point2,
		      TEST=Prob.testX,EQUATION=Prob.X-Aone,GLOBLOC=pointGL,
		      TENSORCOORDSYS=CartTCS));
      
      // Resolution:
      
      Solver S((PROBLEM=Prob,CONSTR=ProbBC1,CONSTR=ProbBC2,
		CONVERGENCE=1.0e-4,DIVERGENCE=1.0e4,NITMAX=6));
      S.NewtonRaphson();
      
      refTCoord=Array((TYPE="arrayField",ORDER=0,DERIV=false,NAME="refTCoord",
		       COMPONENT()=refTCoordField.component(0),TENSORCOORDSYS=CartTCS));
      
      cerr << endl ;
      cerr << " ******************************************************** " << endl ; 
      cerr << " *                                                      * " << endl ; 
      cerr << " *    End of reference coordinate Field construction    * " << endl ; 
      cerr << " *                                                      * " << endl ; 
      cerr << " ******************************************************** " << endl ; 
      cerr << endl ;
    }
    catchAll("kinCondXY::kinCondXY",
	     "at the end of reference coordinates.");    
    
    // Quantities related to velocities:
    
    Int veloFieldNbrC=veloField.getNbrComponents();
    if (veloFieldNbrC<2||veloFieldNbrC>3)
      errThrow("kinCondXY::kinCondXY",
	       "veloField has an invalid number of components.");
    
    Array testMvt;
    try {
      if (veloFieldNeeded) {
	if (eqDim==2)
	  velocities=Array((TYPE="arrayField",ORDER=1,DIM=2,NAME="Velocities",
			    DERIV=true,TENSORCOORDSYS=tcs,
			    COMPONENT(0)=veloField.component(0),
			    COMPONENT(1)=veloField.component(1)));
	else {
	  if (veloField.getNbrComponents()==3)
	    velocities=Array((TYPE="arrayField",ORDER=1,DIM=3,NAME="Velocities",
			      DERIV=true,TENSORCOORDSYS=tcs,
			      COMPONENT(0)=veloField.component(0),
			      COMPONENT(1)=veloField.component(1),
			      COMPONENT(2)=veloField.component(2)));
	  else if (veloField.getNbrComponents()==2)
	    velocities=Array((TYPE="arrayField",ORDER=1,DIM=3,NAME="Velocities",
			      DERIV=true,TENSORCOORDSYS=tcs,
			      COMPONENT(0)=veloField.component(0),
			      COMPONENT(1)=veloField.component(1)));
	}
	testMvt=Array((TYPE="arrayTest",TEST=velocities,
		       NAME="Momentum equations"));
      }
    }
    catchAll("kinCondXY::kinCondXY",
	     "at the end of quantities related to velocities.");
    
    // Tangential remeshing:
    
    Array thompson1DExpr;
    try {
      Array testRefTCoord=Array((TYPE="arrayTest",TEST=refTCoord,
				 NAME="Test for reference tangential coordinates"));
      thompson1DExpr=accumulation(grad(testCartY,CartGL),grad(refTCoord,CartGL));
      thompson1DExpr.setName("Tangential remeshing expression");
    }
    catchAll("kinCondXY::kinCondXY",
	     "at the end of tengential remeshing equations.");    
    
    // Kinematic condition:
    
    Array kinCondExpr;
    try {
      Array testConvect;
      
      if (boolUpwindKinem) {
        double dInterp;
        if (coordField.getInterpolationTypeName()=="none"
            ||coordField.getInterpolationTypeName()=="constantOnDomain"
            ||coordField.getInterpolationTypeName()=="constantPerElement"
            ||coordField.getInterpolationTypeName()=="linear"
            ||coordField.getInterpolationTypeName()=="linearPerElement")
          dInterp=1.0;
        else
          dInterp=2.0;
	  
        if (boolTransient)
  	  testConvect=upwindF(velocities,1.0,0.0,testX,
	                      coordinatesTimeDeriv,gl,dInterp);
  	else
	  testConvect=upwindF(velocities,1.0,0.0,testX,0.0,gl,dInterp);
      }
      else
  	testConvect=testX;
      testConvect.setName("testKinem");
      
      Array normale=gl.getNormal(tcs);
      if (boolTransient)
        kinCondExpr=testConvect*accumulation(normale,velocities-coordinatesTimeDeriv);
      else
        kinCondExpr=testConvect*accumulation(normale,velocities);
      kinCondExpr.setName("Kinematical condition expression");
    }
    catchAll("kinCondXY::kinCondXY",
	     "at the end of kinematical condition equations.");
    
    // Surface tension:
    
    Array capillarityExpr;
    try {
      Array coordinates2((TYPE="arrayField",ORDER=1,DIM=globDim,
			  NAME="coordinates with tcs",DERIV=true,
			  COMPONENT(0)=coordField.component(0),
		          COMPONENT(1)=coordField.component(1),
			  TENSORCOORDSYS=tcs));
      surfaceTension=Array((TYPE="array",ORDER=0,COMPONENT()=dSurfaceTension,
			    TENSORCOORDSYS=tcs,NAME="surface tension"));      
      capillarityExpr=surfaceTension*accumulation(grad(testMvt,gl),grad(coordinates2,gl));
      capillarityExpr.setName("Capillarity expression");
    }
    catchAll("kinCondXY::kinCondXY",
	     "at the end of momentum equations.");
    
    // Integration of all the equations:
    
    try {
      if (boolAxisym) {
	Array axisymExpr=kinCondExpr+capillarityExpr;
	axisymExpr.setName("addition of axisymmetric expressions");
	Array CartExpr=thompson1DExpr;
	CartExpr.setName("addition of Cartesian expressions");
	
	Array::operator=(integrate(axisymExpr,dom,itg)
			 +integrate(CartExpr,dom,CartITG));
	setName(objectName);
      }
      else {
	Array totalExpr=thompson1DExpr+kinCondExpr+capillarityExpr;
	totalExpr.setName("addition of all expressions");
	
	Array::operator=(integrate(totalExpr,dom,itg));
	setName(objectName);
      }
    }
    catchAll("kinCondXY::kinCondXY","at the end of integration.");
  }
  catchAll("kinCondXY::kinCondXY","at the very end.");
}
