
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <string>
#include <stdio.h>

#include "streamFunction2D.h"
#include "util/util.h"

#include "Array.h"
#include "ArrayOperators.h"
#include "GlobLoc.h"
#include "TensorCoordSys.h"
#include "Domain.h"
#include "Field.h"
#include "Integ.h"
#include "param.h"
#include "paramList.h"
#include "err.h"

using namespace std ;
using namespace keyFE2 ;
using namespace keyFlow ;



streamFunction2D::streamFunction2D(paramList PL) {
  try{
    
    string problemTypeName,globLocTypeName,fieldTypeNames,arrayTypeNames;
    int locDim=-1,globDim=-1;
    
    // Problem type:
    
    problemTypeName=PL.extractStringData("type",true);
    separateProblemType(problemTypeName,globLocTypeName,fieldTypeNames,
			arrayTypeNames);
    
    // GlobLoc type name:
    
    if (globLocTypeName=="2D2D"||globLocTypeName=="3D2D") {
      globDim=GlobLoc::getGlobDimFromName(globLocTypeName);
      locDim=GlobLoc::getLocDimFromName(globLocTypeName);
      if (locDim!=2)
	errThrow("streamFunction2D::streamFunction2D",
		 "invalid GlobLoc type name: "+globLocTypeName+".");	
    }
    else
      errThrow("streamFunction2D::streamFunction2D",
	       "invalid GlobLoc type name: "+globLocTypeName+".");
    
    // Field type names:
    
    if (fieldTypeNames!="psi")
      errThrow("streamFunction2D::streamFunction2D",
	       "unknown problem type: "+problemTypeName+".");
    
    // Array type names:
    
    if (arrayTypeNames!="V")
      errThrow("streamFunction2D::streamFunction2D",
	       "velocities Array is expected.");
    
    // Boolean parameters:
    
    bool boolAxisym=false;
    try {    
      if (PL.getNbrParamsNamed("axisym")>0)
	boolAxisym=PL.extractBoolData("axisym");
    }
    catchAll("streamFunction2D::streamFunction2D",
	     "at the end bools recovery.");
    
    // String parameters:
    
    string integRuleName,objectName;
    try {    
      integRuleName=PL.extractStringData("integRule",true);
      objectName=PL.extractStringData("name");
    }
    catchAll("streamFunction2D::streamFunction2D",
	     "at the end of strings recovery.");
        
    // Domain:
    
    Domain dom;
    try {  
      dom=PL.extractDomainData("domain",true);
      if (dom.getDim()!=locDim)
        errThrow("streamFunction2D::streamFunction2D","invalid domain dimension.");
    }
    catchAll("streamFunction2D::streamFunction2D","at the end of domain recovery.");
    
    // Recovering of all possible Field's:
    
    Field coordField,psiField,veloField;
    try {      
      coordField=PL.extractFieldData("coord",true);
      if (coordField.getNbrComponents()!=2)
        errThrow("streamFunction2D::streamFunction2D",
		 "invalid number of components for coordinates field.");

      psiField=PL.extractFieldData("psi",true);
      if (psiField.getNbrComponents()!=1)
        errThrow("streamFunction2D::streamFunction2D",
		 "invalid number of components for coordinates field.");
      
      veloField=PL.extractFieldData("velo",true);
      if (veloField.getNbrComponents()<2||veloField.getNbrComponents()>3)
        errThrow("streamFunction2D::streamFunction2D",
		 "invalid number of components for velocity field.");
    }
    catchAll("streamFunction2D::streamFunction2D",
	     "at the end of Fields' recovery.");
    
    // Illegal parameters:
    
    if (PL.getNbrParams()!=0) {
      cerr << "Illegal parameters were found:" << endl ;
      PL.print();
      errThrow("streamFunction2D::streamFunction2D","");
    }
    
    
    // Coordinates and associated entities:

    if (boolAxisym&&(globDim!=3))
      errThrow("streamFunction2D::streamFunction2D",
	       "global dimension must be 3 for axisymmetric problems.");
    
    try {
      coordinates=Array((TYPE="arrayField",ORDER=1,DIM=globDim,DERIV=false,
			 NAME="Coordinates",
			 COMPONENT(0)=coordField.component(0),
			 COMPONENT(1)=coordField.component(1)));
      
      gl=GlobLoc((COORD=coordinates,GLTYPE=globLocTypeName));
      
      if (boolAxisym)
	tcs=TensorCoordSys((TCSTYPE="CylindricalRZT",COORD=coordinates));
      else if (globDim==2)
	tcs=TensorCoordSys((TCSTYPE="Cartesian2D",COORD=coordinates));
      else
	tcs=TensorCoordSys((TCSTYPE="Cartesian3D",COORD=coordinates));
      
      GlobLoc::setDefaultGL(gl);
      itg=Integ((GLOBLOC=gl,TENSORCOORDSYS=tcs,RULENAME=integRuleName));
    }
    catchAll("streamFunction2D::streamFunction2D",
	     "at the end of coordinates and associated entities.");
    
    // Velocities :
    
    try {
      if (veloField.getNbrComponents()==2)
        velocities=Array((TYPE="arrayField",ORDER=1,DIM=globDim,DERIV=false,
			  NAME="Velocities",
			  TENSORCOORDSYS=tcs,
			  COMPONENT(0)=veloField.component(0),
			  COMPONENT(1)=veloField.component(1)));
      else
        velocities=Array((TYPE="arrayField",ORDER=1,DIM=globDim,DERIV=false,
  			  NAME="Velocities",
			  TENSORCOORDSYS=tcs,
			  COMPONENT(0)=veloField.component(0),
			  COMPONENT(1)=veloField.component(1),
			  COMPONENT(2)=veloField.component(2)));
    }
    catchAll("streamFunction2D::streamFunction2D",
	     "at the end of velocities and associated entities.");    
    
    // Stream function equations:
    
    try {
      psi=Array((TYPE="arrayField",ORDER=0,DERIV=true,NAME="stream function",
		 COMPONENT()=psiField.component(0),TENSORCOORDSYS=tcs));
      
      Array trspArray((TYPE="array",ORDER=2,DIM(0)=globDim,DIM(1)=velocities.getDim(0),
		       COMPONENT(0,1)=-1.0,COMPONENT(1,0)=1.0,
		       NAME="transposer of velocities"));
      Array trspVelo=contractedMultiplication(trspArray,1,velocities,0);
      trspVelo.setTCS(tcs);
      trspVelo.setName("trspVelo");
      
      Array gradPsi=grad(psi);
      gradPsi.setName("gradPsi");
      Array testPsi((TYPE="arrayTest",TEST=psi,NAME="testPsi"));  
      Array gradTestPsi=grad(testPsi);
      gradTestPsi.setName("gradTestPsi");
      
      Array eqStream;
      if (boolAxisym) {
	Array R((TYPE="arrayComp",ORDER=0,NAME="R",
		 COMPONENT()=coordinates.component(0),TENSORCOORDSYS=tcs));
	Array deuxPiR=(2.0*3.141592653589793)*R;
	deuxPiR.setName("deuxPiR");
	eqStream=accumulation(gradTestPsi,gradPsi-deuxPiR*trspVelo);
      }
      else
        eqStream=accumulation(gradTestPsi,gradPsi-trspVelo);
      
      Array::operator=(integrate(eqStream,dom,itg));
      setName(objectName);
    }
    catchAll("streamFunction2D::streamFunction2D",
	     "at the end of stream function equations.");
  }
  catchAll("streamFunction2D::streamFunction2D","at the very end.");
}
