
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <stdio.h>
#include <iostream>
#include <iomanip>
#include <typeinfo>

#include "array_base.h"
#include "arrayField.h"
#include "arrayList.h"
#include "arrayOpN.h"
#include "coordSys/globLoc_base.h"
#include "coordSys/tensorCoordSys_base.h"
#include "err/err.h"
#include "field/field_base.h"
#include "print/printTree.h"
#include "scal/scal.h"
#include "step/step.h"
#include "step/stepList.h"
#include "util/utils.h"

using namespace std ;
using namespace keyFE2 ;



array::array(void) : name(), newPoint(), parents(), boolZero() {
  try {
    layer=0;
  
    order=0;
    dim=NULL;
    nbrVal=1;
    
    componentValues=new scal[nbrVal];
    
    currentComponent=-1;
    
    TCS=NULL;
  }
  catchAll("array::array(void)","")
}



array::array(const array &in) : name(), newPoint(), parents(), boolZero(in) {
  Int i;

  try {
    layer=in.layer;
    
    order=in.order;
    if (order>MAX_ORDER_ARRAY)
      errThrow("array constructor",string("   maximum order of an array is ")
	       +strFromInt(MAX_ORDER_ARRAY)+".");
    
    nbrVal=1;
    if (order>0) {
      dim=new Int[order];
      for (i=0;i<order;i++) {
	dim[i]=in.dim[i];
	nbrVal*=dim[i];
      }
    }
    else
      dim=NULL;
    
    componentValues=new scal[nbrVal];
    for (i=0;i<nbrVal;i++)
      componentValues[i]=in.componentValues[i];
    
    currentComponent=-1;
    
    TCS=NULL;
    setTCS(in.getTCS());
  }
  catchAll("array::array(const array &in)","")
}



array::array(double d) : name(), newPoint(), parents(), boolZero() {
  try {
    layer=0;
  
    order=0;
    dim=NULL;
    nbrVal=1;
    
    componentValues=new scal[nbrVal];
    componentValues[0]=d;
    
    currentComponent=-1;
    
    TCS=NULL;
  }
  catchAll("array::array(double d)","")
}



void array::copyArrayData(const array &in) {
  Int i;
  
  try {
    layer=in.layer;
    
    order=in.order;
    if (order>MAX_ORDER_ARRAY)
      errThrow("array::copyArrayData",string("   maximum order of an array is ")
	       +strFromInt(MAX_ORDER_ARRAY)+".");
    
    delete[] dim;
    nbrVal=1;
    if (order>0) {
      dim=new Int[order];
      for (i=0;i<order;i++) {
	dim[i]=in.dim[i];
	nbrVal*=dim[i];
      }
    }
    else
      dim=NULL;
    
    delete[] componentValues;
    componentValues=new scal[nbrVal];
    for (i=0;i<nbrVal;i++)
      componentValues[i]=in.componentValues[i];
    
    currentComponent=-1;
    
    updateBoolZeroValue();
    
    setTCS(in.getTCS());
  }
  catchAll("void array::copyArrayData(const array &in)","")
}



array::~array(void) {
  try {
    layer=0;
    
    order=-1;
    delete[] dim;
    dim=NULL;
    nbrVal=0;
    delete[] componentValues;
    componentValues=NULL;
    
    currentComponent=-1;
    
    setTCS(NULL);
  }
  catchAll("array::~array(void)","")
}



array* array::makeCloneForArrayTime(timeDep &TD) {
  try {
    array *newArray=new array;
    newArray->copyArrayData(*this);
    return(newArray);
  }
  catchAll("array::makeCloneForArrayTime","")
}



void array::setLayer(Int nbr) {
   if (nbr<MIN_LAYER) {
     errThrow("array::setLayer(Int nbr)",
              "Cannot set layer below "+string(MIN_LAYER_STRING))
   }
  layer=nbr;
}



Int array::getLayer(void) const {
  return layer;
}



void array::setOrder(Int nbr) {
  Int i;
  
  try {
    delete[] dim;
    dim=NULL;
    delete[] componentValues;
    componentValues=NULL;
    
    if (nbr>MAX_ORDER_ARRAY)
      errThrow("array::setOrder",string("   maximum order of an array is ")
	       +strFromInt(MAX_ORDER_ARRAY)+".");
    
    order=nbr;
    if (TCS==NULL) {
      if (order>0)
	dim=new Int[order];
      else
	dim=NULL;
      for (i=0;i<order;i++)
	dim[i]=1;
      nbrVal=1;
      componentValues=new scal[nbrVal];
    }
    else {
      if (order>0)
	dim=new Int[order];
      else
	dim=NULL;
      for (i=0;i<order;i++)
	dim[i]=TCS->getCoord()->getDim(0);
      nbrVal=1;
      for (i=0;i<order;i++)
	nbrVal*=dim[i];
      componentValues=new scal[nbrVal];
    }
    
    currentComponent=-1;
    
    updateBoolZeroValue();
  }
  catchAll("void array::setOrder(Int nbr)","")
}



Int array::getOrder(void) const {
  return(order);
}



void array::setDim(Int index,Int nbrComponents) {
  Int i;
  
  try {
    if (index>=order)
      errThrow("array::setDim","   index larger than array order.");
    
    dim[index]=nbrComponents;
    
    delete[] componentValues;
    
    nbrVal=1;
    for (i=0;i<order;i++)
      nbrVal*=dim[i];
    componentValues=new scal[nbrVal];
    
    currentComponent=-1;
    
    updateBoolZeroValue();
  }
  catchAll("void array::setDim(Int index,Int nbrComponents)","")
}



Int array::getDim(Int index) const {
  try {
    if (index>=order)
      errThrow("array::getDim","   index larger than array order.");
    
    return(dim[index]);
  }
  catchAll("Int array::getDim(Int index) const","")
}



Int array::getNbrVal(void) const {
  return(nbrVal);
}



void array::setTCS(tensorCoordSys *tcs) {
  try {
    if (tcs!=NULL) {
      for (Int i=0;i<order;i++) {
	if (dim[i]!=tcs->getCoord()->getDim(0))
	  errThrow("array::setTCS",
		   "   all dimensions must be equal to the number of coordinates.");
      }
    }
    delinkAndDelete(TCS,&TCS);
    TCS=tcs;
    if (TCS!=NULL)
      TCS->addParent(&TCS);
  }
  catchAll("void array::setTCS(tensorCoordSys *tcs)","")
}



tensorCoordSys* array::getTCS(void) const {
  return(TCS);
}



bool array::updateBoolZeroValue(void) {
  try {
    for (Int i=0;i<nbrVal;i++) {
      if (!componentValues[i].isZero()) {
	setBoolZeroValue(false);
	return(false);
      }
    }
    setBoolZeroValue(true);
    return(true);
  }
  catchAll("bool array::updateBoolZeroValue(void)","")
}



bool array::isDiagonal(void) const {
  static Int i,j,Dim;
  
  try {
    if (order!=2) 
      return(false);
    if (dim[0]!=dim[1])
      return(false);
    
    Dim=dim[0];
    for (i=0;i<Dim;i++)
      for (j=0;j<Dim;j++)
	if (componentValues[Dim*i+j]!=0.0&&i!=j)
	  return(false);
    
    return(true);
  }
  catchAll("bool array::isDiagonal(void) const","")
}



Int array::getIndexFromIndices(const Int indices[MAX_ORDER_ARRAY]) const {
  Int index,i,nbr;
  
  try {
    for (nbr=0;nbr<MAX_ORDER_ARRAY;nbr++)
      if (indices[nbr]==-1)
	break;
    if (nbr!=order)
      errThrow("array::getIndexFromIndices",
	       string("   the order of the array is ")+strFromInt(order)
	       +" and not "+strFromInt(nbr)+".");
    
    index=0;
    Int size=1;
    for (i=order-1;i>=0;i--) {
      if (indices[i]>=dim[i])
	errThrow("array::getIndexFromIndices",
		 string("   the number of components for index ")
		 +strFromInt(i)+" is "+strFromInt(dim[i])+".");
      
      index+=indices[i]*size;
      size*=dim[i];
    }
    
    return(index);
  }
  catchAll("array::getIndexFromIndices","")
}



void array::getIndicesFromIndex(Int index,Int indices[MAX_ORDER_ARRAY]) const {
  static Int i;
  static Int size;
  
  try {
    if (index>nbrVal)
      errThrow("array::getIndicesFromIndex",
	       string("   the number of elements of the array is ")
	       +strFromInt(nbrVal)+".");
    
    indices[order]=-1;
    size=nbrVal;
    for (i=0;i<order;i++) {
      size/=dim[i];
      indices[i]=index/size;
      index-=indices[i]*size;
    }
  }
  catchAll("array::getIndicesFromIndex","")
}



array& array::component(const Int &index0,const Int &index1,
                        const Int &index2,const Int &index3,
                        const Int &index4,const Int &index5,
                        const Int &index6,const Int &index7) {
  static Int argIndex[MAX_ORDER_ARRAY];
  
  try {
    argIndex[0]=index0;
    argIndex[1]=index1;
    argIndex[2]=index2;
    argIndex[3]=index3;
    argIndex[4]=index4;
    argIndex[5]=index5;
    argIndex[6]=index6;
    argIndex[7]=index7;
    
    if (order>0)
      currentComponent=getIndexFromIndices(argIndex);
    else
      currentComponent=0;
    
    return(*this);
  }
  catchAll("array::component","")
}



scal& array::componentScal(const Int &index0,const Int &index1,
                           const Int &index2,const Int &index3,
                           const Int &index4,const Int &index5,
                           const Int &index6,const Int &index7) {
  static Int argIndex[MAX_ORDER_ARRAY];
  static Int comp;
  
  try {
    argIndex[0]=index0;
    argIndex[1]=index1;
    argIndex[2]=index2;
    argIndex[3]=index3;
    argIndex[4]=index4;
    argIndex[5]=index5;
    argIndex[6]=index6;
    argIndex[7]=index7;
    
    if (order>0)
      comp=getIndexFromIndices(argIndex);
    else
      comp=0;
    
    return(componentValues[comp]);
  }
  catchAll("array::componentScal","")
}



array& array::component(const Int index[MAX_ORDER_ARRAY]) {
  try {
    if (order>0)
      currentComponent=getIndexFromIndices(index);
    else
      currentComponent=0;
    
    return(*this);
  }
  catchAll("array& array::component(const Int index[MAX_ORDER_ARRAY])","")
}



scal& array::componentScal(const Int index[MAX_ORDER_ARRAY]) {
  static Int comp;
  
  try {
    if (order>0)
      comp=getIndexFromIndices(index);
    else
      comp=0;
    
    return(componentValues[comp]);
  }
  catchAll("scal& array::componentScal(const Int index[MAX_ORDER_ARRAY])","")
}



void array::setCurrentComponent(Int nbr) {
  currentComponent=nbr;
}



Int array::getCurrentComponent(void) const {
  return(currentComponent);
}



scal* array::getComponentValues(void) const {
  return(componentValues);
}



void array::operator=(Double d) {
  try {
    if (currentComponent==-1)
      errThrow("array = double","   only components may be assigned to.");
    
    componentValues[currentComponent]=d;
    currentComponent=-1;
    if (d!=0.0)
      boolZeroValue=false;
    else
      updateBoolZeroValue();
  }
  catchAll("void array::operator=(Double d)","")
}



void array::operator=(scal &arg) {
  try {
    if (currentComponent==-1)
      errThrow("array = scal","   only components may be assigned to.");
    
    componentValues[currentComponent]=arg;
    currentComponent=-1;
    if (!(arg.isZero()))
      boolZeroValue=false;
    else
      updateBoolZeroValue();
  }
  catchAll("void array::operator=(scal &arg)","")
}



void array::operator=(field&) {
  errThrow("array = field",
	   "   a field only may be assigned to an arrayField.");
}



void array::operator=(array&) {
  errThrow("array = array",
	   "   an array only may be assigned to an arrayComp.");
}



void array::makeAtPoint(object *element,Int direct,Double *localCoord,
			bool makeValue,bool makeDeriv) {
  try {
    updateBoolZeroValue(); // On pourra le supprimer plus tard.
  }
  catchAll("array::makeAtPoint","")
}



void array::makeOnObject(object *element,Int direct,
			 bool makeValue,bool makeDeriv) {
  try {
    updateBoolZeroValue(); // On pourra le supprimer plus tard.
  }
  catchAll("array::makeOnObject","")
}



array* array::makeDeriv(Int nbrScal,scal **scalList,globLoc *GL,
			string derivType) const {
  Int i,j,newIndex;
  
  try {
    if (!(derivType=="explicit"||derivType=="coord"||derivType=="corloc"))
      errThrow("array::makeDeriv",
	       string("   unknown derivation type : ")+derivType+".");
    
    if (derivType=="coord") {
      if (GL!=NULL) {
	nbrScal=GL->getCoordComponentsNbr();
	scalList=GL->getCoordComponents();
      }
      else
	errThrow("array::makeDeriv",
		 "   no coord derivative can be calculated if GL==NULL");
    }
    
    array *newArray=new array;
    newArray->setOrder(order+1);
    for (i=0;i<order;i++)
      newArray->setDim(i,dim[i]);
    newArray->setDim(order,nbrScal);
    
    if (derivType=="explicit"||derivType=="coord") {
      for (i=0;i<nbrVal;i++) {
	for (j=0;j<nbrScal;j++) {
	  newIndex=i*nbrScal+j;
	  if (componentValues+i==scalList[j])
	    newArray->componentValues[newIndex]=1.0;
	  else
	    newArray->componentValues[newIndex]=0.0;
	}
      }
    }
    else if (derivType=="corloc")
      for (i=0;i<newArray->nbrVal;i++)
	newArray->componentValues[i]=0.0;
    
    newArray->updateBoolZeroValue();
    
    return(newArray);
  }
  catchAll("array::makeDeriv","")
}



array* array::makeExplicitDeriv(Int nbrScal,scal **scalList) const {
  try {
    return(makeDeriv(nbrScal,scalList,NULL,"explicit"));
  }
  catchAll("array::makeExplicitDeriv","")
}



array* array::makeCoordDeriv(globLoc *GL) const {
  try {
    return(makeDeriv(0,NULL,GL,"coord"));
  }
  catchAll("array::makeCoordDeriv","")
}



array* array::makeCorlocDeriv(Int nbrDP) const {
  try {
    return(makeDeriv(nbrDP,NULL,NULL,"corloc"));
  }
  catchAll("array::makeCorlocDeriv","")
}



array* array::buildPhysicalDeriv(globLoc *GL) {
  try {
    if (TCS==NULL)
      errThrow("array::buildPhysicalDeriv :","   I am not a tensor.");
    
    arrayOpN *tmpArray=new arrayOpN;
    tmpArray->setOrder(order+1);
    for (Int i=0;i<order;i++)
      tmpArray->setDim(i,dim[i]);
    tmpArray->setDim(order,GL->getCoord()->getDim(0));
    tmpArray->setTCS(TCS);
    tmpArray->setArrayArgsNbr(2);
    tmpArray->setArrayArg(0,this);
    tmpArray->setArrayArg(1,makeCoordDeriv(GL));
    tmpArray->setType("physicalDeriv");
    
    return(tmpArray);
  }
  catchAll("array::buildPhysicalDeriv","")
}



array* array::makePhysicalDeriv(globLoc *GL) {
  try {
    return(buildPhysicalDeriv(GL));
  }
  catchAll("array::makePhysicalDeriv","")
}



printTree array::getPrintTreeData(void) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf;
    printTree pt;
    Int i;
    
    if (getName()=="")
      sprintf(cbuf,"   anonymous");
    else
      sprintf(cbuf,"   name = %s",getName().c_str());
    sbuf=cbuf;
    pt+=sbuf;
    
    if (layer!=0) {
      sprintf(cbuf,"   layer = %d",layer);
      sbuf=cbuf;
      pt+=sbuf;
    }
    
    sprintf(cbuf,"   order = %d",order);
    sbuf=cbuf;
    pt+=sbuf;
    
    if (order>0) {
      sprintf(cbuf,"   dim =%2d",dim[0]);
      sbuf=cbuf;
      for (i=1;i<order;i++) {
	sprintf(cbuf,",%2d",dim[i]);
	sbuf+=cbuf;
      }
      pt+=sbuf;
    }
    
    set<field*>::iterator it;
    for (it=FL.begin();it!=FL.end();it++)
      pt+=string("   FL: ")+(*it)->getName();
    
    return(pt);
  }
  catchAll("array::getPrintTreeData","")
}



printTree array::getPrintTreeValues(void) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf;
    printTree pt;
    Int index,indices[MAX_ORDER_ARRAY];
    Int i;
    Double d;
    
    sbuf="   values of the components:";
    pt+=sbuf;
    
    for (index=0;index<nbrVal;index++) {
      getIndicesFromIndex(index,indices);
      sbuf="      component[";
      for (i=0;i<order;i++) {
	if (i==0) 
	  sprintf(cbuf,"%d",indices[i]);
	else
	  sprintf(cbuf,",%d",indices[i]);
	sbuf+=cbuf;
      }
      sprintf(cbuf,"] = ");
      sbuf+=cbuf; 
      d=componentValues[index].getVal();
      if (componentValues[index].isZero())
	sprintf(cbuf,"%g (is zero)",d);
      else
	sprintf(cbuf,"%g (is not zero)",d);
      sbuf+=cbuf;
      pt+=sbuf;
    }
    
    return(pt);
  }
  catchAll("array::getPrintTreeValues","")
} 



printTree array::getPrintTree(bool values) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf;
    printTree pt;
    
    sprintf(cbuf,"array %p",static_cast<const void*>(this));
    sbuf=cbuf;
    pt+=sbuf;
    
    pt+=getPrintTreeData();
    
    if (values)
      pt+=getPrintTreeValues();
    
    sbuf="   ";
    pt+=sbuf;
    
    return(pt);
  }
  catchAll("array::getPrintTree","")
}



stepList array::getStepList(void) {
  try {
    if (order!=0)
      errThrow("array::getStepList",
	       "   the array has more than one component.");
    
    static stepList SL;
    SL.init();
    step tmp(this);
    SL.pushBack(tmp);
    return(SL);
  }
  catchAll("array::getStepList","")
}



printTree array::getPrintTreeSummary(void) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf,typeName;
    printTree pt;
    
    sprintf(cbuf,"array %p",static_cast<const void*>(this));
    pt+=cbuf;
    
    if (getName()=="")
      sprintf(cbuf,"   anonymous");
    else
      sprintf(cbuf,"   name = %s",getName().c_str());
    pt+=cbuf;
    
    sbuf="   ";
    pt+=sbuf;
    
    return(pt);
  }
  catchAll("array::getPrintTreeSummary","")
}

    
    
set<field*>* array::buildActiveFieldList(void) {
  try {
    FL=set<field*>();
    return(&FL);
  }
  catchAll("array::buildActiveFieldList","")
}



set<field*>* array::getFieldList(void) {
  return(&FL);
}



set<parents*> array::getListOfType(string typeName) {
  set<parents*> S;
  
  try {
    if (isOfType(typeName))
      S.insert(dynamic_cast<parents*>(this));
    
    return(S);
  }
  catchAll("array::getListOfType","")
}



scal& array::getDeterminant(void) const {
  static scal determinant,tmpScal;
  Int factorial,fact;
  Int i,j,index;
  Double sign;
  static Int *permutation=NULL,*perm;
  static Int permutationSize=0,permutationCapacity=0;
  Int permutationID,tmpInt;
  
  try {
    if (order!=2)
      errThrow("array::getDeterminant",
	       "   order of the array must be 2 for inversion.");
    
    if (dim[0]!=dim[1]) {
      determinant=0.0;
      return(determinant);
    }
    
    if (dim[0]>permutationCapacity) {
      delete[] permutation;
      permutation=new Int[dim[0]];
      permutationCapacity=dim[0];
    }
    permutationSize=dim[0];
    
    factorial=1;
    for (i=1;i<=dim[0];i++)
      factorial*=i;
    
    determinant=0.0;
    for (i=0;i<factorial;i++) {
      
      for (j=0;j<dim[0];j++)
	permutation[j]=j;
      fact=factorial;
      sign=1.0;
      perm=permutation;
      permutationID=i;
      
      for (j=0;j<dim[0];j++) {
	fact/=(dim[0]-j);
	index=permutationID/fact;
	if (index!=0) {
	  sign*=-1.0;
	  tmpInt=perm[index];
	  perm[index]=perm[0];
	  perm[0]=tmpInt;
	}
	perm+=1;
	permutationID%=fact;
      }
      tmpScal=sign;
      for (j=0;j<dim[0];j++)
	tmpScal*=componentValues[dim[0]*j+permutation[j]];
      determinant+=tmpScal;
    }
    
    return(determinant);
  }
  catchAll("array::getDeterminant","")
}



