#ifndef ARRAY_BASE_H
#define ARRAY_BASE_H

//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <set>
#include <string>

#include "print/print.h"
#include "types/allTypes.h"
#include "util/utils.h"

namespace keyFE2 {
  
  class field ;
  class function1 ;
  class object ;
  class scal ;
  class stepList ;
  class tensorCoordSys ;
  class timeDep ;
  
  class array : public virtual name, public virtual newPoint, 
		public virtual parents, public virtual boolZero, 
		public virtual print {
											
    friend class Array ;
    friend class arrayComp ;
    friend class arrayField ;
    friend class arrayFilter ;
    friend class arrayInteg ;
    friend class arrayOp1 ;
    friend class arrayOp2 ;
    friend class arrayOpN ;
    friend class arrayTime ;
    friend class globLoc ;
    friend class tensorCoordSys ;
    
    protected :
    
      Int layer ;
      
      Int order ;
      Int *dim ;
      Int nbrVal ;
      scal *componentValues ;

      Int currentComponent ;
    
      tensorCoordSys *TCS ;
    
      std::set<field*> FL ;
    
    public : 
      
      array(void) ;
      array(const array &in) ;
      array(double d) ;
      void copyArrayData(const array &in) ;
      virtual ~array(void) ;
    
      virtual array* makeCloneForArrayTime(timeDep &TD) ;
      
      void setLayer(Int nbr) ;
      Int getLayer(void) const ;
      
      virtual void setOrder(Int nbr) ;
      Int getOrder(void) const ;
      virtual void setDim(Int index,Int nbrComponents) ;
      Int getDim(Int index) const ;
      Int getNbrVal(void) const ;
      
      void setTCS(tensorCoordSys *TCS) ;
      tensorCoordSys* getTCS(void) const ;
      
      virtual bool updateBoolZeroValue() ;
      bool isDiagonal(void) const ;
      
      Int getIndexFromIndices(const Int indices[MAX_ORDER_ARRAY]) const ;
      void getIndicesFromIndex(Int index,Int indices[MAX_ORDER_ARRAY]) const ;
      virtual array& component(const Int &index0=-1,const Int &index1=-1,
			       const Int &index2=-1,const Int &index3=-1,
			       const Int &index4=-1,const Int &index5=-1,
			       const Int &index6=-1,const Int &index7=-1) ;
      virtual scal& componentScal(const Int &index0=-1,const Int &index1=-1,
				  const Int &index2=-1,const Int &index3=-1,
				  const Int &index4=-1,const Int &index5=-1,
				  const Int &index6=-1,const Int &index7=-1) ;
      virtual array& component(const Int index[MAX_ORDER_ARRAY]) ;
      virtual scal& componentScal(const Int index[MAX_ORDER_ARRAY]) ;
      void setCurrentComponent(Int) ;
      Int getCurrentComponent(void) const ;
      scal* getComponentValues(void) const ;
      virtual void operator=(Double) ;
      virtual void operator=(scal&) ;
      virtual void operator=(field&) ;
      virtual void operator=(array&) ;
      
      virtual void makeAtPoint(object *element,Int direct,Double *localCoord,
			       bool makeValue=true,bool makeDeriv=true) ;
      virtual void makeOnObject(object *element,Int direct,bool makeValue,
                                bool makeDeriv) ;
      
      virtual array* makeDeriv(Int nbrScal,scal **scalList,globLoc *GL,
			       std::string derivType) const ;
      virtual array* makeExplicitDeriv(Int nbrScal,scal **scalList) const ;
      virtual array* makeCoordDeriv(globLoc *GL) const ;
      virtual array* makeCorlocDeriv(Int nbrDP) const ;
      virtual array* buildPhysicalDeriv(globLoc *GL) ;
      virtual array* makePhysicalDeriv(globLoc *GL) ;
      
      virtual printTree getPrintTreeData(void) const ;
      virtual printTree getPrintTreeValues(void) const ;
      virtual printTree getPrintTree(bool values) const ;
      virtual printTree getPrintTreeSummary(void) const ;
      
      virtual stepList getStepList(void) ;
      
      virtual std::set<field*>* buildActiveFieldList(void) ;
      std::set<field*>* getFieldList(void) ;
      virtual std::set<parents*> getListOfType(std::string typeName) ;
      
      scal& getDeterminant(void) const ;

      // Execution of the various operators at a given point :
      
      void transpositionExec(array *arrayArg,Int iParam1,Int iParam2) ;
      void contractionExec(array *arrayArg,Int iParam1,Int iParam2) ;
      void inversionExec(array *arrayArg) ;
      void functionExec(array *arrayArg,function1 *fParam,bool makeValue=true,
			bool makeDeriv=true) ;
      void indicesPermutationExec(array *arrayArg,Int *IList) ;
      void discretizedTranspositionExec(array *arrayArg) ;
      
      void specialContractExec(array *arrayArg1,array *arrayArg2,Int IParam1) ;
      void additionExec(array *arrayArg1,array *arrayArg2) ;
      void soustractionExec(array *arrayArg1,array *arrayArg2) ;
      void exteriorMultiplicationExec(array *arrayArg1,array *arrayArg2) ;
      void vectProductExec(array *arrayArg1,array *arrayArg2) ;
      void contractedMultiplicationExec(array *arrayArg1,array *arrayArg2,
					Int IParam1,Int IParam2) ;
      void accumulationExec(array *arrayArg1,array *arrayArg2,Int IParam1) ;
      void substitutionExec(array *arrayArg1,array *arrayArg2,Int IParam1,
			    Int IParam2) ;
      
  } ;
   
}

#endif
    
    
