
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <iostream>
#include <iomanip>
#include <typeinfo>

#include "tensorCoordSys_base.h"
#include "array/array_base.h"
#include "array/arrayFunctionsBuild.h"
#include "array/arrayOp1.h"
#include "array/arrayOp2.h"
#include "array/arrayComp.h"
#include "err/err.h"
#include "function/functions.h"

using namespace std ;
using namespace keyFE2 ;



namespace keyFE2 {
  
  
  
  tensorCoordSys* build1DCartesianTCS(array *coord) {
    try {
      if (coord->getOrder()!=1)
	errThrow("build1DCartesianTCS",
		 "order of the coordinates array should be 1.");
      if (coord->getDim(0)!=1)
	errThrow("build1DCartesianTCS",
		 "dimension of the coordinate array should be 1.");
      Int dim=coord->getDim(0);
      
      tensorCoordSys *TCS=new tensorCoordSys;
      
      if (coord->getName()=="")
	coord->setName("Cartesian 1D TCS");
      TCS->setCoord(coord);
      
      array *metric;
      metric=new array;
      metric->setOrder(2);
      metric->setDim(0,dim);
      metric->setDim(1,dim);
      metric->setName("Cartesian covariant metric");
      for (Int i=0;i<dim;i++)
	metric->component(i,i)=1.0;
      TCS->setCovaMetric(metric);
      TCS->setContraMetric(metric);
      
      array* christoffel;
      christoffel=new array;
      christoffel->setOrder(3);
      christoffel->setDim(0,dim);
      christoffel->setDim(1,dim);
      christoffel->setDim(2,dim);
      christoffel->setName("Cartesian first Christoffel Symbol");
      TCS->setChristoffel(christoffel);
      
      TCS->setType("Cartesian1D");
      
      return(TCS);
    }
    catchAll("build1DCartesianTCS","")
  }
  
  
  
  tensorCoordSys* build2DCartesianTCS(array *coord) {
    try {
      if (coord->getOrder()!=1)
	errThrow("build2DCartesianTCS",
		 "order of the coordinates array should be 1.");
      if (coord->getDim(0)!=2)
	errThrow("build2DCartesianTCS",
		 "dimension of the coordinate array should be 2.");
      Int dim=coord->getDim(0);
      
      tensorCoordSys *TCS=new tensorCoordSys;
      
      if (coord->getName()=="")
	coord->setName("Cartesian 2D TCS");
      TCS->setCoord(coord);
      
      array *metric;
      metric=new array;
      metric->setOrder(2);
      metric->setDim(0,dim);
      metric->setDim(1,dim);
      metric->setName("Cartesian covariant metric");
      for (Int i=0;i<dim;i++)
	metric->component(i,i)=1.0;
      TCS->setCovaMetric(metric);
      TCS->setContraMetric(metric);
      
      array* christoffel;
      christoffel=new array;
      christoffel->setOrder(3);
      christoffel->setDim(0,dim);
      christoffel->setDim(1,dim);
      christoffel->setDim(2,dim);
      christoffel->setName("Cartesian first Christoffel Symbol");
      TCS->setChristoffel(christoffel);
      
      TCS->setType("Cartesian2D");
      
      return(TCS);
    }
    catchAll("build2DCartesianTCS","")
  }
  
  
  
  tensorCoordSys* build3DCartesianTCS(array *coord) {
    try {
      if (coord->getOrder()!=1)
	errThrow("build3DCartesianTCS",
		 "order of the coordinates array should be 1.");
      if (coord->getDim(0)!=3)
	errThrow("build3DCartesianTCS",
		 "dimension of the coordinate array should be 3.");
      Int dim=coord->getDim(0);
      
      tensorCoordSys *TCS=new tensorCoordSys;
      
      if (coord->getName()=="")
	coord->setName("Cartesian 3D TCS");
      TCS->setCoord(coord);
      
      array *metric;
      metric=new array;
      metric->setOrder(2);
      metric->setDim(0,dim);
      metric->setDim(1,dim);
      metric->setName("Cartesian covariant metric");
      for (Int i=0;i<dim;i++)
	metric->component(i,i)=1.0;
      TCS->setCovaMetric(metric);
      TCS->setContraMetric(metric);
      
      array* christoffel;
      christoffel=new array;
      christoffel->setOrder(3);
      christoffel->setDim(0,dim);
      christoffel->setDim(1,dim);
      christoffel->setDim(2,dim);
      christoffel->setName("Cartesian first Christoffel Symbol");
      TCS->setChristoffel(christoffel);
      
      TCS->setType("Cartesian3D");
      
      return(TCS);
    }
    catchAll("build3DCartesianTCS","")
  }
  
  
  
  tensorCoordSys* buildRZTCylindricalTCS(array *coord) {
    
    // NB: the coordinate components are r, theta and z.
    
    try {
      
      if (coord->getOrder()!=1)
	errThrow("builRZTCylindricalTCS",
		 "order of the coordinates array should be 1.");
      Int dim=coord->getDim(0);
      if (dim!=3)
	errThrow("builRZTCylindricalTCS",
		 "dimension of the coordinate array should be 3.");
      
      tensorCoordSys *TCS=new tensorCoordSys;
      
      if (coord->getName()=="")
	coord->setName("Cylindrical coordinates");
      TCS->setCoord(coord);
      
      array *metric;
      
      metric=new array;
      metric->setOrder(2);
      metric->setDim(0,dim);
      metric->setDim(1,dim);
      metric->setName("Cylindrical covariant metric");
      TCS->setCovaMetric(metric);
      
      metric=new array;
      metric->setOrder(2);
      metric->setDim(0,dim);
      metric->setDim(1,dim);
      metric->setName("Cylindrical contravariant metric");
      TCS->setContraMetric(metric);
      
      array* christoffel;
      christoffel=new array;
      christoffel->setOrder(3);
      christoffel->setDim(0,dim);
      christoffel->setDim(1,dim);
      christoffel->setDim(2,dim);
      christoffel->setName("Cylindrical second Christoffel Symbol");
      TCS->setChristoffel(christoffel);
      
      TCS->setType("CylindricalRZT");
      
      return(TCS);
    }
    catchAll("buildRZTCylindricalTCS","")
  }
  
  
  
}
