
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <set>
#include <iostream>
#include <stdio.h>

#include "domain_base.h"
#include "err/err.h"
#include "mesh/mesh_base.h"
#include "object/object_base.h"
#include "object/objectTypes.h"
#include "print/printTree.h"

using namespace std ;
using namespace keyFE2 ;



domain::domain(void) : name(), parents() {
  baseMesh=NULL;

  dim=-1;
  objectsNbr=0;
  objectsLst=NULL;
  objectsDirect=NULL;

  allObjectsNbr=0;
  allObjectsLst=NULL;
}



domain::domain(string s) : name(s), parents() {
  baseMesh=NULL;

  dim=-1;
  objectsNbr=0;
  objectsLst=NULL;
  objectsDirect=NULL;

  allObjectsNbr=0;
  allObjectsLst=NULL;
  makeAllObjectsLst();
}



domain::domain(const domain &dom) : name(), parents() {
  try {
    baseMesh=dom.baseMesh;
    
    dim=dom.dim;
    objectsNbr=dom.objectsNbr;
    if (objectsNbr!=0) {
      objectsLst=new object* [objectsNbr];
      objectsDirect=new Int[objectsNbr];
    }
    else {
      objectsLst=NULL;
      objectsDirect=NULL;
    }
    for (Int i=0;i<objectsNbr;i++) {
      objectsLst[i]=dom.getObject(i);
      objectsDirect[i]=dom.getObjectDirect(i);
    }
    
    allObjectsNbr=0;
    allObjectsLst=NULL;
    makeAllObjectsLst();
  }
  catchAll("domain::domain(const domain &dom)","")
}



domain::~domain(void) {
  try {
    baseMesh=NULL;
    dim=-1;
    
    objectsNbr=0;
    if (objectsLst!=NULL) {
      delete[] objectsLst;
      objectsLst=NULL;
    }
    if (objectsDirect!=NULL) {
      delete[] objectsDirect;
      objectsDirect=NULL;
    }
    
    allObjectsNbr=0;
    if (allObjectsLst!=NULL) {
      delete[] allObjectsLst;
      allObjectsLst=NULL;
    }
  }
  catchAll("domain::~domain(void)","")
}



void domain::setBaseMesh(mesh *BM) {
  baseMesh=BM;
}



mesh* domain::getBaseMesh(void) const {
  return(baseMesh);
}



void domain::setObjectsNbr(Int nbr) {
  try {
    objectsNbr=nbr;
    
    if (objectsLst!=NULL)
      delete[] objectsLst;
    if (objectsDirect!=NULL)
      delete[] objectsDirect;
    if (objectsNbr!=0) {
      objectsLst=new object* [nbr];
      objectsDirect=new Int[nbr];
    }
    else {
      objectsLst=NULL;
      objectsDirect=NULL;
    }
    
    allObjectsNbr=0;
    if (allObjectsLst!=NULL) {
      delete[] allObjectsLst;
      allObjectsLst=NULL;
    }
  }
  catchAll("void domain::setObjectsNbr(Int nbr)","")
}



Int domain::getObjectsNbr(void) const {
  return(objectsNbr);
}



void domain::setObject(Int pos,object *element,Int orientation) {
  objectsLst[pos]=element;
  objectsDirect[pos]=orientation;
}



object* domain::getObject(Int index) const {
  return(objectsLst[index]);
}



Int domain::getObjectDirect(Int index) const {
  return(objectsDirect[index]);
}



void domain::setAllObjectsNbr(Int nbr) {
  try {
    allObjectsNbr=nbr;
    
    if (allObjectsLst!=NULL)
      delete[] allObjectsLst;
    if (allObjectsNbr!=0)
      allObjectsLst=new object* [nbr];
    else
      allObjectsLst=NULL;
  }
  catchAll("void domain::setAllObjectsNbr(Int nbr)","")
}



Int domain::getAllObjectsNbr(void) const {
  return(allObjectsNbr);
}



void domain::setAllObject(Int pos,object *element) {
  allObjectsLst[pos]=element;
}



object* domain::getAllObject(Int index) const {
  return(allObjectsLst[index]);
}



Int domain::getAllObject(object *obj) const {
  try {
    Int up=allObjectsNbr-1,down=0;
    Int index;
    
    if (allObjectsLst[up]<obj)
      return(-1);
    if (obj<allObjectsLst[down])
      return(-1);
    
    while (up-down>1) {
      index=(up+down)/2;
      if (obj==allObjectsLst[index])
	return(index);
      else if (obj<allObjectsLst[index])
	up=index;
      else if (allObjectsLst[index]<obj)
	down=index;
    }
    if (obj==allObjectsLst[up])
      return(up);
    else if (obj==allObjectsLst[down])
      return(down);
    
    return(-1);
  }
  catchAll("Int domain::getAllObject(object *obj) const","")
}



void domain::makeAllObjectsLst(void) {
  try {
    allObjectsNbr=0;
    if (allObjectsLst!=NULL) {
      delete[] allObjectsLst;
      allObjectsLst=NULL;
    }
    
    set<object*>* list;
    list=new set<object*> [getDim()+1];
    set<object*>::iterator soIt;
    
    // One copies the list of highest dimension objects :
    
    for (Int i=0;i<getObjectsNbr();i++)
      list[getDim()].insert(getObject(i));
    
    // One recursively builds the lower order lists :
    
    for (Int dim=getDim()-1;dim>=0;dim--) {
      for (soIt=list[dim+1].begin();soIt!=list[dim+1].end();soIt++) {
	Int nbr=(*soIt)->getMinusNbr();
	for (Int j=0;j<nbr;j++)
	  list[dim].insert((*soIt)->getMinus(j));
      }
    }
    
    // One counts the total number of objects :
    
    Int size=0;
    for (Int dim=getDim();dim>=0;dim--)
      size+=list[dim].size();
    
    // One allocates and fills the total list of objects :
    
    Int counter=0;
    allObjectsNbr=size;
    if (allObjectsNbr!=0)
      allObjectsLst=new object* [allObjectsNbr];
    else
      allObjectsLst=NULL;
    for (Int dim=0;dim<=getDim();dim++)
      for (soIt=list[dim].begin();soIt!=list[dim].end();soIt++) {
      allObjectsLst[counter]=(*soIt);
      counter++;
    }
    
    delete[] list;
  }
  catchAll("void domain::makeAllObjectsLst(void)","")
}



printTree domain::getPrintTreeData(void) const {
  char cbuf[STRING_LENGTH];
  string sbuf;
  printTree pt;
  Int i;
  Int nbrNodes,nbrEdges,nbrTriangles,nbrQuadrangles;
  
  try {
    if (getName()=="")
      sprintf(cbuf,"   anonymous");
    else
      sprintf(cbuf,"   name = %s",getName().c_str());
    sbuf=cbuf;
    pt+=sbuf;
    
    sprintf(cbuf,"   number of objects = %d",objectsNbr);
    sbuf=cbuf;
    pt+=sbuf;
    
    nbrNodes=nbrEdges=nbrTriangles=nbrQuadrangles=0;
    for (i=0;i<objectsNbr;i++) {
      switch(objectsLst[i]->type) {
      case node :
	nbrNodes++;
	break;
      case segment :
	nbrEdges++;
	break;
      case triangle :
	nbrTriangles++;
	break;
      case quadrangle :
	nbrQuadrangles++;
	break;
      }
    }
    if (nbrNodes>0) {
      sprintf(cbuf,"   number of nodes = %d",nbrNodes);
      sbuf=cbuf;
      pt+=sbuf;
    }
    if (nbrEdges>0) {
      sprintf(cbuf,"   number of edges = %d",nbrEdges);
      sbuf=cbuf;
      pt+=sbuf;
    }
    if (nbrTriangles>0) {
      sprintf(cbuf,"   number of triangles = %d",nbrTriangles);
      sbuf=cbuf;
      pt+=sbuf;
    }
    if (nbrQuadrangles>0) {
      sprintf(cbuf,"   number of quadrangles = %d",nbrQuadrangles);
      sbuf=cbuf;
      pt+=sbuf;
    }
    
    return(pt);
  }
  catchAll("domain::getPrintTreeData","")  
}



printTree domain::getPrintTreeValues(void) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf;
    printTree pt;
    Int i;
    object *objZero;
    
    objZero=baseMesh->getObjectsLst();
    
    sbuf="   list of objects:";
    pt+=sbuf;
        
    if (objectsNbr!=0) {
      sbuf="   ";
      for (i=0;i<objectsNbr;i++) {
	if (i!=0&&i%10==0) {
	  pt+=sbuf;
	  sbuf="   ";
	}
	if (objectsDirect[i]>=0)
	  sprintf(cbuf,"%6d",(objectsLst[i]+1)-objZero);
	else
	  sprintf(cbuf,"%6d",objZero-(objectsLst[i]+1));
	sbuf+=cbuf;
      }
      pt+=sbuf;
    }
    
    return(pt);
  }
  catchAll("domain::getPrintTreeValues","")
} 



printTree domain::getPrintTree(bool values) const {
  char cbuf[STRING_LENGTH];
  string sbuf;
  printTree pt;
  
  try {
    sprintf(cbuf,"domain %p",static_cast<const void*>(this));
    sbuf=cbuf;
    pt+=sbuf;
    
    pt+=getPrintTreeData();
    
    if (values)
      pt+=getPrintTreeValues();
    
    sbuf="   ";
    pt+=sbuf;
    
    return(pt);
  }
  catchAll("domain::getPrintTree","")
}



printTree domain::getPrintTreeSummary(void) const {
  char cbuf[STRING_LENGTH];
  string sbuf;
  printTree pt;
  
  try {
    sprintf(cbuf,"domain %p",static_cast<const void*>(this));
    sbuf=cbuf;
    pt+=sbuf;
    
    if (getName()=="")
      sprintf(cbuf,"   anonymous");
    else
      sprintf(cbuf,"   name = %s",getName().c_str());
    sbuf=cbuf;
    pt+=sbuf;
    
    sbuf="   ";
    pt+=sbuf;
    
    return(pt);
  }
  catchAll("domain::getPrintTreeSummary","")
}



void domain::splitWrite(ostream &myCout) {
  try {
    Int direction[20];
    object* liste[20];
    Int nbrDown,i,j;
    
    for (i=0;i<objectsNbr;i++) {
      myCout << "Splitting object " << i << " (" 
	<< objectsLst[i]-baseMesh->getObject(0)+1 << ") of type "
	  << objectsLst[i]->getTypeName() << " :" << endl ;
      nbrDown=objectsLst[i]->split(objectsDirect[i],liste,direction);
      for (j=0;j<nbrDown;j++) {
	myCout << "   " << j << " (" << liste[j]-baseMesh->getObject(0)+1
	  << ") of type " << liste[j]->getTypeName()
	    << " and direction " << direction[j] << "." << endl ;
      }
    }
  }
  catchAll("void domain::splitWrite(ostream &myCout)","")
}



void domain::splitWrite(string outFileName) {
  try {
    ofstream outFile(outFileName.c_str());
    this->splitWrite(outFile);
    outFile.close();
  }
  catchAll("void domain::splitWrite(string outFileName)","")
}





