
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <string>

#include "Field.h"
#include "field_base.h"
#include "array/Array.h"
#include "array/array_base.h"
#include "domain/Domain.h"
#include "domain/domain_base.h"
#include "err/err.h"
#include "mesh/Mesh.h"
#include "mesh/mesh_base.h"
#include "param/param.h"
#include "param/paramList.h"
#include "print/printTree.h"

using namespace std ;
using namespace keyFE2 ;



Field::Field(void) {
  try {
    field *f=new field;
    data=NULL;
    setData(f);
  }
  catchAll("Field::Field(void)","")
}



Field::Field(const Field &F) {
  try {
    data=NULL;
    setData(F.getData());
  }
  catchAll("Field::Field(const Field &F)","")
}



Field::Field(param P) {
  try {
    paramList PL;
    PL.pushBack(P);
    Field F(PL);
    data=NULL;
    setData(F.getData());
  }
  catchAll("Field::Field(param P)","")
}



Field::Field(paramList PL) {
  try {
    domain *dom=NULL;
    Int nbrComponents;
    string interpolationTypeName="";
    string fieldName="";
    
    paramList plDomain=PL.getListOfType(DOM);
    PL=PL.getListNotOfType(DOM);
    paramList plNbrComponents=PL.getListOfType(NBRCOMPONENTS);
    PL=PL.getListNotOfType(NBRCOMPONENTS);
    paramList plInterpolation=PL.getListOfType(INTERPOLATION);
    PL=PL.getListNotOfType(INTERPOLATION);
    paramList plName=PL.getListOfType(NAME);
    PL=PL.getListNotOfType(NAME);
    paramList plComponentName=PL.getListOfType(COMPONENTNAME);
    PL=PL.getListNotOfType(COMPONENTNAME);
    paramList plInterpMesh=PL.getListOfType(INTERPMESH);
    PL=PL.getListNotOfType(INTERPMESH);
    paramList plMinRef=PL.getListOfType(MINREF);
    PL=PL.getListNotOfType(MINREF);
        
    if (PL.getNbrParams()!=0) {
      string tmpString=strFromInt(PL.getNbrParams())+" invalid parameters :";
      for (Int i=0;i<PL.getNbrParams();i++) {
	tmpString+=" ";
	tmpString+=PL[i].getName();
      }
      tmpString+=".\n";
      errWrite("Field construction",tmpString);
    }
    
    if (plDomain.getNbrParams()>1)
      errWrite("Field construction","only one DOM argument is allowed.");
    if (plDomain.getNbrParams()==1) {
      dom=dynamic_cast<domain*>(plDomain[0].getParentData());
      if (dom==NULL)
	errWrite("Field construction",
		 "DOM= has to be followed by a Domain argument.");
    }
    
    if (plNbrComponents.getNbrParams()>1)
      errWrite("Field construction",
	       "only one NBRCOMPONENTS argument is allowed.");
    
    if (plInterpolation.getNbrParams()>1)
      errWrite("Field construction",
	       "only one INTERPOLATION argument is allowed.");
    
    if (plInterpMesh.getNbrParams()>1)
      errWrite("Field construction",
	       "only one INTERPMESH argument is allowed.");
    
    if (plName.getNbrParams()>1) {
      errWrite("Field construction","only one NAME argument is allowed.");
    }
    else if (plName.getNbrParams()==1) {
      fieldName=plName[0].getStringData();
      if (fieldName=="")
	errWrite("Field construction",
		 "NAME= has to be followed by a string argument.");
    }
    
    Double minRef=1.0e-8;
    if (plMinRef.getNbrParams()>1)
      errWrite("Field construction","maximum one MINREF argument is allowed.");
    if (plMinRef.getNbrParams()==1) {
      minRef=plMinRef[0].getDoubleData();
      if (minRef<=0.0)
	errWrite("Field construction",
		 "MINREF= has to be followed by a positive Double argument.");
    }
    
    if (!err::getOk())
      errThrow("Field construction","bailing out due to preceding errors.");
    
    if (plNbrComponents.getNbrParams()==1) {
      nbrComponents=plNbrComponents[0].getIntData();
    }
    else
      nbrComponents=1;
    
    if (plInterpolation.getNbrParams()==1) {
      interpolationTypeName=plInterpolation[0].getStringData();
    }
    else
      interpolationTypeName="none";
    
    string* compNames=NULL;
    bool* compDefined=NULL;
    compNames=new string[nbrComponents];
    compDefined=new bool[nbrComponents];
    for (Int i=0;i<nbrComponents;i++) {
      compNames[i]="";
      compDefined[i]=false;
    }
    if (plComponentName.getNbrParams()>nbrComponents) {
      errWrite("Field construction",
	       "cannot have more COMPONENTNAME= arguments than order of the array.");
    }
    else {
      for (Int i=0;i<plComponentName.getNbrParams();i++) {
	param PCN(plComponentName[i]);
	const Int *comps=PCN.getComponentsData();
	string compName=PCN.getStringData();
	if (comps[1]!=-1) {
	  errWrite("Field construction",
		   "exactly one component index allowed for fields.");
	}
	else if (comps[0]<0||comps[0]>=nbrComponents) {
	  errWrite("Field construction",
		   string("invalid value of component index : ")
		   +strFromInt(comps[0])+".");
	}
	else {
	  if (compDefined[comps[0]]) {
	    errWrite("Field construction",
		     string("name of component ")+strFromInt(comps[0])
		     +" defined twice.");
	  }
	  else {
	    compNames[comps[0]]=compName;
	    compDefined[comps[0]]=true;
	  }
	}  
      }
    }
    
    if (!err::getOk())
      errThrow("Field construction","bailing out due to preceding errors");
    
    field *f=new field(dom,nbrComponents,interpolationTypeName);
    f->setName(fieldName);
    for (Int i=0;i<nbrComponents;i++)
      f->setComponentName(i,compNames[i]);
    
    if (plInterpMesh.getNbrParams()==1) {
      param pInterpMesh(plInterpMesh[0]);
      mesh *MESH;
      MESH=dynamic_cast<mesh*>(pInterpMesh.getParentData());
      field *coord=MESH->getCoordinateField(dom);
      f->interpolate(coord,dom,coord);
      delete coord;
    }
    
    f->setMinRefForRelativeVariation(minRef);
    
    delete[] compNames;
    delete[] compDefined;
    
    data=NULL;
    setData(f);  
  }
  catchAll("Field::Field(paramList PL)","")
}



Field::~Field(void) {
  try {
    setData(NULL);
  }
  catchAll("Field::~Field(void)","")
}



void Field::setData(field *newFieldData) {
  try {
    field *f=getData();
    delinkAndDelete(f,&data);
    if (newFieldData!=NULL)
      newFieldData->addParent(&data);
    data=static_cast<void*>(newFieldData);
    //  field::updateAllFieldsIndices();
  }
  catchAll("void Field::setData(field *newFieldData)","")
}



field* Field::getData(void) const {
  return(static_cast<field*>(data));
}



void Field::setName(string s) {
  try {
    getData()->setName(s);
  }
  catchAll("void Field::setName(string s)","")
}



string Field::getName(void) const {
  try {
    return(getData()->getName());
  }
  catchAll("string Field::getName(void) const","")
}



void Field::operator=(const Field &F) {
  if (this==&F)
    return;
  
  try {
    setData(F.getData());
  }
  catchAll("void Field::operator=(const Field &F)","")
}



void Field::makeCopy(const Field &F) {
  try {
    setData(new field(*(F.getData())));
  }
  catchAll("void Field::makeCopy(const Field &F)","")
}



Int Field::getNbrComponents(void) const {
  try {
    return(getData()->getNbrComponents());
  }
  catchAll("Int Field::getNbrComponents(void) const","")
}



string Field::getComponentName(Int index) const {
  try {
    if (index>getData()->getNbrComponents()||index<0) {
      errThrow("Field::getComponentName(Int index)","invalid index value");
    }
    else
      return(getData()->getComponentName(index));
  }
  catchAll("string Field::getComponentName(Int index) const","")
}



Int Field::getInterpolationType(void) const {
  try {
    return(getData()->getInterpolationType());
  }
  catchAll("Int Field::getInterpolationType(void) const","")
}



string Field::getInterpolationTypeName(void) const {
  try {
    return(getData()->getInterpolationTypeName());
  }
  catchAll("string Field::getInterpolationTypeName(void) const","")
}



void Field::interpolate(Array &origArray,Domain &interpDom,Array &coordArray) {
  try {
    field *f=getData();
    array *a=origArray.getData();
    domain *dom=interpDom.getData();
    array *c=coordArray.getData();
    f->interpolate(a,dom,c);
  }
  catchAll("void Field::interpolate(Array &origArray,Domain &interpDom,Array &coordArray)","")
}



void Field::interpolate(Field &origField,Domain &interpDom,Field &coordField) {
  try {
    field *f2=getData();
    field *f1=origField.getData();
    domain *dom=interpDom.getData();
    field *c=coordField.getData();
    f2->interpolate(f1,dom,c);
  }
  catchAll("void Field::interpolate(Field &origField,Domain &interpDom)","")
}



printTree Field::getPrintTree(bool values) const {
  try {
    return(getData()->getPrintTree(values));
  }
  catchAll("printTree Field::getPrintTree(bool values) const","")
}



Field& Field::component(Int index) {
  try {
    field *f=getData();
    if (index>f->getNbrComponents()||index<0) {
      errThrow("Field::getComponentName(Int index)","invalid index value");
    }
    else {
      f->component(index);
      return(*this);
    }
  }
  catchAll("Field& Field::component(Int index)","")
}



void Field::write(ostream &os) const {
  try {
    getData()->write(os);
  }
  catchAll("void Field::write(ostream &os) const","")
}



void Field::write(void) const {
  try {
    getData()->write();
  }
  catchAll("void Field::write(void) const","")
}



void Field::writeValues(ostream &os) const {
  try {
    getData()->writeValues(os);
  }
  catchAll("void Field::writeValues(ostream &os) const","")
}



void Field::writeValues(void) const {
  try {
    getData()->writeValues();
  }
  catchAll("void Field::writeValues(void) const","")
}



