
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <algorithm>
#include <iomanip>
#include <list>
#include <map>
#include <vector>

#include "mesh_base.h"
#include "array/array_base.h"
#include "domain/Domain.h"
#include "domain/domain_base.h"
#include "err/err.h"
#include "object/object_base.h"
#include "scal/scal.h"

using namespace std ;
using namespace keyFE2 ;



domain* mesh::setDomainsNbr(Int nbr) {
  try {
    Int i,j;
    domain *currentDom;
    void **p;
    if (domainsLst!=NULL) {
      for (i=0;i<domainsNbr;i++) {
	currentDom=domainsLst+i;
	for (j=0;j<currentDom->getParentsNbr();j++) {
	  p=static_cast<void**>(currentDom->getParent(0));
	  if (p!=static_cast<void*>(domainsLst+i)) {
	    *p=static_cast<void*>(NULL);
	    currentDom->supParent(p);
	  }
	}
      }
      delete[] domainsLst;
      domainsLst=NULL;
    }
    if (domainsIntId!=NULL) {
      delete[] domainsIntId;
      domainsIntId=NULL;
    }
    domainsNbr=nbr;
    
    if (domainsNbr!=0) {
      domainsLst=new domain[domainsNbr];
      for (i=0;i<domainsNbr;i++)
	domainsLst[i].addParent(domainsLst+i);
      domainsIntId=new Int[domainsNbr];
    }
    
    return(domainsLst);
  }
  catchAll("domain* mesh::setDomainsNbr(Int nbr)","")
}



Int mesh::getDomainsNbr(void) const {
  return(domainsNbr);
}



Domain mesh::getDefDomain(void) const {
  try {
    Int i;
    domain defDom;
    
    // One calculates the dimension of the domain :
    Int dim=0;
    for (i=0;i<objectsNbr;i++)
      dim=max(dim,objectsLst[i].getDim());
    
    // One calculates the number of objects having the maximum dimension :
    Int nbr=0;
    for (i=0;i<objectsNbr;i++)
      if (objectsLst[i].getDim()==dim)
	nbr++;
    
    // One creates a domain collecting those objects :
    defDom.setObjectsNbr(nbr);
    Int counter=0;
    for (i=0;i<objectsNbr;i++)
      if (objectsLst[i].getDim()==dim) {
      defDom.setObject(counter,objectsLst+i,
		       objectsLst[i].getMinusNbr());
      counter++;
    }
    
    // Additional information :
    defDom.setDim();
    defDom.setBaseMesh(const_cast<mesh*>(this));
    defDom.makeAllObjectsLst();
    
    if (defDom.getDim()==1) {
      domain tmp;
      tmp.reorder1DDomain(defDom);
      defDom=tmp;
    }
    
    // The member data :
    
    domain *dom=new domain(defDom);
    Domain Dom;
    Dom.setData(dom);
    return(Dom);
  }
  catchAll("Domain mesh::getDefDomain(void) const","")
}



Domain mesh::getDomainFromIntId(Int index) const {
  try {
    Domain Dom;
    domain *dom;
    for (Int i=0;i<domainsNbr;i++) {
      if (domainsIntId[i]==index) {
	dom=new domain(domainsLst[i]);
	Dom.setData(dom);
	return(Dom);
      }
    }
    errThrow("mesh::getDomainFromIntId",
	     "   invalid domain identifier : "+strFromInt(index)+".");
  }
  catchAll("Domain mesh::getDomainFromIntId(Int index) const","")
}



Domain mesh::getDomain(Int index) const {
  try {
    if (index<0||index>=domainsNbr) {
      errThrow("mesh::getDomain",
	       "   invalid domain index : "+strFromInt(index)+".");
    }
    domain *dom=new domain(domainsLst[index]);
    Domain Dom;
    Dom.setData(dom);
    return(Dom);
  }
  catchAll("Domain mesh::getDomain(Int index) const","")
}



Int mesh::getDomainIntId(Int index) const {
  try {
    if (index<0||index>=domainsNbr) {
      errThrow("mesh::getDomain",
	       "   invalid domain index : "+strFromInt(index)+".");
    }
    return(domainsIntId[index]);
  }
  catchAll("Int mesh::getDomainIntId(Int index) const","")
}



// What follows is devoted to the renumbering of meshes :



static vector<object*> elemVect;
static map<object*,Int> elemsAppear;
static map<object*,Int> elemsDisappear;
static vector<object*> nodeVect;
static map<object*,bool> nodeMap;
static map<object*,Int> nodesAppear;
static map<object*,Int> nodesDisappear;

class compare_elements {
  public :
    bool operator()(object *elem1,object *elem2) {
    if (elemsDisappear[elem1]<elemsDisappear[elem2])
      return(true);
    else if (elemsDisappear[elem1]>elemsDisappear[elem2])
      return(false);
    if (elemsAppear[elem1]<elemsAppear[elem2])
      return(true);
    else if (elemsAppear[elem1]>elemsAppear[elem2])
      return(false);
    return(true);
  }
} ;
        
class compare_nodes {
  public :
    bool operator()(object *node1,object *node2) {
    if (nodesDisappear[node1]<nodesDisappear[node2])
      return(true);
    else if (nodesDisappear[node1]>nodesDisappear[node2])
      return(false);
    if (nodesAppear[node1]<nodesAppear[node2])
      return(true);
    else if (nodesAppear[node1]>nodesAppear[node2])
      return(false);
    return(true);
  }
} ;
        
void mesh::renumberDefDomainFrontal(const mesh &arg,domain &dom) {
  try {
    Int i,j;
    Int counter,nodeIndex,elemIndex;
    Domain DefDomain;
    domain *defDomain;
    object* objList[50];
    object *firstNode,*firstElem;
    Int objDirect[50],objNbr;
    list<object*>::iterator loIt,loItB;
    vector<object*>::iterator voItA;
    
    DefDomain=arg.getDefDomain();
    defDomain=DefDomain.getData();
    
    // Les lments pointent vers leurs noeuds :
    
    list<object*> *elemsNodes=new list<object*>[defDomain->objectsNbr];
    firstElem=NULL;
    for (i=0;i<defDomain->objectsNbr;i++) {
      if (firstElem==NULL||defDomain->objectsLst[i]<firstElem)
	firstElem=defDomain->objectsLst[i];
      objNbr=defDomain->objectsLst[i]->split(defDomain->objectsDirect[i],
					     objList,objDirect);
      for (j=0;j<objNbr;j++)
	if (objList[j]->getType()==1)
	  elemsNodes[i].push_back(objList[j]);
    }
    
    // Les noeuds pointent vers leurs lments :
    
    list<object*> *nodesElems=new list<object*>[arg.nodesNbr];
    firstNode=arg.objectsLst;
    for (i=0;i<defDomain->objectsNbr;i++) {
      for (loIt=elemsNodes[i].begin();loIt!=elemsNodes[i].end();loIt++)
	nodesElems[*loIt-firstNode].push_back(firstElem+i);
    }
    
    // On dfinit et remplit les listes rordonnes de mailles et noeuds :
    
    map<object*,int> reorderedElems;
    map<object*,int> reorderedNodes;
    Int elemCounter=0;
    Int nodeCounter=0;
    
    for (i=0;i<dom.objectsNbr;i++) {
      objNbr=dom.objectsLst[i]->split(dom.objectsDirect[i],objList,objDirect);
      for (j=0;j<objNbr;j++) {
	if (objList[j]->getType()==1&&nodeMap.find(objList[j])==nodeMap.end()) {
	  nodeVect.push_back(objList[j]);
	  nodeMap[objList[j]]=true;
	}
      }
    }
    for (voItA=nodeVect.begin();voItA!=nodeVect.end();voItA++,nodeCounter++)
      reorderedNodes[*voItA]=nodeCounter;
    
    while(nodeCounter!=arg.nodesNbr||elemCounter!=defDomain->getObjectsNbr()) {
      elemVect.clear();
      elemsAppear.clear();
      elemsDisappear.clear();
      counter=0;
      for (voItA=nodeVect.begin();voItA!=nodeVect.end();voItA++,counter++) {
	nodeIndex=(*voItA)-firstNode;
	for (loItB=nodesElems[nodeIndex].begin();
	     loItB!=nodesElems[nodeIndex].end();loItB++) {
	  if (reorderedElems.find(*loItB)==reorderedElems.end()) {
	    if (elemsAppear.find(*loItB)==elemsAppear.end()) {
	      elemsAppear[*loItB]=counter;
	      elemVect.push_back(*loItB);
	    }
	    elemsDisappear[*loItB]=counter;
	  }
	}
      }
      sort(elemVect.begin(),elemVect.end(),compare_elements());
      for (voItA=elemVect.begin();voItA!=elemVect.end();voItA++,elemCounter++)
	reorderedElems[*voItA]=elemCounter;
      
      nodeVect.clear();
      nodesAppear.clear();
      nodesDisappear.clear();
      counter=0;
      for (voItA=elemVect.begin();voItA!=elemVect.end();voItA++,counter++) {
	elemIndex=*voItA-firstElem;
	for (loItB=elemsNodes[elemIndex].begin();
	     loItB!=elemsNodes[elemIndex].end();loItB++) {
	  if (reorderedNodes.find(*loItB)==reorderedNodes.end()) {
	    if (nodesAppear.find(*loItB)==nodesAppear.end()) {
	      nodesAppear[*loItB]=counter;
	      nodeVect.push_back(*loItB);
	    }
	    nodesDisappear[*loItB]=counter;
	  }
	}
      }
      sort(nodeVect.begin(),nodeVect.end(),compare_nodes());
      for (voItA=nodeVect.begin();voItA!=nodeVect.end();voItA++,nodeCounter++)
	reorderedNodes[*voItA]=nodeCounter;
    }
    
    // One fills member data of the new mesh :
    
    type=arg.type;
    
    nodesNbr=arg.nodesNbr;
    if (nodesNbr!=0)
      coord=new Double[nodesNbr][3];
    else
      coord=NULL;
    Int elemsNbr=defDomain->objectsNbr;
    Int edgesNbr=arg.objectsNbr-nodesNbr-elemsNbr;
    
    object *firstNodeB=setObjectsNbr(arg.objectsNbr);
    map<object*,object*> objCorresp;
    for (i=0;i<objectsNbr;i++)
      objCorresp[firstNode+i]=firstNodeB+i;
    
    map<object*,Int>::iterator moiIt;
    for (moiIt=reorderedNodes.begin();moiIt!=reorderedNodes.end();moiIt++)
      objCorresp[moiIt->first]=firstNodeB+moiIt->second;
    for (moiIt=reorderedElems.begin();moiIt!=reorderedElems.end();moiIt++) {
      objCorresp[moiIt->first]=firstNodeB+nodesNbr+edgesNbr+moiIt->second;
    }
    
    object *currentObj,*origObj;
    for (i=0;i<nodesNbr;i++) {
      currentObj=objCorresp[firstNode+i];
      nodeIndex=currentObj-firstNodeB;
      coord[nodeIndex][0]=arg.coord[i][0];
      coord[nodeIndex][1]=arg.coord[i][1];
      coord[nodeIndex][2]=arg.coord[i][2];
    }
    
    for (i=0;i<objectsNbr;i++) {
      origObj=firstNode+i;
      currentObj=objCorresp[origObj];
      currentObj->setType(origObj->getType());
      currentObj->setMinusNbr(origObj->getMinusNbr());
      for (j=0;j<currentObj->getMinusNbr();j++)
	currentObj->setMinus(j,objCorresp[origObj->getMinus(j)],
			     origObj->getMinusDirect(j));
      currentObj->setPlusNbr(origObj->getPlusNbr());
      for (j=0;j<currentObj->getPlusNbr();j++)
	currentObj->setPlus(j,objCorresp[origObj->getPlus(j)]);
    }
    
    domain *currentDom,*origDom;
    setDomainsNbr(arg.domainsNbr);
    for (i=0;i<domainsNbr;i++) {
      origDom=arg.domainsLst+i;
      currentDom=domainsLst+i;
      domainsIntId[i]=arg.domainsIntId[i];
      
      currentDom->setBaseMesh(this);
      currentDom->setObjectsNbr(origDom->getObjectsNbr());
      for (j=0;j<currentDom->getObjectsNbr();j++)
	currentDom->setObject(j,objCorresp[origDom->getObject(j)],
			      origDom->getObjectDirect(j));
      currentDom->setAllObjectsNbr(origDom->getAllObjectsNbr());
      for (j=0;j<currentDom->getAllObjectsNbr();j++)
	currentDom->setAllObject(j,objCorresp[origDom->getObject(j)]);
      currentDom->setDim();
    }
    
    // freeing allocated space of static containers :
    elemVect.resize(0);
    elemsAppear.clear();
    elemsDisappear.clear();
    nodeVect.resize(0);
    nodeMap.clear();
    nodesAppear.clear();
    nodesDisappear.clear();
    
    delete[] elemsNodes;
    delete[] nodesElems;
  }
  catchAll("void mesh::renumberDefDomainFrontal(const mesh &arg,domain &dom)","")
}



static map <object*,Double> objsX;

class compare_objsX {
  public :
    bool operator()(object *obj1,object *obj2) {
    
    if (obj1->getDim()<obj2->getDim())
      return(true);
    if (obj1->getDim()>obj2->getDim())
      return(false);
    
    return(objsX[obj1]<objsX[obj2]);
  }
} ;
        
void mesh::renumberDefDomainArray(const mesh &arg,array &A) {
  try {
    Double localCoord[3];
    Int i,j;
    object *currentObj,*firstNode,*firstNodeB;
    Int objectType,nodeIndex;
    scal *currentScal;
    vector<object*> objsVect;
    
    objsX.clear();
    
    for (i=0;i<arg.objectsNbr;i++) {
      currentObj=arg.objectsLst+i;
      objectType=currentObj->getType();
      switch (objectType) {
      case 1 :
	break;
      case 2 :
	localCoord[0]=0.0;
	break;
      case 3 :
	localCoord[0]=1.0/3.0;
	localCoord[1]=1.0/3.0;
	break;
      case 4 :
	localCoord[0]=0.0;
	localCoord[1]=0.0;
	break;
      }
      A.setNewPoint(true);
      A.makeAtPoint(currentObj,objectType,localCoord,true,false);
      currentScal=A.getComponentValues();
      objsX[currentObj]=currentScal->getVal();
      objsVect.push_back(currentObj);
    }
    
    sort(objsVect.begin(),objsVect.end(),compare_objsX());
    
    // One fills member data of the new mesh :
    
    type=arg.type;
    
    nodesNbr=arg.nodesNbr;
    if (nodesNbr!=0)
      coord=new Double[nodesNbr][3];
    else
      coord=NULL;
    
    firstNode=arg.objectsLst;
    firstNodeB=setObjectsNbr(arg.objectsNbr);
    map<object*,object*> objCorresp;
    vector<object*>::iterator vOIt;
    for (i=0,vOIt=objsVect.begin();i<objectsNbr;i++,vOIt++)
      objCorresp[*vOIt]=firstNodeB+i;
    
    object *origObj;
    for (i=0;i<nodesNbr;i++) {
      currentObj=objCorresp[firstNode+i];
      nodeIndex=currentObj-firstNodeB;
      coord[nodeIndex][0]=arg.coord[i][0];
      coord[nodeIndex][1]=arg.coord[i][1];
      coord[nodeIndex][2]=arg.coord[i][2];
    }
    
    for (i=0;i<objectsNbr;i++) {
      origObj=firstNode+i;
      currentObj=objCorresp[origObj];
      currentObj->setType(origObj->getType());
      currentObj->setMinusNbr(origObj->getMinusNbr());
      for (j=0;j<currentObj->getMinusNbr();j++)
	currentObj->setMinus(j,objCorresp[origObj->getMinus(j)],
			     origObj->getMinusDirect(j));
      currentObj->setPlusNbr(origObj->getPlusNbr());
      for (j=0;j<currentObj->getPlusNbr();j++)
	currentObj->setPlus(j,objCorresp[origObj->getPlus(j)]);
    }
    
    domain *currentDom,*origDom;
    setDomainsNbr(arg.domainsNbr);
    for (i=0;i<domainsNbr;i++) {
      origDom=arg.domainsLst+i;
      currentDom=domainsLst+i;
      domainsIntId[i]=arg.domainsIntId[i];
      
      currentDom->setBaseMesh(this);
      currentDom->setObjectsNbr(origDom->getObjectsNbr());
      for (j=0;j<currentDom->getObjectsNbr();j++)
	currentDom->setObject(j,objCorresp[origDom->getObject(j)],
			      origDom->getObjectDirect(j));
      currentDom->setAllObjectsNbr(origDom->getAllObjectsNbr());
      for (j=0;j<currentDom->getAllObjectsNbr();j++)
	currentDom->setAllObject(j,objCorresp[origDom->getObject(j)]);
      currentDom->setDim();
    }
    
    // freeing allocated space of static containers :
    
    objsX.clear();
  }
  catchAll("void mesh::renumberDefDomainArray(const mesh &arg,array &A)","")
}



