
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <map>
#include <set>
#include <vector>

#include "mesh_base.h"
#include "meshTypes.h"
#include "domain/domain_base.h"
#include "err/err.h"
#include "object/object_base.h"
#include "util/utils.h"

using namespace std ;
using namespace keyFE2 ;



inline void swap(Int &a,Int &b) {
  Int c=a;
  a=b;
  b=c;
}



inline void swap(Double &a,Double &b) {
  Double c=a;
  a=b;
  b=c;
}



void mesh::read_emc2Mesh(istream &inFile) {
  try {
    Int i,k,tmpInt;
    string line;
    char buf[100];
    bool endOfFile;
    
    Int dim,nbrNodes,nbrElements,nbrTriangles,nbrQuadrangles,nbrAretes;
    Int **aretes=NULL,*aretesBoundary=NULL;
    Int **triangles=NULL,*trianglesDomain=NULL;
    Int **quadrangles=NULL,*quadranglesDomain=NULL;
    
    // Reading beginning of mesh file :
    
    type=emc2Mesh;
    
    inFile.getline(buf,99);
    if (!inFile)
      errThrow("mesh::read_emc2Mesh","error in reading first line.");
    line=buf;
    if (line.find("MeshVersionFormatted")==line.npos)
      errThrow("mesh::read_emc2Mesh","unexpected line :  \""+line+"\".");
    
    inFile.getline(buf,99);
    if (!inFile)
      errThrow("mesh::read_emc2Mesh","error in reading second line.");
    line=buf;
    if (line.find("Dimension")==line.npos)
      errThrow("mesh::read_emc2Mesh","unexpected line :  \""+line+"\".");
    inFile >> dim ;
    if (!inFile)
      errThrow("mesh::read_emc2Mesh","error in reading third line.");
    if (dim!=2) 
      errThrow("mesh::read_emc2Mesh","invalid dimension for mesh.");
    inFile.ignore(100,'\n');
    
    nbrNodes=0;
    nbrAretes=0;
    nbrTriangles=0;
    nbrQuadrangles=0;
    nbrElements=0;
    
    // Reading the mesh in input file :
    
    endOfFile=inFile.eof();
    
    while (!endOfFile) {
      inFile.getline(buf,99);
      line=buf;
      endOfFile=inFile.eof();
      
      if (line.find("Vertices")!=line.npos) {
	if (nbrNodes>0)
	  errThrow("mesh::read_emc2Mesh","nbrNodes>0");      
	inFile >> nbrNodes ;
	if (!inFile)
	  errThrow("mesh::read_emc2Mesh","error in reading number of vertices.");
	inFile.ignore(100,'\n');
	nodesNbr=nbrNodes;
	if (nbrNodes!=0)
	  coord=new Double[nbrNodes][3];
	else
	  coord=NULL;
	for (i=0;i<nbrNodes;i++) {
	  inFile >> coord[i][0] >> coord[i][1] >> tmpInt ;
	  if (!inFile)
	    errThrow("mesh::read_emc2Mesh",
		     string("error in reading node ")+strFromInt(i)+".");
	  coord[i][2]=0.0;
	  inFile.ignore(100,'\n');
	}   
      }
      else if (line.find("Edges")!=line.npos) {
	if (nbrAretes>0)
	  errThrow("mesh::read_emc2Mesh","nbrAretes>0");
	inFile >> nbrAretes ;
	if (!inFile)
	  errThrow("mesh::read_emc2Mesh","error in reading number of edges.");
	inFile.ignore(100,'\n');
	{
	  Int *tmp=new Int[2*nbrAretes];
	  aretes=new Int* [nbrAretes];
	  for (i=0;i<nbrAretes;i++)
	    aretes[i]=tmp+2*i;
	  aretesBoundary=new Int[nbrAretes];
	}
	Int aretesCounter=0;
	Int i1,i2,i3;
	for (i=0;i<nbrAretes;i++) {
	  inFile >> i1 >> i2 >> i3 ;
	  if (!inFile)
	    errThrow("mesh::read_emc2Mesh",
		     string("error in reading boundary edge ")+strFromInt(i)+".");
	  inFile.ignore(100,'\n');
	  if (i3!=0) {
	    // On ne tient pas compte des aretes auxquelles aucun numro de 
	    // frontire n'a t assign :
	    aretes[aretesCounter][0]=i1-1; // la numrotation commence  zro
	    aretes[aretesCounter][1]=i2-1; // la numrotation commence  zro
	    aretesBoundary[aretesCounter]=i3;
	    aretesCounter++;
	  }
	}
	nbrAretes=aretesCounter;
      }
      else if (line.find("Triangles")!=line.npos) {
	if (nbrTriangles>0)
	  errThrow("mesh::read_emc2Mesh","nbrTriangles>0");
	inFile >> nbrTriangles ;
	if (!inFile)
	  errThrow("mesh::read_emc2Mesh","error in reading number of triangles.");
	inFile.ignore(100,'\n');
	{
	  Int *tmp=new Int[3*nbrTriangles];
	  triangles=new Int* [nbrTriangles];
	  for (i=0;i<nbrTriangles;i++)
	    triangles[i]=tmp+3*i;
	  trianglesDomain=new Int[nbrTriangles];
	}
	for (i=0;i<nbrTriangles;i++) {
	  inFile >> triangles[i][0] >> triangles[i][1] >> triangles[i][2] 
	    >> trianglesDomain[i] ;
	  if (!inFile)
	    errThrow("mesh::read_emc2Mesh",
		     string("error in reading triangle ")+strFromInt(i)+".");
	  inFile.ignore(100,'\n');
	}
      }
      else if (line.find("Quadrangles")!=line.npos) {
	if (nbrQuadrangles>0)
	  errThrow("mesh::read_emc2Mesh","nbrQuadrangles>0");
	inFile >> nbrQuadrangles ;
	if (!inFile)
	  errThrow("mesh::read_emc2Mesh","error in reading number of quadrangles.");
	inFile.ignore(100,'\n');
	{
	  Int *tmp=new Int[4*nbrQuadrangles];
	  quadrangles=new Int* [nbrQuadrangles];
	  for (i=0;i<nbrQuadrangles;i++)
	    quadrangles[i]=tmp+4*i;
	  quadranglesDomain=new Int[nbrQuadrangles];
	}
	for (i=0;i<nbrQuadrangles;i++) {
	  inFile >> quadrangles[i][0] >> quadrangles[i][1] >> quadrangles[i][2] 
	    >> quadrangles[i][3] >> quadranglesDomain[i] ;
	  if (!inFile)
	    errThrow("mesh::read_emc2Mesh",
		     string("error in reading quadrangle ")+strFromInt(i)+".");
	  inFile.ignore(100,'\n');
	}
      }
      else if (line.find("End")!=line.npos)
	endOfFile=true;
      else
	errThrow("mesh::read_emc2Mesh","unexpected line :  "+line);
    } // while(!endOfFile)
    
    // Here, the mesh file has been read.
    
    nbrElements=nbrTriangles+nbrQuadrangles;
    Int** elements=NULL;
    Int* elementsType=NULL;
    Int* elementsDomain=NULL;
    {
       elementsType=new Int[nbrElements];
       elementsDomain=new Int[nbrElements];
       Int* tmpInts=new Int[4*nbrElements];
       elements=new Int* [nbrElements];
       for (Int i=0;i<nbrElements;i++)
          elements[i]=tmpInts+4*i;
    }
    for (i=0;i<nbrTriangles;i++) {
      elements[i][0]=triangles[i][0]-1; // la numrotation commence  zro
      elements[i][1]=triangles[i][1]-1; // la numrotation commence  zro
      elements[i][2]=triangles[i][2]-1; // la numrotation commence  zro
      elements[i][3]=-1;
      elementsType[i]=3;
      elementsDomain[i]=trianglesDomain[i];
    }
    for (i=0;i<nbrQuadrangles;i++) {
      elements[nbrTriangles+i][0]=quadrangles[i][0]-1; // la numrotation commence  zro
      elements[nbrTriangles+i][1]=quadrangles[i][1]-1; // la numrotation commence  zro
      elements[nbrTriangles+i][2]=quadrangles[i][2]-1; // la numrotation commence  zro
      elements[nbrTriangles+i][3]=quadrangles[i][3]-1; // la numrotation commence  zro
      elementsType[nbrTriangles+i]=4;
      elementsDomain[nbrTriangles+i]=quadranglesDomain[i];
    }
    
    // A list of edges is created :
    
    set<Int>* edges=NULL;
    edges=new set<Int> [nbrNodes];
    for (Int i=0;i<nbrElements;i++) {
      for (Int j=0;j<elementsType[i];j++) {
	edges[elements[i][j]].insert(elements[i][(j+1)%elementsType[i]]);
	edges[elements[i][(j+1)%elementsType[i]]].insert(elements[i][j]);
      }
    }
    Int nbrEdges=0;
    for (Int i=0;i<nbrNodes;i++) {
      set<Int>::iterator j;
      for(j=edges[i].begin();j!=edges[i].end();j++)
	if (*j>i)
	  nbrEdges++;
    }
    
    // The objects of the domain are created :
    
    objectsNbr=nbrNodes+nbrEdges+nbrElements;
    if (objectsNbr!=0)
      objectsLst=new object[objectsNbr];
    else
      objectsLst=NULL;
    
    object* nodes=objectsLst;
    for (Int i=0;i<nbrNodes;i++)
      nodes[i].setType("node");
    
    object* segment=objectsLst+nbrNodes;
    for (Int i=0;i<nbrEdges;i++) {
      segment[i].setType("segment");
      segment[i].setMinusNbr(2);
    }
    
    object* surfElt=objectsLst+nbrNodes+nbrEdges;
    for (Int i=0;i<nbrElements;i++) {
      if (elementsType[i]==3) 
	surfElt[i].setType("triangle");
      else if (elementsType[i]==4) 
	surfElt[i].setType("quadrangle");
      surfElt[i].setMinusNbr(elementsType[i]);
    }
    
    // The segments are down connected to the nodes :
    
    map<Int,object*>* edgeIndexPointer=NULL;
    edgeIndexPointer=new map<Int,object*>[nbrNodes];
    Int edgeIndex=0;
    for (Int i=0;i<nbrNodes;i++) {
      set<Int>::iterator j;
      for (j=edges[i].begin();j!=edges[i].end();j++) {
	if (*j>i) {
	  segment[edgeIndex].setMinus(0,nodes+i,0);
	  segment[edgeIndex].setMinus(1,nodes+(*j),0);
	  edgeIndexPointer[i][*j]=segment+edgeIndex;
	  edgeIndexPointer[*j][i]=segment+edgeIndex;
	  ++edgeIndex;
	}
      }
    }
    
    // The nodes are up connected to the segments :
    
    {
      Int* nodeIndex=NULL;
      nodeIndex=new Int[nbrNodes];
      for (Int i=0;i<nbrNodes;i++) {
	nodeIndex[i]=0;
	nodes[i].setPlusNbr(edges[i].size());
      }
      Int edgeIndex=0;
      for (Int i=0;i<nbrNodes;i++) {
	set<Int>::iterator j;
	for(j=edges[i].begin();j!=edges[i].end();j++) {
	  if (*j>i) {
	    nodes[i].setPlus(nodeIndex[i],segment+edgeIndex);
	    nodeIndex[i]++;
	    nodes[*j].setPlus(nodeIndex[*j],segment+edgeIndex);
	    nodeIndex[*j]++;
	    ++edgeIndex;
	  }
	}
      }
      delete[] nodeIndex;
    }
    
    // The triangles and quadrangles are created and down connected to the 
    // segments (and the segments are up connected to triangles and quadrangles) :
    
    {
      Int node1,node2;
      Int* nbrUp=NULL;
      nbrUp=new Int[nbrEdges];
      for (Int i=0;i<nbrEdges;i++)
	nbrUp[i]=0;
      for (Int i=0;i<nbrElements;i++) {
	surfElt[i].setMinusNbr(elementsType[i]);
	for (Int j=0;j<elementsType[i];j++) {
	  node1=elements[i][j];
	  node2=elements[i][(j+1)%elementsType[i]];
	  if (node2>node1)
	    surfElt[i].setMinus(j,edgeIndexPointer[node1][node2],2);
	  else
	    surfElt[i].setMinus(j,edgeIndexPointer[node2][node1],-2);
	  nbrUp[(edgeIndexPointer[node1][node2]-segment)]++;
	}
      }
      for (Int i=0;i<nbrEdges;i++)
	segment[i].setPlusNbr(nbrUp[i]);
      for (Int i=0;i<nbrEdges;i++)
	nbrUp[i]=0;
      for (Int i=0;i<nbrElements;i++) {
	for (Int j=0;j<elementsType[i];j++) {
	  Int edgeIndex=surfElt[i].getMinus(j)-segment;
	  surfElt[i].getMinus(j)->setPlus(nbrUp[edgeIndex],surfElt+i);
	  nbrUp[edgeIndex]++;
	}
      }
      delete[] nbrUp;
    }
    
    // Definition of subdomains and boundaries :      
    
    {
      set<Int> subDomains,boundaries;
      set<Int>::iterator j;
      for (Int i=0;i<nbrElements;i++)
	subDomains.insert(elementsDomain[i]);
      Int nbrSubDomains=subDomains.size();
      for (Int i=0;i<nbrAretes;i++)
	boundaries.insert(aretesBoundary[i]);
      Int nbrBoundaries=boundaries.size();
      
      setDomainsNbr(nbrSubDomains+nbrBoundaries);
      
      Int domainIndex=0;
      for (j=subDomains.begin();j!=subDomains.end();j++) {
	Int nbrES=0;
	for (Int i=0;i<nbrElements;i++)
	  if (elementsDomain[i]==(*j))
	    nbrES++;
	domainsLst[domainIndex].setObjectsNbr(nbrES);
	Int indexES=0;
	for (Int i=0;i<nbrElements;i++) {
	  if (elementsDomain[i]==(*j)) {
	    // One assumes that the orientation definition is conserved
	    // in the domain description.
	    domainsLst[domainIndex].setObject(indexES,surfElt+i,
					      surfElt[i].getMinusNbr());
	    indexES++;
	  }
	}
	domainsLst[domainIndex].setDim();
	domainsIntId[domainIndex]=(*j);
	domainIndex++;
      }
      
      Int boundaryIndex=0;
      Int indexAB;
      for (j=boundaries.begin();j!=boundaries.end();j++) {
	Int nbrAB=0;
	domain tmp;
	for (Int i=0;i<nbrAretes;i++)
	  if (aretesBoundary[i]==(*j))
	    nbrAB++;
	domainsLst[nbrSubDomains+boundaryIndex].setObjectsNbr(nbrAB);
	object *theEdge,*theMaille;
	Int direct;
	indexAB=0;
	for (Int i=0;i<nbrAretes;i++) {
	  if (aretesBoundary[i]==(*j)) {
	    theEdge=edgeIndexPointer[aretes[i][0]][aretes[i][1]];
	    if (theEdge->plusNbr==1) {
	      // The edge belongs to a boundary.
	      direct=0;
	      theMaille=theEdge->plusLst[0];
	      for (k=0;k<theMaille->minusNbr;k++)
		if (theMaille->minusLst[k]==theEdge)
		  direct=theMaille->minusDirect[k];
	      if (direct==0)
		errThrow("mesh::read_emc2Mesh",
			 string("   invalid edge for boundary ")
			 +strFromInt(aretes[i][0])
			 +" "+strFromInt(aretes[i][1])+".");
	    }
	    else
	      direct=2;
	    domainsLst[nbrSubDomains+boundaryIndex]
	      .setObject(indexAB,theEdge,direct);
	    indexAB++;
	  }
	}
	domainsLst[nbrSubDomains+boundaryIndex].setDim();
	domainsIntId[nbrSubDomains+boundaryIndex]=(*j);
	tmp.reorder1DDomain(domainsLst[nbrSubDomains+boundaryIndex]);
	domainsLst[nbrSubDomains+boundaryIndex]=tmp;
	boundaryIndex++;
      }
    }
    
    // cleaning allocated data :
    
    if (nbrAretes!=0) {
      delete[] aretes[0];
      delete[] aretes;
      delete[] aretesBoundary;
    }
    if (nbrTriangles!=0) {
      delete[] triangles[0];
      delete[] triangles;
      delete[] trianglesDomain;
    }
    if (nbrQuadrangles!=0) {
      delete[] quadrangles[0];
      delete[] quadrangles;
      delete[] quadranglesDomain;
    }
    if (nbrElements!=0) {
      delete[] elements[0];
      delete[] elements;
      delete[] elementsType;
      delete[] elementsDomain;
    }
    delete[] edges;
    delete[] edgeIndexPointer;
  }
  catchAll("void mesh::read_emc2Mesh(istream &inFile)","")
}



void mesh::write_emc2Mesh(ostream &outFile) const {
  try {
    map<object*,Int> elemsDoms;
    domain *currentDom;
    Int objsNbr,domainId,segmentsNbr=0,elementsNbr=0,trianglesNbr=0,quadranglesNbr=0;
    object *obj;
    for (Int i=0;i<domainsNbr;i++) {
      currentDom=domainsLst+i;
      domainId=domainsIntId[i];
      objsNbr=currentDom->getObjectsNbr();
      for (Int j=0;j<domainsLst[i].getObjectsNbr();j++) {
	obj=currentDom->getObject(j);
	if (obj->getType()==3) {
	  elementsNbr++;
	  trianglesNbr++;
	  elemsDoms[obj]=domainId;
	}
	if (obj->getType()==4) {
	  elementsNbr++;
	  quadranglesNbr++;
	  elemsDoms[obj]=domainId;
	}
	else if (obj->getType()==2)
	  segmentsNbr++;
      }
    }
    
    // Writing of preliminary data :
    
    outFile << " MeshVersionFormatted 0" << endl ;
    if (!outFile)
      errThrow("mesh::write_emc2Mesh","error in writing first line.");
    outFile << " Dimension" << endl ;
    if (!outFile)
      errThrow("mesh::write_emc2Mesh","error in writing second line.");
    outFile << " 2" << endl ;
    if (!outFile)
      errThrow("mesh::write_emc2Mesh","error in writing third line.");
    
    // Writing of nodes :
    
    outFile << " Vertices" << endl ;
    if (!outFile)
      errThrow("mesh::write_emc2Mesh","error in writing nodes.");
    outFile << " " << nodesNbr << endl ;
    if (!outFile)
      errThrow("mesh::write_emc2Mesh","error in writing nodes.");
    outFile.setf(ios::scientific);
    outFile << setprecision(7);
    for (Int i=0;i<nodesNbr;i++) {
      for (Int j=0;j<2;j++) {
	outFile << setw(16) << coord[i][j] ;
	if (!outFile)
	  errThrow("mesh::write_emc2Mesh",
		   string("error in writing node ")+strFromInt(i+1)+".");
      }
      outFile << " " << 0 << endl ;
      if (!outFile)
	errThrow("mesh::write_emc2Mesh",
		 string("error in writing node ")+strFromInt(i+1)+".");
    }
    
    object* objList[50];
    Int objDirect[50],objNbr;
    Int elemsCounter;
    
    // Writing of edges :
    
    outFile << " Edges" << endl ;
    if (!outFile)
      errThrow("mesh::write_emc2Mesh","error in writing edges.");
    outFile << " " << segmentsNbr << endl ;
    if (!outFile)
      errThrow("mesh::write_emc2Mesh","error in writing edges.");
    for (Int i=0;i<domainsNbr;i++) {
      currentDom=domainsLst+i;
      domainId=domainsIntId[i];
      objsNbr=currentDom->getObjectsNbr();
      for (Int j=0;j<domainsLst[i].getObjectsNbr();j++) {
	obj=currentDom->getObject(j);
	if (obj->getType()==2) {
	  objNbr=obj->split(currentDom->getObjectDirect(j),objList,objDirect);
	  for (Int k=0;k<objNbr;k++) {
	    if (objList[k]->getType()==1) {
	      outFile << " " << objList[k]-objectsLst+1 ;
	      if (!outFile)
		errThrow("mesh::write_emc2Mesh",
			 string("error in writing domain ")
			 +strFromInt(i+1)+".");
	    }
	  }
	  outFile << " " << domainId << endl ;
	  if (!outFile)
	    errThrow("mesh::write_emc2Mesh",
		     string("error in writing domain ")+strFromInt(i+1)+".");
	}
      }
    }
    
    // Writing of triangles :
    
    outFile << " Triangles" << endl ;
    if (!outFile)
      errThrow("mesh::write_emc2Mesh","error in writing triangles.");
    outFile << " " << trianglesNbr << endl ;
    if (!outFile)
      errThrow("mesh::write_emc2Mesh","error in writing triangles.");
    elemsCounter=0;
    for (Int i=0;i<objectsNbr;i++) {
      obj=objectsLst+i;
      if (obj->getType()==3) {
	elemsCounter++;
	objNbr=obj->split(obj->getType(),objList,objDirect);
	for (Int j=0;j<objNbr;j++) {
	  if (objList[j]->getType()==1) {
	    outFile << " " << objList[j]-objectsLst+1 ;
	    if (!outFile)
	      errThrow("mesh::write_emc2Mesh",
		       string("error in writing triangle ")
		       +strFromInt(elemsCounter)+".");
	  }
	}
	if (elemsDoms.find(obj)!=elemsDoms.end()) {
	  outFile << " " << elemsDoms[obj] ;
	  if (!outFile)
	    errThrow("mesh::write_emc2Mesh",
		     string("error in writing triangle ")
		     +strFromInt(elemsCounter)+".");
	}
	else {
	  outFile << " 0" ;
	  if (!outFile)
	    errThrow("mesh::write_emc2Mesh",
		     string("error in writing triangle ")
		     +strFromInt(elemsCounter)+".");
	}
	outFile << endl ;
	if (!outFile)
	  errThrow("mesh::write_emc2Mesh",
		   string("error in writing triangle ")+strFromInt(elemsCounter)+".");
      }
    }
    if (elemsCounter!=trianglesNbr) 
      errThrow("mesh::write_emc2Mesh","triangle counters mismatch");
    
    // Writing of quadrangles :
    
    outFile << " Quadrangles" << endl ;
    if (!outFile)
      errThrow("mesh::write_emc2Mesh","error in writing quadrangles.");
    outFile << " " << quadranglesNbr << endl ;
    if (!outFile)
      errThrow("mesh::write_emc2Mesh","error in writing quadrangles.");
    elemsCounter=0;
    for (Int i=0;i<objectsNbr;i++) {
      obj=objectsLst+i;
      if (obj->getType()==4) {
	elemsCounter++;
	objNbr=obj->split(obj->getType(),objList,objDirect);
	for (Int j=0;j<objNbr;j++) {
	  if (objList[j]->getType()==1) {
	    outFile << " " << objList[j]-objectsLst+1 ;
	    if (!outFile)
	      errThrow("mesh::write_emc2Mesh",
		       string("error in writing quadrangle ")
		       +strFromInt(elemsCounter)+".");
	  }
	}
	if (elemsDoms.find(obj)!=elemsDoms.end()) {
	  outFile << " " << elemsDoms[obj] ;
	  if (!outFile)
	    errThrow("mesh::write_emc2Mesh",
		     string("error in writing quadrangle ")
		     +strFromInt(elemsCounter)+".");
	}
	else {
	  outFile << " 0" ;
	  if (!outFile)
	    errThrow("mesh::write_emc2Mesh",
		     string("error in writing quadrangle ")
		     +strFromInt(elemsCounter)+".");
	}
	outFile << endl ;
	if (!outFile)
	  errThrow("mesh::write_emc2Mesh",
		   string("error in writing quadrangle ")+strFromInt(elemsCounter)+".");
	elemsCounter++;
      }
    }
    
    outFile << " End" ;
    if (!outFile)
      errThrow("mesh::write_emc2Mesh",
	       string("error in writing \"End\"."));
    elemsCounter++;
  }
  catchAll("void mesh::write_emc2Mesh(ostream &outFile) const","")
}



