
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <fstream>
#include <iomanip>
#include <list>
#include <map>
#include <set>
#include <vector>

#include "mesh_base.h"
#include "meshTypes.h"
#include "domain/domain_base.h"
#include "err/err.h"
#include "object/object_base.h"
#include "util/utils.h"

using namespace std ;
using namespace keyFE2 ;



void mesh::read_gmsh2dMesh(istream &inFile) {
  try {
    Int i,j,index,nbrElm,elmType,direct,nodeIndex,pos;
    string line;
    char buf[100];
    intList tmpIntList,tmpIntList2;
    list<intList> lines;
    list<intList>::iterator ilIt1;
    object *currentObj;
    
    // Type of the mesh:
  
    type=gmsh2dMesh;
    
    // Reading the nodes:
    
    map<Int,Int> nodesMap;
    map<Int,Int> nodeIndices;
    
    inFile.getline(buf,99);
    if (!inFile)
      errThrow("mesh::read_gmsh2dMesh","error in reading first line.");
    line=buf;
    if (line.find("$NOD")==line.npos)
      errThrow("mesh::read_gmsh2dMesh","unexpected line :  \""+line+"\" ($NOD).");
      
    if (nodesNbr>0)
      errThrow("mesh::read_gmsh2dMesh","nodesNbr>0");	 
    inFile >> nodesNbr ;
    if (!inFile)
      errThrow("mesh::read_gmsh2dMesh","error in reading number of nodes.");
    inFile.ignore(100,'\n');
    if (nodesNbr!=0)
      coord=new Double[nodesNbr][3];
    else
      coord=NULL;
    for (i=0;i<nodesNbr;i++) {
      inFile >> nodeIndex >> coord[i][0] >> coord[i][1] >> coord[i][2] ;
      if (!inFile)
    	errThrow("mesh::read_gmsh2dMesh",
    		 string("error in reading node ")+strFromInt(i)+".");
      inFile.ignore(100,'\n');
      nodesMap[nodeIndex]=i;
      nodeIndices[i]=nodeIndex;
    }	
    
    inFile.getline(buf,99);
    if (!inFile)
      errThrow("mesh::read_gmsh2dMesh","error in reading first line.");
    line=buf;
    if (line.find("$ENDNOD")==line.npos)
      errThrow("mesh::read_gmsh2dMesh","unexpected line :  \""+line+"\" ($ENDNOD).");
      
    // Reading the lines corresponding to elements:
    
    inFile.getline(buf,99);
    if (!inFile)
      errThrow("mesh::read_gmsh2dMesh","error in reading line.");
    line=buf;
    if (line.find("$ELM")==line.npos)
      errThrow("mesh::read_gmsh2dMesh","unexpected line :  \""+line+"\" ($ELM).");
      
    bool physicalDefined=false;
    map <int,int> reg_phys,reg_elem; // objects counter for regions.
    inFile >> nbrElm ;
    if (!inFile)
      errThrow("mesh::read_gmsh2dMesh","error in reading number of ELMs.");
    inFile.ignore(100,'\n');
    for (i=0;i<nbrElm;i++) {
      inFile >> tmpIntList.ilist[0]
             >> tmpIntList.ilist[1]
             >> tmpIntList.ilist[2]
             >> tmpIntList.ilist[3]
             >> tmpIntList.ilist[4] ;
      tmpIntList.setSize(5+tmpIntList.ilist[4]);
      for (j=0;j<tmpIntList.ilist[4];j++)
        inFile >> tmpIntList.ilist[5+j] ;
      if (!inFile)
    	errThrow("mesh::read_gmsh2dMesh",
    		 string("error in reading ELM at pos ")+strFromInt(i)+".");
      inFile.ignore(100,'\n');
      lines.push_back(tmpIntList);
      if (tmpIntList.ilist[2]!=tmpIntList.ilist[3])
        physicalDefined=true;
      reg_phys[tmpIntList.ilist[2]]++;
      reg_elem[tmpIntList.ilist[3]]++;
    }	
   
    inFile.getline(buf,99);
    if (!inFile)
      errThrow("mesh::read_gmsh2dMesh","error in reading line.");
    line=buf;
    if (line.find("$ENDELM")==line.npos)
      errThrow("mesh::read_gmsh2dMesh","unexpected line :  \""+line+"\" ($ENDELM).");
      
    // Here, the file has been read!
    
    // One builds a list of all the edges that have been read :
    
    Int nodeId1,nodeId2,nbrEdges,nbrElm2d;
    list<intList> edges,elm2d;
    tmpIntList.setSize(2);
    for (ilIt1=lines.begin();ilIt1!=lines.end();ilIt1++) {
       elmType=ilIt1->ilist[1];
       switch (elmType) {
          case 1 : 
	     // Segment:
	     nodeId1=ilIt1->ilist[5];
	     nodeId2=ilIt1->ilist[6];
	     if (nodeId1<nodeId2) {
	        tmpIntList.ilist[0]=nodeId1;
	        tmpIntList.ilist[1]=nodeId2;
	     }
	     else {
	        tmpIntList.ilist[0]=nodeId2;
	        tmpIntList.ilist[1]=nodeId1;
	     }
	     edges.push_back(tmpIntList);
	     break;
          case 2 : 
	     // Triangle:
          case 3 : 
	     // Quadrangle:
	     nbrEdges=ilIt1->ilist[4];
	     tmpIntList2.setSize(nbrEdges);
	     for (i=0;i<nbrEdges;i++) {
	        nodeId1=ilIt1->ilist[5+i];
	        nodeId2=ilIt1->ilist[5+(i+1)%nbrEdges];
	     	if (nodeId1<nodeId2) {
	     	   tmpIntList.ilist[0]=nodeId1;
	     	   tmpIntList.ilist[1]=nodeId2;
	     	}
	     	else {
	     	   tmpIntList.ilist[0]=nodeId2;
	     	   tmpIntList.ilist[1]=nodeId1;
	     	}
	     	edges.push_back(tmpIntList);
		tmpIntList2.ilist[i]=nodeId1;
             }
	     elm2d.push_back(tmpIntList2);
	     break;
          case 15 :
	     // Point: nothing to do !
	     break;
	  default :
    	     errThrow("mesh::read_gmsh2dMesh",
    		      string("invalid element type: ")+strFromInt(elmType)+".");
	     break;
       } // switch ...
    } // for lines ...
    edges.sort();
    edges.unique();
    nbrEdges=edges.size();
    elm2d.sort();
    elm2d.unique();
    nbrElm2d=elm2d.size();
    
    // Creation of the objects:
    
    objectsNbr=nodesNbr+nbrEdges+nbrElm2d;
    if (objectsNbr!=0)
      objectsLst=new object[objectsNbr];
    else
      objectsLst=NULL;
    
    object* nodes=objectsLst;
    for (Int i=0;i<nodesNbr;i++)
      nodes[i].setType("node");
    
    object* segment=objectsLst+nodesNbr;
    map<intList,object*> elm1dMap; // Correspondence between 1D elements and objects.
    i=0;
    for (ilIt1=edges.begin();ilIt1!=edges.end();ilIt1++) {
      segment[i].setType("segment");
      segment[i].setMinusNbr(2);
      elm1dMap[*ilIt1]=segment+i;
      i++;
    }
    
    object* surfElt=objectsLst+nodesNbr+nbrEdges;
    map<intList,object*> elm2dMap; // Correspondence between 2D elements and objects.
    i=0;
    for (ilIt1=elm2d.begin();ilIt1!=elm2d.end();ilIt1++) {
      if (ilIt1->size==3) 
	surfElt[i].setType("triangle");
      else if (ilIt1->size==4) 
	surfElt[i].setType("quadrangle");
      surfElt[i].setMinusNbr(ilIt1->size);
      elm2dMap[*ilIt1]=surfElt+i;
      i++;
    }
    
    // The segments are down connected to the nodes:
    
    map<Int,Int> nodesUpCounter; // Number of segments connected to each node.
    for (i=0;i<nodesNbr;i++)
       nodesUpCounter[nodeIndices[i]]=0;
    index=0;
    for (ilIt1=edges.begin();ilIt1!=edges.end();ilIt1++) {
       nodeId1=ilIt1->ilist[0];
       nodeId2=ilIt1->ilist[1];
       nodesUpCounter[nodeId1]++;
       nodesUpCounter[nodeId2]++;
       segment[index].setMinus(0,nodes+nodesMap[nodeId1],0);
       segment[index].setMinus(1,nodes+nodesMap[nodeId2],0);       
       index++;
    }
    
    // The nodes are up connected to the segments:
    
    for (i=0;i<nodesNbr;i++)
       (nodes+nodesMap[nodeIndices[i]])
          ->setPlusNbr(nodesUpCounter[nodeIndices[i]]);
    for (i=0;i<nodesNbr;i++)
       nodesUpCounter[nodeIndices[i]]=0;
    index=0;
    for (ilIt1=edges.begin();ilIt1!=edges.end();ilIt1++) {
       nodeId1=ilIt1->ilist[0];
       nodeId2=ilIt1->ilist[1];
       currentObj=segment+index;
       (nodes+nodesMap[nodeId1])->setPlus(nodesUpCounter[nodeId1]++,currentObj);
       (nodes+nodesMap[nodeId2])->setPlus(nodesUpCounter[nodeId2]++,currentObj);
       index++;
    }
    
    // The triangles and quadrangles are down connected to segments:
    
    object *currentEdge;
    map<object*,Int> segmentsUpCounter; // Number of segments connected to each segment.
    for (i=0;i<nbrEdges;i++)
       segmentsUpCounter[segment+i]=0;
    tmpIntList.setSize(2);
    for (ilIt1=elm2d.begin();ilIt1!=elm2d.end();ilIt1++) {
       currentObj=elm2dMap[*ilIt1];
       for (i=0;i<ilIt1->size;i++) {
          nodeId1=ilIt1->ilist[i];
          nodeId2=ilIt1->ilist[(i+1)%ilIt1->size];
	  if (nodeId1<nodeId2) {
	     tmpIntList.ilist[0]=nodeId1;
	     tmpIntList.ilist[1]=nodeId2;
	     direct=2;
	  }
	  else {
	     tmpIntList.ilist[0]=nodeId2;
	     tmpIntList.ilist[1]=nodeId1;
	     direct=-2;
	  }
	  currentEdge=elm1dMap[tmpIntList];
	  currentObj->setMinus(i,currentEdge,direct);
	  segmentsUpCounter[currentEdge]++;
       }
    }
    
    // The segments are up-connected to 2D elements:
    
    for (i=0;i<nbrEdges;i++)
       (segment+i)->setPlusNbr(segmentsUpCounter[segment+i]);
    for (i=0;i<nbrEdges;i++)
       segmentsUpCounter[segment+i]=0;
    tmpIntList.setSize(2);
    for (ilIt1=elm2d.begin();ilIt1!=elm2d.end();ilIt1++) {
       currentObj=elm2dMap[*ilIt1];
       for (i=0;i<ilIt1->size;i++) {
          nodeId1=ilIt1->ilist[i];
          nodeId2=ilIt1->ilist[(i+1)%ilIt1->size];
	  if (nodeId1<nodeId2) {
	     tmpIntList.ilist[0]=nodeId1;
	     tmpIntList.ilist[1]=nodeId2;
	     direct=2;
	  }
	  else {
	     tmpIntList.ilist[0]=nodeId2;
	     tmpIntList.ilist[1]=nodeId1;
	     direct=-2;
	  }
	  currentEdge=elm1dMap[tmpIntList];
	  currentEdge->setPlus(segmentsUpCounter[currentEdge]++,currentObj);
       }
    }
    
    // Definition of domains:
    
    map<Int,Int>::iterator miiIt;
    map<Int,Int> regions;
    if (physicalDefined) {
       regions=reg_phys;
       pos=2;
    }
    else {
       regions=reg_elem;
       pos=3;
    }
    
    setDomainsNbr(regions.size());
    i=0;
    map<Int,domain*> mIntDom;
    for (miiIt=regions.begin();miiIt!=regions.end();miiIt++) {
       mIntDom[miiIt->first]=domainsLst+i;
       domainsIntId[i]=miiIt->first;
       domainsLst[i++].setObjectsNbr(miiIt->second);
       miiIt->second=0;
    }
    
    i=0;
    domain *dom;
    Int domId;
    for (ilIt1=lines.begin();ilIt1!=lines.end();ilIt1++) {
       domId=ilIt1->ilist[pos];
       dom=mIntDom[domId];
       elmType=ilIt1->ilist[1];
       switch (elmType) {
          case 1 : 
	     // Segment:
	     nodeId1=ilIt1->ilist[5];
	     nodeId2=ilIt1->ilist[6];
	     tmpIntList.setSize(2);
	     if (nodeId1<nodeId2) {
	     	tmpIntList.ilist[0]=nodeId1;
	     	tmpIntList.ilist[1]=nodeId2;
	        direct=2; 
	     }
	     else {
	     	tmpIntList.ilist[0]=nodeId2;
	     	tmpIntList.ilist[1]=nodeId1;
	        direct=-2;
	     }
             dom->setObject(regions[domId],elm1dMap[tmpIntList],direct);
	     regions[domId]++;
	     break;
          case 2 : 
	     // Triangle:
          case 3 : 
	     // Quadrangle:
	     nbrEdges=ilIt1->ilist[4];
	     tmpIntList.setSize(nbrEdges);
	     for (i=0;i<nbrEdges;i++)
		tmpIntList.ilist[i]=ilIt1->ilist[5+i];
             dom->setObject(regions[domId],elm2dMap[tmpIntList],nbrEdges);
	     regions[domId]++;
	     break;
          case 15 :
	     // Point:
	     nodeIndex=ilIt1->ilist[5];
             dom->setObject(regions[domId],nodes+nodesMap[nodeIndex],0);
	     regions[domId]++;
	     break;
	  default :
    	     errThrow("mesh::read_gmsh2dMesh",
    		      string("invalid element type: ")+strFromInt(elmType)+".");
	     break;
       } // switch ...
	  
	  
       
    }
    

  }
  catchAll("void mesh::read_gmsh2dMesh(istream &inFile)","")
}



void mesh::write_gmsh2dMesh(ostream &outFile) const {
  try {
        
    // Writing of nodes :
    
    outFile << "$NOD" << endl ;
    outFile << setprecision(16);
    if (!outFile)
      errThrow("mesh::write_gmsh2dMesh","error in writing $NOD.");
    outFile << nodesNbr << endl ;
    if (!outFile)
      errThrow("mesh::write_gmsh2dMesh","error in writing nodesNbr.");
    for (Int i=0;i<nodesNbr;i++) {
      outFile << i+1 ;
      for (Int j=0;j<3;j++) {
	outFile << " " << coord[i][j] ;
	if (!outFile)
	  errThrow("mesh::write_gmsh2dMesh",
		   string("error in writing node ")+strFromInt(i+1)+".");
      }
      outFile << endl ;
      if (!outFile)
	errThrow("mesh::write_gmsh2dMesh",
		 string("error in writing node ")+strFromInt(i+1)+".");
    }
    outFile << "$ENDNOD" << endl ;
    if (!outFile)
      errThrow("mesh::write_gmsh2dMesh","error in writing $ENDNOD.");
    
    object* objList[50];
    Int objDirect[50],objNbr,domainId,objsNbr;
    domain *currentDom;
    object *obj;
    
    
    // Writing of elements:
    
    Int nbrElements=0;
    for (Int i=0;i<domainsNbr;i++)
       nbrElements+=domainsLst[i].getObjectsNbr();
    outFile << "$ELM" << endl ;
    if (!outFile)
      errThrow("mesh::write_gmsh2dMesh","error in writing $ELM.");
    outFile << nbrElements << endl ;
    if (!outFile)
      errThrow("mesh::write_gmsh2dMesh","error in writing nbrElements.");
    
    Int elmNumber=1;
    for (Int i=0;i<domainsNbr;i++) {
      currentDom=domainsLst+i;
      domainId=domainsIntId[i];
      objsNbr=currentDom->getObjectsNbr();
      for (Int j=0;j<domainsLst[i].getObjectsNbr();j++) {
	obj=currentDom->getObject(j);
	outFile << elmNumber << " " ;
	switch (obj->getType()) {
	  case 1:
	    // Point:
	    outFile << 15 << " " << domainsIntId[i] << " " << domainsIntId[i] << " " 
	            << 1 << " " ;
	    outFile << obj-objectsLst+1 << endl ;
	    break;
	  case 2:
	    // Segment:
	    outFile << 1 << " " << domainsIntId[i] << " " << domainsIntId[i] << " " 
	            << 2 ;
	    objNbr=obj->split(currentDom->getObjectDirect(j),objList,objDirect);
	    for (Int k=0;k<objNbr;k++) {
	      if (objList[k]->getType()==1) {
	  	outFile << " " << objList[k]-objectsLst+1 ;
	  	if (!outFile)
	  	  errThrow("mesh::write_gmsh2dMesh",
	  		   string("error in writing domain ")
	  		   +strFromInt(i+1)+".");
	      }
	    }
	    outFile << endl ;
	    if (!outFile)
	      errThrow("mesh::write_gmsh2dMesh",
	  	       string("error in writing domain ")+strFromInt(i+1)+".");
	    break;
	  case 3:
	    outFile << 2 << " " << domainsIntId[i] << " " << domainsIntId[i] << " " 
	            << 3 ;
	    objNbr=obj->split(currentDom->getObjectDirect(j),objList,objDirect);
	    for (Int k=0;k<objNbr;k++) {
	      if (objList[k]->getType()==1) {
	  	outFile << " " << objList[k]-objectsLst+1 ;
	  	if (!outFile)
	  	  errThrow("mesh::write_gmsh2dMesh",
	  		   string("error in writing domain ")
	  		   +strFromInt(i+1)+".");
	      }
	    }
	    outFile << endl ;
	    if (!outFile)
	      errThrow("mesh::write_gmsh2dMesh",
	  	       string("error in writing domain ")+strFromInt(i+1)+".");
	    break;
	  case 4:
	    outFile << 3 << " " << domainsIntId[i] << " " << domainsIntId[i] << " " 
	            << 4 ;
	    objNbr=obj->split(currentDom->getObjectDirect(j),objList,objDirect);
	    for (Int k=0;k<objNbr;k++) {
	      if (objList[k]->getType()==1) {
	  	outFile << " " << objList[k]-objectsLst+1 ;
	  	if (!outFile)
	  	  errThrow("mesh::write_gmsh2dMesh",
	  		   string("error in writing domain ")
	  		   +strFromInt(i+1)+".");
	      }
	    }
	    outFile << endl ;
	    if (!outFile)
	      errThrow("mesh::write_gmsh2dMesh",
	  	       string("error in writing domain ")+strFromInt(i+1)+".");
	    break;
	  default:
    	    errThrow("mesh::read_gmsh2dMesh",
    	             string("invalid element type: ")+strFromInt(obj->getType())+".");
	    break;
	} // switch (obj->getType()) ...
	elmNumber++;
      } // for (j...
    } // for (i...
    
    outFile << "$ENDELM" << endl ;
    if (!outFile)
      errThrow("mesh::write_gmsh2dMesh","error in writing $ENDELM.");
  }
  catchAll("void mesh::write_gmsh2dMesh(ostream &outFile) const","")
}



