
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <map>
#include <set>
#include <vector>

#include "object_base.h"
#include "objectTypes.h"
#include "err/err.h"
#include "types/types.h"
#include "util/utils.h"

using namespace std ;
using namespace keyFE2 ;



object::object(void) {
  type=0;
  
  minusNbr=0;
  minusDirect=NULL;
  minusLst=NULL;
  
  plusNbr=0;
  plusLst=NULL;
}


object::object(string tp) {
  if (tp=="node") 
    type=node;
  else if (tp=="segment")
    type=segment;
  else if (tp=="triangle")
    type=triangle;
  else if (tp=="quadrangle")
    type=quadrangle;
  else
    errThrow("object constructor",string("unknown object type :  ")+tp+".");
  minusNbr=0;
  minusDirect=NULL;
  minusLst=NULL;
  plusNbr=0;
  plusLst=NULL;
}



object::~object(void) {
  try {
    if (minusDirect!=NULL) {
      delete[] minusDirect;
      minusDirect=NULL;
    }
    if (minusLst!=NULL) {
      delete[] minusLst;
      minusLst=NULL;
    }
    if (plusLst!=NULL) {
      delete[] plusLst;
      plusLst=NULL;
    }
  }
  catchAll("object::~object(void)","")
}



void object::setType(Int tp) {
   type=tp;
}



void object::setType(string tp) {
  if (tp=="node") 
    type=node;
  else if (tp=="segment")
    type=segment;
  else if (tp=="triangle")
    type=triangle;
  else if (tp=="quadrangle")
    type=quadrangle;
  else
    errThrow("object::setType",string("unknown object type :  ")+tp+".");
}



Int object::getType(void) const {
  return(type);
}



string object::getTypeName(void) const {
  if (type==node)
    return("node");
  else if (type==segment)
    return("segment");
  else if (type==triangle)
    return("triangle");
  else if (type==quadrangle)
    return("quadrangle");
  else
    errThrow("object::getTypeName",string("unknown object type :  ")
	     +strFromInt(type)+".");
}



void object::setMinusNbr(Int nbr) {
  try {
    if (minusDirect!=NULL)
      delete[] minusDirect;
    if (minusLst!=NULL)
      delete[] minusLst;
    
    minusNbr=nbr;
    if (minusNbr!=0) {
      minusDirect=new Int[minusNbr];
      minusLst=new object* [minusNbr];
      for (Int i=0;i<minusNbr;i++) {
	minusDirect[i]=0;
	minusLst[i]=NULL;
      }
    }
    else {
      minusDirect=NULL;
      minusLst=NULL;
    } 
  }
  catchAll("void object::setMinusNbr(Int nbr)","")
}



void object::setMinus(Int pos,object *minus,Int orientation) {
  try {
    minusDirect[pos]=orientation;
    minusLst[pos]=minus;
  }
  catchAll("object::setMinus","")
}



Int object::getMinusNbr(void) const {
  return(minusNbr);
}



object* object::getMinus(Int index) const {
  return(minusLst[index]);
}



Int object::getMinusDirect(Int index) const {
  return(minusDirect[index]);
}



void object::setPlusNbr(Int nbr) {
  try {
    if (plusLst!=NULL)
      delete[] plusLst;
    
    plusNbr=nbr;
    if (plusNbr!=0) {
      plusLst=new object* [plusNbr];
      for (Int i=0;i<plusNbr;i++)
	plusLst[i]=NULL;
    }
    else
      plusLst=NULL;
  }
  catchAll("void object::setPlusNbr(Int nbr)","")
}



void object::setPlus(Int pos,object *plus) {
  plusLst[pos]=plus;
}



Int object::getPlusNbr(void) const {
  return(plusNbr);
}



object* object::getPlus(Int index) const {
  return(plusLst[index]);
}



Int object::getDim(void) const {
  switch(type) {
  case node :
    return(0);
  case segment :
    return(1);
  case triangle :
  case quadrangle :
    return(2);
  default :
    errThrow("object::getDim",string("unknown object type :  ")
	     +strFromInt(type)+".");
  }
}



Int object::split(Int thisDirect,object **objectsLst,Int *objectsDirect) {
  try {
    Int dim,j,counter,index;
    Int thisDim=getDim();
    Int upDirect;
    object *upCurrent;
    Int downNbr,currentDirect;
    object *current;
    
    vector<object*>* list;
    list=new vector<object*>[thisDim+1];
    vector<object*>::iterator voIt;
    map<object*,Int> objDir;
    
    // This object :
    
    list[thisDim].push_back(this);
    objDir[this]=thisDirect;
    
    // One recursively builds the lower order lists :
    
    for (dim=thisDim-1;dim>=0;dim--) {
      for (voIt=list[dim+1].begin();voIt!=list[dim+1].end();voIt++) {
	upCurrent=*voIt;
	upDirect=objDir[upCurrent];
	downNbr=upCurrent->getMinusNbr();
	for (j=0;j<downNbr;j++) {
	  if (upDirect>0) {
	    index=(upDirect+j)%downNbr;
	    current=upCurrent->getMinus(index);
	    currentDirect=upCurrent->getMinusDirect(index);
	  }
	  else {
	    index=(2*downNbr+upDirect-j-1)%downNbr;
	    current=upCurrent->getMinus(index);
	    currentDirect=
	      -upCurrent->getMinusDirect(index);
	  }
	  if (objDir.find(current)==objDir.end()) {
	    list[dim].push_back(current);
	    objDir[current]=currentDirect;
	  }
	}
      }
    }
    
    // One fills the list of objects :
    
    counter=0;
    for (dim=0;dim<=thisDim;dim++) {
      for (voIt=list[dim].begin();voIt!=list[dim].end();voIt++) {
	current=*voIt;
	objectsLst[counter]=current;
	objectsDirect[counter]=objDir[current];
	counter++;
      }
    }
    
    delete[] list;
    
    return(counter);
  }
  catchAll("object::split","")
}



inline object* min(object *obj1,object *obj2) {
  if (obj1<obj2)
    return(obj1);
  else
    return(obj2);
}



inline object* max(object *obj1,object *obj2) {
  if (obj1>obj2)
    return(obj1);
  else
    return(obj2);
}



void object::getPlusMinMax(object* &objMin,object* &objMax) {
  try {
    Int i,dim,thisDim,downNbr;
    object *downCurrent,*current;
    
    thisDim=getDim();
    vector<object*> objList[4];
    vector<object*>::iterator voIt;
    
    objList[thisDim].push_back(this);
    
    for (dim=thisDim+1;dim<4;dim++) {
      for (voIt=objList[dim-1].begin();voIt!=objList[dim-1].end();voIt++) {
	downCurrent=*voIt;
	downNbr=downCurrent->getPlusNbr();
	for (i=0;i<downNbr;i++) {
	  current=downCurrent->getPlus(i);
	  objList[dim].push_back(current);
	}
      }
    }
    
    for (dim=3;dim>=0;dim--) {
      if (objList[dim].begin()!=objList[dim].end()) {
	voIt=objList[dim].begin();
	objMin=objMax=*voIt;
	for (voIt=objList[dim].begin();voIt!=objList[dim].end();voIt++) {
	  objMin=min(objMin,*voIt);
	  objMax=max(objMax,*voIt);
	}
	return;
      }
    }
  }
  catchAll("object::getPlusMinMax","")
}



Int object::getDim(const Int &objectType) {
  switch(objectType) {
  case -1 :
    return(0);
  case node :
    return(0);
  case segment :
    return(1);
  case triangle :
  case quadrangle :
    return(2);
  default :
    errThrow("object::getDim",string("unknown object type :  ")
	     +strFromInt(objectType)+".");
  }
}

