
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <iostream>

#include "array/Array.h"
#include "coordSys/GlobLoc.h"
#include "coordSys/TensorCoordSys.h"
#include "domain/Domain.h"
#include "field/Field.h"
#include "integ/Integ.h"
#include "mesh/Mesh.h"
#include "constr/Constr.h"
#include "timeDep/TimeDep.h"
#include "problem/Problem.h"
#include "result/Result.h"

#include "array/array_base.h"
#include "coordSys/globLoc_base.h"
#include "coordSys/tensorCoordSys_base.h"
#include "domain/domain_base.h"
#include "field/field_base.h"
#include "integ/integ_base.h"
#include "mesh/mesh_base.h"
#include "constr/constr_base.h"
#include "timeDep/timeDep_base.h"
#include "problem/problem_base.h"
#include "result/result_base.h"

#include "param/paramList.h"
#include "err/err.h"
#include "util/utils.h"

using namespace std ;
using namespace keyFE2 ;



paramList::paramList(void) {
  paramSize=0;
  paramCapacity=10;
  params=new param[paramCapacity];
}



paramList::paramList(const paramList &PL) {
  paramSize=PL.paramSize;
  paramCapacity=2*paramSize+1;
  params=new param[paramCapacity];
  for (Int i=0;i<paramSize;i++)
    params[i].equals(PL.params[i]);
}



paramList::~paramList(void) {
   paramSize=0;
   paramCapacity=0;
   delete[] params;
}



void paramList::operator=(const paramList &PL) {
  if (this==&PL)
    return;

  paramSize=PL.paramSize;
  paramCapacity=2*paramSize+1;
  delete[] params;
  params=new param[paramCapacity];
  for (Int i=0;i<paramSize;i++)
    params[i].equals(PL.params[i]);
}



param paramList::operator[](Int index) const {
  if (index<0||index>=paramSize)
    errThrow("paramList::operator[]",string("invalid value of index : ")
	     +strFromInt(index)+".");
  return(params[index]);
}


    
void paramList::pushBack(const param &p) {
  if (paramSize==paramCapacity) {
    param *tmpParams=params;
    paramCapacity=2*paramCapacity+1;
    params=new param[paramCapacity];
    for (Int i=0;i<paramSize;i++)
      params[i].equals(tmpParams[i]);
    delete[] tmpParams;
  }
  params[paramSize].equals(p);
  paramSize++;
}



paramList& paramList::operator,(const param &p) {
  pushBack(p);
  return(*this);
}



paramList& paramList::operator,(const paramList &pl) {
  for (Int i=0;i<pl.paramSize;i++)
    pushBack(pl.params[i]);
  return(*this);
}



paramList paramList::getListOfType(const string &s) const {
  paramList PL;
  for (Int i=0;i<paramSize;i++)
    if (params[i].getName()==s)
      PL.pushBack(params[i]);
  return(PL);
}



paramList paramList::getListNotOfType(const string &s) const {
  paramList PL;
  for (Int i=0;i<paramSize;i++)
    if (params[i].getName()!=s)
      PL.pushBack(params[i]);
  return PL;
}



paramList paramList::getListOfType(const param &p) const {
  paramList PL;
  for (Int i=0;i<paramSize;i++)
    if (params[i].getName()==p.getName())
      PL.pushBack(params[i]);
  return PL;
}



paramList paramList::getListNotOfType(const param &p) const {
  paramList PL;
  for (Int i=0;i<paramSize;i++)
    if (params[i].getName()!=p.getName())
      PL.pushBack(params[i]);
  return PL;
}



Int paramList::getNbrParams(void) const {
  return paramSize;
}



Int paramList::getNbrParamsNamed(const string &sid) const {
  return getListOfType(sid).paramSize;
}



param paramList::extractParamNamed(const string &sId,bool exactlyOne) {
  try {
    paramList PL=getListOfType(sId);
    if (!exactlyOne) {
      if (PL.getNbrParams()>1)
        errThrow("paramList::extractParamNamed",
		 "maximum one parameter named \""
	         +sId+"\" should be present"); 
    }
    else {
      if (PL.getNbrParams()!=1)
        errThrow("paramList::extractParamNamed",
		 "exactly one parameter named \""+sId+
		 "\" should be present");
    }
    param P;
    if (PL.getNbrParams()==1)
      P=PL[0];
    operator=(getListNotOfType(sId));
    return P;
  }
  catchAll("paramList::extractParamNamed","at the very end");
}



Int paramList::extractIntData(string sId,bool exactlyOne) {
  try {
    param P=extractParamNamed(sId,exactlyOne);
    if (P.getTypeName()!="Int"&&P.getTypeName()!="none")
      errThrow("paramList::extractIntData","param named "+sId+
	       " should have \"Int\" type");    
    return P.getIntData();
  }
  catchAll("paramList::extractIntData","at the very end");
}



Double paramList::extractDoubleData(string sId,bool exactlyOne) {
  try {
    param P=extractParamNamed(sId,exactlyOne);
    if (P.getTypeName()!="Double"&&P.getTypeName()!="none")
      errThrow("paramList::extractDoubleData","param named "+sId+
	       " should have \"Double\" type");    
    return P.getDoubleData();
  }
  catchAll("paramList::extractDoubleData","at the very end");
}



string paramList::extractStringData(string sId,bool exactlyOne) {
  try {
    param P=extractParamNamed(sId,exactlyOne);
    if (P.getTypeName()!="string"&&P.getTypeName()!="none")
      errThrow("paramList::extractStringData","param named "+sId+
	       " should have \"string\" type");    
    return P.getStringData();
  }
  catchAll("paramList::extractStringData","at the very end");
}



bool paramList::extractBoolData(string sId,bool exactlyOne) {
  try {
    param P=extractParamNamed(sId,exactlyOne);
    if (P.getTypeName()!="bool"&&P.getTypeName()!="none")
      errThrow("paramList::extractBoolData","param named "+sId+
	       " should have \"bool\" type");    
    return P.getBoolData();
  }
  catchAll("paramList::extractBoolData","at the very end");
}



const Int* paramList::extractComponentsData(string sId,bool exactlyOne) {
  try {
    param P=extractParamNamed(sId,exactlyOne);
    return P.getComponentsData();
  }
  catchAll("paramList::extractComponentsData","at the very end");
}



parents* paramList::extractParentData(string sId,bool exactlyOne) {
  try {
    param P=extractParamNamed(sId,exactlyOne);
    if (P.getTypeName()!="parents"&&P.getTypeName()!="none")
      errThrow("paramList::extractParentData","param named "+sId+
	       " should have \"parents\" type");    
    return P.getParentData();
  }
  catchAll("paramList::extractParentData","at the very end");
}



Array paramList::extractArrayData(string sId,bool exactlyOne) {
  try {
    Array A;
    A.setData(dynamic_cast<array*>(extractParentData(sId,exactlyOne)));
    return A;
  }
  catchAll("paramList::extractArrayData","at the very end");
}



GlobLoc paramList::extractGlobLocData(string sId,bool exactlyOne) {
  try {
    GlobLoc GL;
    GL.setData(dynamic_cast<globLoc*>(extractParentData(sId,exactlyOne)));
    return GL;
  }
  catchAll("paramList::extractGlobLocData","at the very end");
}



TensorCoordSys paramList::extractTensorCoordSysData(string sId,bool exactlyOne) {
  try {
    TensorCoordSys TCS;
    TCS.setData(dynamic_cast<tensorCoordSys*>(extractParentData(sId,exactlyOne)));
    return TCS;
  }
  catchAll("paramList::extractTensorCoordSysData","at the very end");
}



Domain paramList::extractDomainData(string sId,bool exactlyOne) {
  try {
    Domain D;
    D.setData(dynamic_cast<domain*>(extractParentData(sId,exactlyOne)));
    return D;
  }
  catchAll("paramList::extractDomainData","at the very end");
}



Field paramList::extractFieldData(string sId,bool exactlyOne) {
  try {
    Field F;
    F.setData(dynamic_cast<field*>(extractParentData(sId,exactlyOne)));
    return F;
  }
  catchAll("paramList::extractFieldData","at the very end");
}



Integ paramList::extractIntegData(string sId,bool exactlyOne) {
  try {
    Integ I;
    I.setData(dynamic_cast<integ*>(extractParentData(sId,exactlyOne)));
    return I;
  }
  catchAll("paramList::extractIntegData","at the very end");
}



Mesh paramList::extractMeshData(string sId,bool exactlyOne) {
  try {
    Mesh M;
    M.setData(dynamic_cast<mesh*>(extractParentData(sId,exactlyOne)));
    return M;
  }
  catchAll("paramList::extractMeshData","at the very end");
}



Constr paramList::extractConstrData(string sId,bool exactlyOne) {
  try {
    Constr C;
    C.setData(dynamic_cast<constr*>(extractParentData(sId,exactlyOne)));
    return C;
  }
  catchAll("paramList::extractConstrData","at the very end");
}



TimeDep paramList::extractTimeDepData(string sId,bool exactlyOne) {
  try {
    TimeDep TD;
    TD.setData(dynamic_cast<timeDep*>(extractParentData(sId,exactlyOne)));
    return TD;
  }
  catchAll("paramList::extractTimeDepData","at the very end");
}



Problem paramList::extractProblemData(string sId,bool exactlyOne) {
  try {
    Problem P;
    P.setData(dynamic_cast<problem*>(extractParentData(sId,exactlyOne)));
    return P;
  }
  catchAll("paramList::extractProblemData","at the very end");
}



Result paramList::extractResultData(string sId,bool exactlyOne) {
  try {
    Result R;
    R.setData(dynamic_cast<result*>(extractParentData(sId,exactlyOne)));
    return R;
  }
  catchAll("paramList::extractResultData","at the very end");
}



void paramList::print(void) const {
  for (Int i=0;i<paramSize;i++)
    params[i].print();
}




