
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <iostream>
#include <iomanip>

#include "Plot.h"
#include "plot_base.h"
#include "array/Array.h"
#include "array/array_base.h"
#include "domain/domain_base.h"
#include "err/err.h"
#include "param/param.h"
#include "param/paramList.h"
#include "result/result_base.h"

using namespace std ;
using namespace keyFE2 ;



Plot::Plot(void) {
  try {
    plot *P=new plot;
    data=NULL;
    setData(P);
  }
  catchAll("Plot::Plot(void)","")
}



Plot::Plot(const Plot &in) {
  try {
    data=NULL;
    setData(in.getData());
  }
  catchAll("Plot::Plot(const Plot &in)","")
}



Plot::Plot(param P) {
  try {
    paramList PL;
    PL.pushBack(P);
    Plot p(PL);
    data=NULL;
    setData(p.getData());
  }
  catchAll("Plot::Plot(param P)","")
}



Plot::Plot(paramList PL) {
  try {
    paramList plName=PL.getListOfType(NAME);
    PL=PL.getListNotOfType(NAME);
    paramList plType=PL.getListOfType(TYPE);
    PL=PL.getListNotOfType(TYPE);
    paramList plDomain=PL.getListOfType(DOM);
    PL=PL.getListNotOfType(DOM);
    paramList plCoordinates=PL.getListOfType(COORD);
    PL=PL.getListNotOfType(COORD);
    paramList plResults=PL.getListOfType(RESULT);
    PL=PL.getListNotOfType(RESULT);
    
    // Checks for illegal arguments :
    
    if (PL.getNbrParams()!=0) {
      string tmpString=strFromInt(PL.getNbrParams())+" invalid parameters :";
      for (Int i=0;i<PL.getNbrParams();i++) {
	tmpString+=" ";
	tmpString+=PL[i].getName();
      }
      tmpString+=".\n";
      errWrite("Plot construction",tmpString);
    }
    
    plot *outputP=new plot;
    
    // Sets various plot parameters :
    
    if (plName.getNbrParams()>1) {
      errWrite("Plot construction","maximum one NAME parameter is needed.");
    }
    else if (plName.getNbrParams()==1) {
      param PC=plName[0];
      if (PC.getTypeName()!="string") {
	errWrite("Plot construction",
		 "NAME= must be followed by a string argument.");
      }
      else {
        string tmpS=PC.getStringData();
	outputP->setName(tmpS);
      }
    }
    else
      outputP->setName("");
    
    if (plType.getNbrParams()>1) {
      errWrite("Plot construction","maximum one TYPE parameter is needed.");
    }
    else if (plType.getNbrParams()==1) {
      param PC=plType[0];
      if (PC.getTypeName()!="string") {
	errWrite("Plot construction",
		 "TYPE= must be followed by a string argument.");
      }
      else {
        string tmpS=PC.getStringData();
	outputP->setType(tmpS);
      }
    }
    else
      outputP->setType("vigie2D");
    
    if (plCoordinates.getNbrParams()!=1) {
      errWrite("Plot construction","exactly one COORD parameter is needed.");
    }
    else {
      param PC=plCoordinates[0];
      array *tmpA=dynamic_cast<array*>(PC.getParentData());
      if (tmpA==NULL) {
	errWrite("Plot construction",
		 "COORD= must be followed by an Array argument.");
      }
      else
	outputP->setCoordinates(tmpA);
    }
    
    if (plDomain.getNbrParams()!=1) {
      errWrite("Plot construction","exactly one DOM parameter is needed.");
    }
    else {
      param PD=plDomain[0];
      domain *dom=dynamic_cast<domain*>(PD.getParentData());
      if (PD.getParentData()==NULL) {
	errWrite("Plot construction",
		 "DOM= must be followed by a Domain argument.");
      }
      else
	outputP->setDomain(dom);
    }
    
    // Sets the Results printed in the plot :
    
    for (Int i=0;i<plResults.getNbrParams();i++) {
      param PR=plResults[i];
      array *tmpA=dynamic_cast<array*>(PR.getParentData());
      result *tmpR=dynamic_cast<result*>(PR.getParentData());
      if (tmpA==NULL&&tmpR==NULL) {
	errWrite("Plot construction",
		 "RESULT= must be followed by an Array or a Result argument.");
      }
      else {
        if (tmpA!=NULL)
  	  outputP->addResult(tmpA);
	else if (tmpR!=NULL)
  	  outputP->addResult(tmpR);
      }
    }
    
    if (!err::getOk())
      errThrow("Plot construction","bailing out due to preceding errors.");
    
    data=NULL;
    setData(outputP);  
  }
  catchAll("Plot::Plot(paramList PL)","")
}



Plot::~Plot(void) {
  try {
    setData(NULL);
  }
  catchAll("Plot::~Plot(void)","")
}



void Plot::setData(plot *newPlotData) {
  try {
    plot *S=getData();
    delinkAndDelete(S,&data);
    if (newPlotData!=NULL)
      newPlotData->addParent(&data);
    data=static_cast<void*>(newPlotData);
  }
  catchAll("void Plot::setData(plot *newPlotData)","")
}



plot* Plot::getData(void) const {
  return(static_cast<plot*>(data));
}



void Plot::setName(string s) {
  try {
    plot *plotData=getData();
    plotData->setName(s);
  }
  catchAll("void Plot::setName(string s)","")
}



string Plot::getName(void) const {
  try {
    plot *plotData=getData();
    return(plotData->getName());
  }
  catchAll("string Plot::getName(void) const","")
}



void Plot::write(void) const {
  try {
    plot *plotData=getData();
    plotData->write();
  }
  catchAll("Plot::write","");
}



void Plot::writeValues(void) const {
  try {
    plot *plotData=getData();
    plotData->writeValues();
  }
  catchAll("Plot::writeValues","");
}



void Plot::write(ostream &out) const {
  try {
    plot *plotData=getData();
    plotData->write(out);
  }
  catchAll("Plot::write","");
}



void Plot::writeValues(ostream &out) const {
  try {
    plot *plotData=getData();
    plotData->writeValues(out);
  }
  catchAll("Plot::writeValues","");
}



void Plot::operator=(const Plot &Sin) {
  if (this==&Sin)
    return;
  
  try {
    setData(new plot(*(Sin.getData())));
  }
  catchAll("void Plot::operator=(const Plot &Sin)","")
}



void Plot::exec(string s) {
  try {
    plot *plotData=getData();
    plotData->exec(s);
  }
  catchAll("void Plot::exec(string s)","")
}



