
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <stdio.h>
#include <string>

#include "problem_base.h"
#include "array/array_base.h"
#include "constr/constr_base.h"
#include "err/err.h"
#include "print/printTree.h"

using namespace std ;
using namespace keyFE2 ;



problem::problem(void) : name(), parents() {
  try {
    arrayCapacity=10;
    arraySize=0;
    aList=new array* [arrayCapacity];
    for (Int i=0;i<arrayCapacity;i++)
      aList[i]=NULL;
    constrCapacity=10;
    constrSize=0;
    cList=new constr* [constrCapacity];
    for (Int i=0;i<constrCapacity;i++)
      cList[i]=NULL;
  }
  catchAll("problem::problem(void)","")
}



problem::problem(const problem &in) : name(in), parents(in) {
  try {
    arrayCapacity=in.arrayCapacity;
    arraySize=0;
    aList=new array* [arrayCapacity];
    for (Int i=0;i<in.arraySize;i++)
      addArray(in.aList[i]);
    constrCapacity=in.constrCapacity;
    constrSize=0;
    cList=new constr* [constrCapacity];
    for (Int i=0;i<in.constrSize;i++)
      addConstr(in.cList[i]);
  }
  catchAll("problem::problem(const problem &in)","")
}



problem::~problem(void) {
  try {
    for (Int i=0;i<arraySize;i++) {
      delinkAndDelete(aList[i],aList+i);
      aList[i]=NULL;
    }
    delete[] aList;
    for (Int i=0;i<constrSize;i++) {
      delinkAndDelete(cList[i],cList+i);
      cList[i]=NULL;
    }
    delete[] cList;
  }
  catchAll("problem::~problem(void)","")
}



void problem::addArray(array *inArray) {
  try {
    if (arraySize==arrayCapacity) {
      array **tmpArray=new array* [arrayCapacity*2+1];
      for (Int i=0;i<arraySize;i++) {
	tmpArray[i]=aList[i];
	tmpArray[i]->addParent(tmpArray+i);
	delinkAndDelete(aList[i],aList+i);
	aList[i]=NULL;
      }
      delete[] aList;
      aList=tmpArray;
      arrayCapacity=2*arrayCapacity+1;
    }
    aList[arraySize]=inArray;
    inArray->addParent(aList+arraySize);
    arraySize++;
  }
  catchAll("void problem::addArray(array *inArray)","")
}



void problem::addConstr(constr *inConstr) {
  try {
    if (constrSize==constrCapacity) {
      constr **tmpConstr=new constr* [constrCapacity*2+1];
      for (Int i=0;i<constrSize;i++) {
	tmpConstr[i]=cList[i];
	tmpConstr[i]->addParent(tmpConstr+i);
	delinkAndDelete(cList[i],cList+i);
	cList[i]=NULL;
      }
      delete[] cList;
      cList=tmpConstr;
      constrCapacity=2*constrCapacity+1;
    }
    cList[constrSize]=inConstr;
    inConstr->addParent(cList+constrSize);
    constrSize++;
  }
  catchAll("void problem::addConstr(constr *inConstr)","")
}



Int problem::getNbrArrays(void) const {
   return arraySize;
}


 
array* problem::getArray(Int pos) const {
  try {
    return aList[pos];
  }
  catchAll("array* problem::getArray(Int pos)","")
}



Int problem::getNbrConstrs(void) const {
   return constrSize;
}



constr* problem::getConstr(Int pos) const {
  try {
    return cList[pos];
  }
  catchAll("constr* problem::getConstr(Int pos)","")
}



void problem::setNewPoint(bool b) {
  for (Int i=0;i<arraySize;i++)
    aList[i]->setNewPoint(b);
  for (Int i=0;i<constrSize;i++)
    cList[i]->setNewPoint(b);
  newPoint::setNewPoint(b);
}



void problem::accessNbrInit(void) {
  for (Int i=0;i<arraySize;i++)
    aList[i]->accessNbrInit();
  for (Int i=0;i<constrSize;i++)
    cList[i]->accessNbrInit();
  newPoint::accessNbrInit();
}



printTree problem::getPrintTreeData(void) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf;
    printTree pt;
    
    if (getName()=="")
      sprintf(cbuf,"   anonymous");
    else
      sprintf(cbuf,"   name = %s",getName().c_str());
    sbuf=cbuf;
    pt+=sbuf;
    
    return(pt);
  }
  catchAll("problem::getPrintTreeData","")
}



printTree problem::getPrintTreeValues(void) const {
  try {
    printTree pt;
    return(pt);
  }
  catchAll("problem::getPrintTreeValues","")
}



printTree problem::getPrintTree(bool values) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf,typeName;
    printTree pt,ptArg;
    Int i;
    
    sprintf(cbuf,"problem %p",static_cast<const void*>(this));
    sbuf=cbuf;
    pt+=sbuf;
    
    pt+=getPrintTreeData();
    
    if (values)
      pt+=getPrintTreeValues();
    
    sbuf="   ";
    pt+=sbuf;
    
    pt.prepareForArguments();

    sprintf(cbuf," | arraySize = %d",arraySize);
    pt+=cbuf;
    sbuf=" | ";
    pt+=sbuf;
    for (i=0;i<arraySize;i++) {
      sprintf(cbuf,"aList[%d]",i);
      ptArg=aList[i]->print::getPrintTreeSummary(cbuf);
      ptArg.prepareForAdd();
      pt+=ptArg;
    }
    
    sprintf(cbuf," | constrSize = %d",constrSize);
    pt+=cbuf;
    sbuf=" | ";
    pt+=sbuf;
    for (i=0;i<constrSize;i++) {
      sprintf(cbuf,"cList[%d]",i);
      ptArg=cList[i]->print::getPrintTreeSummary(cbuf);
      if (i<constrSize-1)
	ptArg.prepareForAdd();
      else
	ptArg.prepareForAddLast();
      pt+=ptArg;
    }
    
    return(pt);
  }
  catchAll("problem::getPrintTree","")
}



printTree problem::getPrintTreeSummary(void) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf,typeName;
    printTree pt;
    
    sprintf(cbuf,"problem %p",static_cast<const void*>(this));
    sbuf=cbuf;
    pt+=sbuf;
    
    if (getName()=="")
      sprintf(cbuf,"   anonymous");
    else
      sprintf(cbuf,"   name = %s",getName().c_str());
    sbuf=cbuf;
    pt+=sbuf;
    
    sbuf="   ";
    pt+=sbuf;
    
    return(pt);
  }
  catchAll("problem::getPrintTreeSummary","")
}

    
