
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <iomanip>
#include <iostream>

#include "shape.h"
#include "interpolationTypes.h"
#include "domain/domain_base.h"
#include "err/err.h"
#include "object/object_base.h"
#include "object/objectTypes.h"

using namespace std ;
using namespace keyFE2 ;



int shape::oneDimSF(const Int &interpType,Double &coord1,Double *F,
		    Int orderDeriv) {
  
  try {
  
    if (orderDeriv>1) {
      errThrow("shape::oneDimSF","orderDeriv>1");
    }
    
    register Double xi=coord1;
    
    switch(interpType) {
      
    case none :
      return(0);
      
    case constantOnDomain :
    case constantPerElement :
      switch(orderDeriv) {
      case 0 :
	F[0]=1.0;
	return(1);
      case 1 :
	F[0]=0.0;
	return(1); 
      }
      break;
      
    case linear :
    case linearPerElement :
      switch(orderDeriv) {
      case 0 :
	F[0]=0.5*(1.0-xi);
	F[1]=0.5*(1.0+xi);
	return(2);
      case 1 :
	F[0]=-0.5;
	F[1]=0.5;
	return(2);
      }
      break;
      
    case quadratic :
    case Serendip :
    case quadraticPerElement :
    case SerendipPerElement :
      switch(orderDeriv) {
      case 0 :
	F[0]=0.5*xi*(xi-1.0);
	F[1]=0.5*xi*(xi+1.0);
	F[2]=1.0-xi*xi;
	return(3);
      case 1 :
	F[0]=xi-0.5;
	F[1]=xi+0.5;
	F[2]=-2.0*xi;
	return(3);
      }
      break;
      
    default :
      errThrow("shape::oneDimSF",
	       string("unknown interpolation type : ")+strFromInt(interpType)
	       +".");
      
    }
    return(0);
  }
  catchAll("shape::oneDimSF","")
}



Int shape::function(const Int &interpType,const Int &objectType,
		    domain *defDomain,Double coord[3],Int corlocDeriv[3],Double *F) {
  
  try {
    
    if (corlocDeriv[0]+corlocDeriv[1]+corlocDeriv[2]>1) {
      errThrow("shape::function",
               "corlocDeriv[0]+corlocDeriv[1]+corlocDeriv[2]>1");
    }
    
    static Int i,domainDim,objectDim;
    static Double tmpF[3][MAX_OBJ_PER_ELT];
    static Double *F0=tmpF[0];
    static Double *F1=tmpF[1];
    register Double xi,eta,zeta;
    
    xi=coord[0];
    eta=coord[1];
    zeta=coord[2];
    
    objectDim=object::getDim(objectType);
    domainDim=defDomain->getDim();
    
    if (interpType==constantPerElement||interpType==linearPerElement
        ||interpType==quadraticPerElement||interpType==SerendipPerElement)
      if (objectDim!=domainDim)
        return(0);
    
    switch(objectDim) {
    case 0 :
      if (corlocDeriv[0]!=0||corlocDeriv[1]!=0||corlocDeriv[2]!=0)
	return(0);
      break;
    case 1 :
      if (corlocDeriv[1]!=0||corlocDeriv[2]!=0)
	return(0);
      break;
    case 2 :
      if (corlocDeriv[2]!=0)
	return(0);
      break;
    default :
      break;
    }
    
    switch(interpType) {
      
    case constantOnDomain :
      
      F[0]=1.0;
      for (i=0;i<objectDim;i++)
	if (corlocDeriv[i]!=0)
	  F[0]=0.0;
      return(1);
      
    case constantPerElement :
      
      F[0]=1.0;
      for (i=0;i<objectDim;i++)
	if (corlocDeriv[i]!=0)
	  F[0]=0.0;
      return(1);
      
    case linear :
    case linearPerElement :
      
      switch (objectType) {
      case node :
	F[0]=1.0;
	return(1);
      case segment :
	return(oneDimSF(interpType,coord[0],F,corlocDeriv[0]));
      case triangle :
	F[0]=0.0;
	F[1]=0.0;
	F[2]=0.0;
	if (corlocDeriv[0]==0&&corlocDeriv[1]==0) {
	  F[0]=1.0-xi-eta;
	  F[1]=xi;
	  F[2]=eta;
	}
	else if (corlocDeriv[0]==1&&corlocDeriv[1]==0) {
	  F[0]=-1.0;
	  F[1]=1.0;
	}
	else if (corlocDeriv[0]==0&&corlocDeriv[1]==1) {
	  F[0]=-1.0;
	  F[2]=1.0;
	}
	return(3);
      case quadrangle :
	oneDimSF(interpType,coord[0],F0,corlocDeriv[0]);
	oneDimSF(interpType,coord[1],F1,corlocDeriv[1]);
	F[0]=F0[1]*F1[1];
	F[1]=F0[0]*F1[1];
	F[2]=F0[0]*F1[0];
	F[3]=F0[1]*F1[0];
	return(4);
	
      } // switch objectType
      
    case quadratic :
    case quadraticPerElement :
      
      switch (objectType) {
      case node :
	F[0]=1.0;
	return(1);
      case segment :
	return(oneDimSF(interpType,coord[0],F,corlocDeriv[0]));
      case triangle :
	F[0]=0.0;
	F[1]=0.0;
	F[2]=0.0;
	F[3]=0.0;
	F[4]=0.0;
	F[5]=0.0;
	if (corlocDeriv[0]==0&&corlocDeriv[1]==0) {
	  F[0]=1.0-3.0*(xi+eta)+2*(xi+eta)*(xi+eta);
	  F[1]=xi*(2.0*xi-1.0);
	  F[2]=eta*(2.0*eta-1.0);
	  F[3]=4.0*xi*(1.0-xi-eta);
	  F[4]=4.0*xi*eta;
	  F[5]=4.0*eta*(1.0-xi-eta);
	}
	else if (corlocDeriv[0]==1&&corlocDeriv[1]==0) {
	  F[0]=-3.0+4.0*(xi+eta);
	  F[1]=4.0*xi-1.0;
	  F[3]=4.0-8.0*xi-4.0*eta;
	  F[4]=4.0*eta;
	  F[5]=-4.0*eta;
	}
	else if (corlocDeriv[0]==0&&corlocDeriv[1]==1) {
	  F[0]=-3.0+4.0*(xi+eta);
	  F[2]=4.0*eta-1.0;
	  F[3]=-4.0*xi;
	  F[4]=4.0*xi;
	  F[5]=4.0-4.0*xi-8.0*eta;
	}
	return(6);
      case quadrangle :
	oneDimSF(interpType,coord[0],F0,corlocDeriv[0]);
	oneDimSF(interpType,coord[1],F1,corlocDeriv[1]);
	F[0]=F0[1]*F1[1];
	F[1]=F0[0]*F1[1];
	F[2]=F0[0]*F1[0];
	F[3]=F0[1]*F1[0];
	F[4]=F0[2]*F1[1];
	F[5]=F0[0]*F1[2];
	F[6]=F0[2]*F1[0];
	F[7]=F0[1]*F1[2];
	F[8]=F0[2]*F1[2];
	return(9);
	
      } // switch objectType
      
    case Serendip :
    case SerendipPerElement :
      
      switch (objectType) {
      case node :
	F[0]=1.0;
	return(1);
      case segment :
	return(oneDimSF(interpType,coord[0],F,corlocDeriv[0]));
      case triangle :
	F[0]=0.0;
	F[1]=0.0;
	F[2]=0.0;
	F[3]=0.0;
	F[4]=0.0;
	F[5]=0.0;
	if (corlocDeriv[0]==0&&corlocDeriv[1]==0) {
	  F[0]=1.0-3.0*(xi+eta)+2*(xi+eta)*(xi+eta);
	  F[1]=xi*(2.0*xi-1.0);
	  F[2]=eta*(2.0*eta-1.0);
	  F[3]=4.0*xi*(1.0-xi-eta);
	  F[4]=4.0*xi*eta;
	  F[5]=4.0*eta*(1.0-xi-eta);
	}
	else if (corlocDeriv[0]==1&&corlocDeriv[1]==0) {
	  F[0]=-3.0+4.0*(xi+eta);
	  F[1]=4.0*xi-1.0;
	  F[3]=4.0-8.0*xi-4.0*eta;
	  F[4]=4.0*eta;
	  F[5]=-4.0*eta;
	}
	else if (corlocDeriv[0]==0&&corlocDeriv[1]==1) {
	  F[0]=-3.0+4.0*(xi+eta);
	  F[2]=4.0*eta-1.0;
	  F[3]=-4.0*xi;
	  F[4]=4.0*xi;
	  F[5]=4.0-4.0*xi-8.0*eta;
	}
	return(6);
      case quadrangle :
	F[0]=0.0;
	F[1]=0.0;
	F[2]=0.0;
	F[3]=0.0;
	F[4]=0.0;
	F[5]=0.0;
	F[6]=0.0;
	F[7]=0.0;
	if (corlocDeriv[0]==0&&corlocDeriv[1]==0) {
	  F[0]=0.25*(1.0+xi)*(1.0+eta)*(xi+eta-1.0);
	  F[1]=0.25*(1.0-xi)*(1.0+eta)*(-xi+eta-1.0);
	  F[2]=0.25*(1.0-xi)*(1.0-eta)*(-xi-eta-1.0);
	  F[3]=0.25*(1.0+xi)*(1.0-eta)*(xi-eta-1.0);
	  F[4]=0.5*(1.0-xi*xi)*(1.0+eta);
	  F[5]=0.5*(1.0-xi)*(1.0-eta*eta);
	  F[6]=0.5*(1.0-xi*xi)*(1.0-eta);
	  F[7]=0.5*(1.0+xi)*(1.0-eta*eta);
	}
	else if (corlocDeriv[0]==1&&corlocDeriv[1]==0) {
	  F[0]=0.25*(1.0+eta)*(2.0*xi+eta);
	  F[1]=0.25*(1.0+eta)*(2.0*xi-eta);
	  F[2]=0.25*(1.0-eta)*(2.0*xi+eta);
	  F[3]=0.25*(1.0-eta)*(2.0*xi-eta);
	  F[4]=-xi*(1.0+eta);
	  F[5]=0.5*(eta*eta-1.0);
	  F[6]=xi*(eta-1.0);
	  F[7]=0.5*(1.0-eta*eta);
	}
	else if (corlocDeriv[0]==0&&corlocDeriv[1]==1) {
	  F[0]=0.25*(1.0+xi)*(2.0*eta+xi);
	  F[1]=0.25*(1.0-xi)*(2.0*eta-xi);
	  F[2]=0.25*(1.0-xi)*(2.0*eta+xi);
	  F[3]=0.25*(1.0+xi)*(2.0*eta-xi);
	  F[4]=0.5*(1.0-xi*xi);
	  F[5]=eta*(xi-1.0);
	  F[6]=0.5*(xi*xi-1.0);
	  F[7]=-eta*(1.0+xi);
	}
	return(8);
	
      } // switch objectType
      
    } // switch interpType
    return(0);
    
  }
  catchAll("shape::function","")
}



Int shape::nbrFunctions(const Int &interpType,const Int &objectType) {

  try{
    
    switch(interpType) {
      
    case constantOnDomain :
      
      return(1);
      
    case constantPerElement :
      
      return(1);
      
    case linear :
    case linearPerElement :
      
      switch (objectType) {
      case node :
	return(1);
      case segment :
	return(2);
      case triangle :
	return(3);
      case quadrangle :
	return(4);
	
      } // switch objectType
      
    case quadratic :
    case quadraticPerElement :
      
      switch (objectType) {
      case node :
	return(1);
      case segment :
	return(3);
      case triangle :
	return(6);
      case quadrangle :
	return(9);
	
      } // switch objectType
      
    case Serendip :
    case SerendipPerElement :
      
      switch (objectType) {
      case node :
	return(1);
      case segment :
	return(3);
      case triangle :
	return(6);
      case quadrangle :
	return(8);
	
      } // switch objectType
      
    } // switch interpType
    return(0);
  }
  catchAll("shape::nbrFunctions","")
}

