
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <cmath>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <list>
#include <map>
#include <stdio.h>
#include <vector>

#include "solver_base.h"
#include "array/array_base.h"
#include "array/arrayFilter.h"
#include "array/arrayList.h"
#include "constr/constr_base.h"
#include "err/err.h"
#include "field/field_base.h"
#include "print/printTree.h"
#include "scal/scal.h"
#include "step/step.h"
#include "step/stepList.h"



using namespace std ;
using namespace keyFE2 ;



inline void swap(Int &a,Int &b) {
  Int c=a;
  a=b;
  b=c;
}



void solver::initForConstruct(void) {
  try {
    problemsCapacity=10;
    problemsSize=0;
    problems=new array* [problemsCapacity];
    for (Int i=0;i<problemsCapacity;i++)
      problems[i]=NULL;
    
    zero=1.0e-8;
    silent=false;
    
    constraintsCapacity=10;
    constraintsSize=0;
    constraints=new constr* [constraintsCapacity];
    for (Int i=0;i<constraintsCapacity;i++)
      constraints[i]=NULL;
    
    unknownsCapacity=0;
    bUnknowns=NULL;
    lUnknowns=NULL;
    nvSize=0;
    NVBackup=NULL;
    NVDelta=NULL; 
    NVConstrs=NULL;
    
    appearance=NULL;
    disappearance=NULL;
    
    nbrActMax=0;
    mat=NULL;
    rhand=NULL;
    active=NULL;
    matPointers=NULL;
    globalPointers=NULL;
    
    map2Max=0;
    ligMap2=NULL;
    colMap2=NULL;
    
    nbrVariables=0;
    varToNVPointers=NULL;
    upperMat=NULL;
    
    tmpSize=0;
    tmpCapacity=0;
    tmpBool=NULL;
    tmpCol=NULL;
    tmpMat=NULL;
    tmpMatB=NULL;
    
    maxNbrDelayed=0;
    autoFix=false;
    autoFixValue=0.0;
    permut=false;
    
    lastState=-1; // none
  }
  catchAll("void solver::initForConstruct(void)","")
}



void solver::clear(void) {
  try {
    frontaleClear();
  }
  catchAll("void solver::clear(void)","")
}



solver::solver(void) {  
  try {
    initForConstruct();
  }
  catchAll("solver::solver(void)","")
}



solver::solver(array *A) {
  try {
    initForConstruct();
    addProblem(A);
  }
  catchAll("solver::solver(array *A)","")
}



solver::solver(const solver &in) {
  try {
    initForConstruct();
    
    for (Int i=0;i<in.problemsSize;i++)
      addProblem(in.problems[i]);
    
    zero=in.zero;
    
    for (Int i=0;i<in.constraintsSize;i++)
      addConstraint(in.constraints[i]);

    maxNbrDelayed=in.maxNbrDelayed;
    autoFix=in.autoFix;
    autoFixValue=in.autoFixValue;
    permut=in.permut;
  }
  catchAll("solver::solver(const solver &in)","")
}



solver::~solver(void) {
  try {
    for (Int i=0;i<problemsSize;i++)
      delinkAndDelete(problems[i],problems+i);
    delete[] problems;
    problems=NULL;
    problemsCapacity=0;
    problemsSize=0;
    
    zero=0.0;
    silent=false;
    
    for (Int i=0;i<constraintsSize;i++)
      delinkAndDelete(constraints[i],constraints+i);
    delete[] constraints;
    constraints=NULL;
    constraintsCapacity=0;
    constraintsSize=0;
    
    clear();
    
    maxNbrDelayed=0;
    autoFix=false;
    autoFixValue=0.0;
    permut=false;
    
    lastState=-1; // none
  }
  catchAll("solver::~solver(void)","")
}



void solver::addProblem(array *A) {
  try {
    if (problemsSize==problemsCapacity) {
      array **tmpProblems=new  array* [problemsCapacity*2+1];
      for (Int i=0;i<problemsSize;i++) {
	tmpProblems[i]=problems[i];
	tmpProblems[i]->addParent(tmpProblems+i);
	delinkAndDelete(problems[i],problems+i);
	problems[i]=NULL;
      }
      delete[] problems;
      problems=tmpProblems;
      problemsCapacity=2*problemsCapacity+1;
    }
    problems[problemsSize]=A;
    A->addParent(problems+problemsSize);
    problemsSize++;
  }
  catchAll("void solver::addProblem(array *A)","")
}



void solver::addConstraint(constr *C) {
  try {
    if (constraintsSize==constraintsCapacity) {
      constr **tmpConstraints=new constr* [constraintsCapacity*2+1];
      for (Int i=0;i<constraintsSize;i++) {
	tmpConstraints[i]=constraints[i];
	tmpConstraints[i]->addParent(tmpConstraints+i);
	delinkAndDelete(constraints[i],constraints+i);
	constraints[i]=NULL;
      }
      delete[] constraints;
      constraints=tmpConstraints;
      constraintsCapacity=constraintsCapacity*2+1;
    }
    constraints[constraintsSize]=C;
    C->addParent(constraints+constraintsSize);
    constraintsSize++;
  }
  catchAll("void solver::addConstraint(constr *C)","")
}



void solver::setZero(Double d) {
  zero=d;
}



void solver::constraintsPrepare(void) {
  try {
    Int i,j;
    
    bool whiteBackup=scal::getWhite();
    scal::setWhite(false);
    
    list<expression> EL1;
    list<expression>::iterator ELit;
    exprCollect tmpEC;
    
    for (i=0;i<constraintsSize;i++) {
      constraints[i]->generateExprCollect();
      expression *exprData=constraints[i]->EC.exprData;
      Int exprDataSize=constraints[i]->EC.size;
      for (j=0;j<exprDataSize;j++)
	EL1.push_back(exprData[j]);
    }
    
    if (EL1.size()!=0) {
      EL1.sort();
      tmpEC.setCapacity(EL1.size());
      for (ELit=EL1.begin();ELit!=EL1.end();ELit++)
	tmpEC.pushBack(*ELit);
      tmpEC.solve();
      EC.setCapacity(tmpEC.size);
      EC=tmpEC;
    }
    
    scal::setWhite(whiteBackup);
  }
  catchAll("void solver::constraintsPrepare(void)","")
}



void solver::frontalePrepare(void) {
  try {
    Int i,j,k,index,index2;
    step *currentStep;
    scal *tmpScal;
    Int ligSize,colSize;
    bool *ligBool,*colBool;
    Int ligIndex,colIndex;
    Int *scalMap;
    set<parents*> tmpS;
    set<parents*>::iterator Sit;
    arrayFilter *tmpAF;
    Int currentLayer;

    // Prparer ici les contraintes :
    constraintsPrepare();
    
    scal::setWhite(true);
    
    // steps :
    
    for (i=0;i<problemsSize;i++) {
      SL+=problems[i]->getStepList();
      problems[i]->buildActiveFieldList();
    }
    SL.sortForFrontale();
    slSize=SL.getSize();
    
    // filters :
    
    SAF.clear();
    for (i=0;i<problemsSize;i++) {
      tmpS=problems[i]->getListOfType("arrayFilter");
      for (Sit=tmpS.begin();Sit!=tmpS.end();Sit++) {
	tmpAF=dynamic_cast<arrayFilter*>(*Sit);
	tmpAF->generateFilters();
	SAF.insert(tmpAF);
      }
    }
    
    // fields and nodal values :
    
    nvSize=field::getAllNodalValuesSize();
    
    if (nvSize>0) {
      NVDelta=new Double[nvSize];
      NVBackup=new Double[nvSize];
      NVConstrs=new expression* [nvSize];
    }
    else {
      NVDelta=NULL;
      NVBackup=NULL;
      NVConstrs=NULL;
    }
    
    // Calculation of appearance and disappearance lists :
    
    unknownsCapacity=nvSize;
    if (nvSize>0) {
      bUnknowns=new bool[nvSize];
      lUnknowns=new Int[nvSize];
    }
    else {
      bUnknowns=NULL;
      lUnknowns=NULL;
    }
    
    if (nvSize>0) {
      appearance=new Int[nvSize];
      disappearance=new Int[nvSize];
      matPointers=new Int[nvSize];
    }
    else {
      appearance=NULL;
      disappearance=NULL;
      matPointers=NULL;
    }
    for (i=0;i<nvSize;i++) {
      bUnknowns[i]=false;
      lUnknowns[i]=MIN_LAYER-1;
      appearance[i]=-1;
      disappearance[i]=-1;
      matPointers[i]=-1;
      NVConstrs[i]=NULL;
    }
    
    for (i=0;i<EC.size;i++)
      NVConstrs[EC.exprData[i].index]=EC.exprData+i;
    
    for (i=0;i<slSize;i++) {
      currentStep=SL[i];
      currentStep->exec();
      
      tmpScal=currentStep->getScalPtr();
      scalMap=tmpScal->getMap();
      currentLayer=currentStep->getEquation()->getLayer();
      
      ligSize=tmpScal->getLigSize();
      ligIndex=tmpScal->getLigIndex();
      map2Max=max(map2Max,ligSize);
      ligBool=tmpScal->getLigBool();
      for (j=0;j<ligSize;j++) {
	if (ligBool[j]) {
	  index=scalMap[ligIndex+j];
	  if (!NVConstrs[index]) {
	    if (!bUnknowns[index])
	      appearance[index]=i;
	    disappearance[index]=i;
	  }
	  else {
	    for (k=0;k<NVConstrs[index]->size;k++) {
	      index2=NVConstrs[index]->indices[k];
	      if (!bUnknowns[index2]) {
		bUnknowns[index2]=true;
		appearance[index2]=i;
	      }
	      disappearance[index2]=i;
	    }
	  }
	  bUnknowns[index]=true;
	}
      }
      
      colSize=tmpScal->getColSize();
      colIndex=tmpScal->getColIndex();
      colBool=tmpScal->getColBool();
      map2Max=max(map2Max,colSize);
      for (j=0;j<colSize;j++) {
	if (colBool[j]) {
	  index=scalMap[colIndex+j];
	  
	  if (lUnknowns[index]<currentLayer)
	    lUnknowns[index]=currentLayer;
	    
	  if (!NVConstrs[index]) {
	    if (!bUnknowns[index])
	      appearance[index]=i;
	    disappearance[index]=i;
	  }
	  else {
	    for (k=0;k<NVConstrs[index]->size;k++) {
	      index2=NVConstrs[index]->indices[k];
	      if (!bUnknowns[index2]) {
		bUnknowns[index2]=true;
		appearance[index2]=i;
	      }
	      disappearance[index2]=i;
	    }	  
	  }
	  bUnknowns[index]=true;
	}
      }      
    } // for (i=0;i<slSize;i++)
    
    // All constrained variables are unknowns : 
    
    for (i=0;i<EC.size;i++) {
      expression *tmpExpr;
      tmpExpr=EC.exprData+i;
      bUnknowns[tmpExpr->index]=true;
      for (k=0;k<tmpExpr->size;k++) {
	index2=tmpExpr->indices[k];
	bUnknowns[index2]=true;
      }
    }
    
    if (map2Max>0) {
      ligMap2=new Int[map2Max];
      colMap2=new Int[map2Max];
    }
    else {
      ligMap2=NULL;
      colMap2=NULL;
    }
    for (i=0;i<map2Max;i++) {
      ligMap2[i]=0;
      colMap2[i]=0;
    }
    
    // Calculation of the number of active variables for each list :
    
    Int* nbrAct=NULL;
    nbrAct=new Int[slSize];
    nbrActMax=0;
    for (i=0;i<slSize;i++)
      nbrAct[i]=0;
    nbrVariables=0;
    for (i=0;i<nvSize;i++) {
      if (bUnknowns[i]) {
	nbrVariables++;
	if (appearance[i]>=0)
	  for (j=appearance[i];j<=disappearance[i];j++)
	    nbrAct[j]++;
      }
    }
    nbrVariables-=EC.size; // On laisse tomber les variables contraintes.
    for (i=0;i<slSize;i++)
      nbrActMax=max(nbrActMax,nbrAct[i]);
    nbrActMax+=maxNbrDelayed;
    
    delete[] nbrAct;
    nbrAct=NULL;
    
    // Intermediate results printing :
    
    if (!silent) {
      cerr << "Total number of active variables :  "
	<< nbrVariables+EC.size << " :" << endl ;
      cerr << "   number of unconstrained variables :  "
	<< nbrVariables << "," << endl ;
      cerr << "   number of constrained variables :  "
	<< EC.size << "." << endl ;
      cerr << "Maximum number of simultaneously active variables :  "
	<< nbrActMax << " (" << nbrActMax-maxNbrDelayed << "+" << maxNbrDelayed
	<< ")." << endl ;
    }
    
    // Allocating working matrix and right hand side :
    
    if (nbrActMax>0) {
      Double *tmpMat=new Double[nbrActMax*nbrActMax];
      mat=new Double* [nbrActMax];
      for (i=0;i<nbrActMax;i++)
	mat[i]=tmpMat+nbrActMax*i;
      rhand=new Double[nbrActMax];
      active=new bool[nbrActMax];
      globalPointers=new Int[nbrActMax];
    }
    else {
      mat=NULL;
      rhand=NULL;
      active=NULL;
      globalPointers=NULL;
    }
    
    if (nbrVariables>0) {
      upperMat=new expression[nbrVariables];
      varToNVPointers=new Int[nbrVariables];
    }
    else {
      upperMat=NULL;
      varToNVPointers=NULL;
    }
    
    scal::setWhite(false);
    
    lastState=0; // still in solver
  }
  catchAll("void solver::frontalePrepare(void)","")
}



void solver::frontaleIteration(void) {
  try {
    Int i,j,k,index,indexB,counter;
    step *currentStep;
    scal *tmpScal;
    Int ligSize,colSize;
    Int ligIndex,colIndex;
    Int *scalMap;
    bool *ligBool,*colBool,constrBool;
    Double **localMat,*localCol;
    Int variableIndex;
    Double delta;
    Int currentLayer;
    bool nonZeroPivotFound,zeroLineFound,lineIsZero;
    expression *ligne;
    Double maxPivot,curPivot;
    Int index1,index2;
        
    // Data for constraints resolution :
    
    static list<Int> LI,LIB;
    static list<Int>::iterator LIit,LIitA,LIitB;
    static map<Int,Int> MII;
    static vector<Int> VILig(50),VICol(50);
    static list<expression*> LE; // List of constraints for this problem.
    static list<expression*>::iterator LEit;
    static expression *tmpExpr;
    static Double coef;
    static Int tmpIndex1,tmpIndex2;
    static Int lIndex;
    
    // Prparer ici les contraintes :
    
    constraintsPrepare();
    for (i=0;i<EC.size;i++)
      NVConstrs[EC.exprData[i].index]=EC.exprData+i;
    
    for (i=0;i<nbrActMax;i++) {
      rhand[i]=0.0;
      for (j=0;j<nbrActMax;j++)
	mat[i][j]=0.0;
    }
    for (i=0;i<nbrActMax;i++)
      active[i]=false;
    for (i=0;i<nbrActMax;i++)
      globalPointers[i]=-1;
    for (i=0;i<nvSize;i++)
      bUnknowns[i]=false;
      
    nbrDelayed=0;
    nbrAutoFix=0;
    nbrPermut=0;
    
    variableIndex=0;
    // Forward elimination :
    for (i=0;i<slSize;i++) {
      currentStep=SL[i];
      currentStep->exec();
      
      tmpScal=currentStep->getScalPtr();
      scalMap=tmpScal->getMap();
      currentLayer=currentStep->getEquation()->getLayer();

      // Allocation of space in working matrix and assembling :
      
      counter=0;
      constrBool=false;
      LI.clear();
      LIB.clear();
      MII.clear();
      VILig.clear();
      VICol.clear();
      LE.clear();
      
      ligSize=tmpScal->getLigSize();
      ligBool=tmpScal->getLigBool();
      ligIndex=tmpScal->getLigIndex();
      for (j=0;j<ligSize;j++) {
	if (ligBool[j]) {
	  index=scalMap[ligIndex+j];
	  if (!NVConstrs[index]) {
	    if (!bUnknowns[index]) {
	      while (active[counter])
		counter++;
	      matPointers[index]=counter;
	      globalPointers[counter]=index;
	      active[counter]=true;
	    }
	    ligMap2[j]=matPointers[index];
	  }
	  else {
	    constrBool=true;
	    for (k=0;k<NVConstrs[index]->size;k++) {
	      indexB=NVConstrs[index]->indices[k];
	      if (!bUnknowns[indexB]) {
		bUnknowns[indexB]=true;
		while (active[counter])
		  counter++;
		matPointers[indexB]=counter;
		globalPointers[counter]=indexB;
		active[counter]=true;
	      }
	      ligMap2[j]=matPointers[indexB];
	    }
	  }
	  bUnknowns[index]=true;
	}
      }
      
      colSize=tmpScal->getColSize();
      colBool=tmpScal->getColBool();
      colIndex=tmpScal->getColIndex();
      for (j=0;j<colSize;j++) {
	if (colBool[j]) {
	  index=scalMap[colIndex+j];
	  if (!NVConstrs[index]) {
	    if (!bUnknowns[index]) {
	      while (active[counter])
		counter++;
	      matPointers[index]=counter;
	      globalPointers[counter]=index;
	      active[counter]=true;
	    }
	    colMap2[j]=matPointers[index];
	  }
	  else {
	    constrBool=true;
	    for (k=0;k<NVConstrs[index]->size;k++) {
	      indexB=NVConstrs[index]->indices[k];
	      if (!bUnknowns[indexB]) {
		bUnknowns[indexB]=true;
		while (active[counter])
		  counter++;
		matPointers[indexB]=counter;
		globalPointers[counter]=indexB;
		active[counter]=true;
	      }
	      colMap2[j]=matPointers[indexB];
	    }
	  }
	  bUnknowns[index]=true;
	}
      }
      
      localMat=tmpScal->getMat();
      localCol=tmpScal->getCol();
      if (constrBool) {
	for (j=0;j<colSize;j++)
	  if (colBool[j])
	    LI.push_back(scalMap[colIndex+j]);
	for (j=0;j<ligSize;j++)
	  if (ligBool[j])
	    LI.push_back(scalMap[ligIndex+j]);
	LI.sort();
	LI.unique();
	LIit=LI.begin();
	while (LIit!=LI.end()) {
	  index=*LIit;
	  if (NVConstrs[index]!=NULL) {
	    for (k=0;k<NVConstrs[index]->size;k++) {
	      indexB=NVConstrs[index]->indices[k];
	      LIB.push_back(indexB);
	    }
	  }
	  LIit++;
	} 
	LIB.sort();
	LIB.unique();
	LI.merge(LIB);
	LI.sort();
	LI.unique(); // LI contains the list of active nodal indices.
	
	for (LIit=LI.begin(),j=0;LIit!=LI.end();LIit++,j++)
	  MII[*LIit]=j; // MII[nodal index]=pos in tmpMat.
	// VILig and VICol contain mappings from the local matrix indices
	//    towards the temporary matrix indices.
	for (j=0;j<ligSize;j++)
	  if (ligBool[j])
	    VILig.push_back(MII[scalMap[ligIndex+j]]);
        else
	  VILig.push_back(-1);
	for (j=0;j<colSize;j++)
	  if (colBool[j])
	    VICol.push_back(MII[scalMap[colIndex+j]]);
        else
	  VICol.push_back(-1);
	
	// Temporary structures for constraints applications :
	if (static_cast<Int>(MII.size())>tmpCapacity) {
	  delete[] tmpBool;
	  delete[] tmpCol;
	  delete[] tmpMat;
	  delete[] tmpMatB;
	  tmpSize=0;
	  tmpCapacity=MII.size();
	  if (tmpCapacity>0) {
	    tmpBool=new bool[tmpCapacity];
	    tmpCol=new Double[tmpCapacity];
	    tmpMatB=new Double[tmpCapacity*tmpCapacity];
	    tmpMat=new Double* [tmpCapacity];
	    for (j=0;j<tmpCapacity;j++)
	      tmpMat[j]=tmpMatB+j*tmpCapacity;
	  }
	  else {
	    tmpBool=NULL;
	    tmpCol=NULL;
	    tmpMatB=NULL;
	    tmpMat=NULL;
	  }
	}
	
	// Filling of the temporary structure :
	tmpSize=static_cast<Int>(MII.size());
	for (j=0;j<tmpSize;j++) {
	  for (k=0;k<tmpSize;k++)
	    tmpMat[j][k]=0.0;
	  tmpBool[j]=true;
	  tmpCol[j]=0.0;
	}
	colIndex=tmpScal->getColIndex();
	for (j=0;j<colSize;j++) {
	  if (colBool[j]&&currentLayer==lUnknowns[scalMap[colIndex+j]]) {
	    for (k=0;k<ligSize;k++)
	      if (ligBool[k])
		tmpMat[VICol[j]][VILig[k]]=localMat[j][k];
	    tmpCol[VICol[j]]=-localCol[j];
	  }
	}
	
	for (LIit=LI.begin();LIit!=LI.end();LIit++)
	  if (NVConstrs[*LIit]!=NULL)
	    LE.push_back(NVConstrs[*LIit]);
	LE.sort();
	
	// Lines and columns operations for constraints :
	for (LEit=LE.begin();LEit!=LE.end();LEit++) {
	  tmpExpr=*LEit;
	  // Line operations :
	  tmpIndex1=MII[tmpExpr->index];
	  for (j=0;j<tmpExpr->size;j++) {
	    tmpIndex2=MII[tmpExpr->indices[j]];
	    coef=tmpExpr->factors[j];
	    for (k=0;k<tmpSize;k++)
	      tmpMat[tmpIndex2][k]+=coef*tmpMat[tmpIndex1][k];
	    tmpCol[tmpIndex2]+=coef*tmpCol[tmpIndex1];
	  }
	  // Column operations :
	  for (j=0;j<tmpExpr->size;j++) {
	    tmpIndex2=MII[tmpExpr->indices[j]];
	    coef=tmpExpr->factors[j];
	    for (k=0;k<tmpSize;k++)
	      tmpMat[k][tmpIndex2]+=coef*tmpMat[k][tmpIndex1];
	  }
	  coef=tmpExpr->independentValue;
	  for (k=0;k<tmpSize;k++)
	    tmpCol[k]-=coef*tmpMat[k][tmpIndex1];
	  // The variable is then eliminated :
	  for (k=0;k<tmpSize;k++) {
	    tmpMat[tmpIndex1][k]=0.0;
	    tmpMat[k][tmpIndex1]=0.0;
	  }
	  tmpCol[tmpIndex1]=0.0;
	  tmpBool[tmpIndex1]=false;
	  tmpExpr=NULL;
	}
	if (false) {
	  for (j=0;j<tmpSize;j++) {
	    cerr << setw(15) << tmpCol[j] << "   " ;
	    for (k=0;k<tmpSize;k++) {
	      if (tmpMat[j][k]>0.0) {
		cerr << "+" ;
	      }
	      else if (tmpMat[j][k]<0.0) {
		cerr << "-" ;
	      }
	      else {
		cerr << " " ;
	      }
	    }
	    cerr << endl ;
	  }
	  cerr << endl ;
	}
	
	// The temporary matrix is assembled in working matrix :
	for (LIit=LI.begin(),j=0;j<tmpSize;LIit++,j++) {
	  if (tmpBool[j]) {
	    lIndex=matPointers[*LIit];
	    for (LIitB=LI.begin(),k=0;k<tmpSize;LIitB++,k++) {
	      if (tmpBool[k]) {
		mat[lIndex][matPointers[*LIitB]]+=tmpMat[j][k];
	      }
	    }
	    rhand[lIndex]+=tmpCol[j];
	  }
	}
      } // if (constrBool)
      else {
	colIndex=tmpScal->getColIndex();
	for (j=0;j<colSize;j++) {
	  if (colBool[j]&&currentLayer==lUnknowns[scalMap[colIndex+j]]) {
	    for (k=0;k<ligSize;k++)
	      if (ligBool[k])
		mat[colMap2[j]][ligMap2[k]]+=localMat[j][k];
	    rhand[colMap2[j]]-=localCol[j]; // "-" pour avoir Ax=b
	  }
	}
      }
      
      // Building of the list of variables to be eliminated at this step:
      for (j=0;j<nbrActMax;j++)
	if (active[j])
	  if (disappearance[globalPointers[j]]==i)
	    elimList.push_back(j);

      // Elimination of the variables:
      
      // Usual elimination:
      nonZeroPivotFound=true;
      index=-1;
      while (nonZeroPivotFound&&elimList.size()>0) {
        nonZeroPivotFound=false;
        for (LIit=elimList.begin();LIit!=elimList.end();LIit++) {
	  index=*LIit;
	  if (abs(mat[index][index])>zero) {
	    nonZeroPivotFound=true;
	    break;
	  }
	}
	  
	if (nonZeroPivotFound) {
	  frontaleIteration_eliminateUnknown(index,variableIndex);
	  elimList.erase(LIit);
        } // if (nonZeroPivotFound)
      } // end while (nonZeroPivotFound)
      
      // Permutation of lines and elimination:
      if (permut&&elimList.size()>0) {
        nonZeroPivotFound=true;
        while (nonZeroPivotFound&&elimList.size()>0) {
	  nonZeroPivotFound=false;
          // One searches the largest pivot:
	  maxPivot=0.0;
	  index1=index2=-1;
	  for (LIitA=elimList.begin();LIitA!=elimList.end();LIitA++) {
	    for (LIitB=elimList.begin();LIitB!=elimList.end();LIitB++) {
	      curPivot=abs(mat[*LIitA][*LIitB]);
	      if (curPivot>maxPivot) {
	        maxPivot=curPivot;
		index1=*LIitA;
		index2=*LIitB;
		LIit=LIitB;
	      }
	    }
	  }
	  // Permutation and elimination:
	  if (maxPivot>zero) {
	    nonZeroPivotFound=true;
	    if (index1!=index2)
	      frontaleIteration_permutLines(index1,index2);
	    frontaleIteration_eliminateUnknown(index2,variableIndex);
	    elimList.erase(LIit);
	  }
	} // while (nonZeroPivotFound&&elimList.size()>0)
      } // if (permut&&elimList.size()>0)
      
      // Automatic fixation for zero lines:
      if (autoFix&&elimList.size()>0) {
	zeroLineFound=true;
	while (zeroLineFound&&elimList.size()>0) {
	  zeroLineFound=false;
          for (LIit=elimList.begin();LIit!=elimList.end();LIit++) {
	    index=*LIit;
	    lineIsZero=true;
	    if (abs(rhand[index])>zero)
	      lineIsZero=false;
	    for (j=0;j<nbrActMax;j++) {
	      if (active[j]&&abs(mat[index][j])>zero) {
	        lineIsZero=false;
		break;
	      }
	    }
	    
	    if (lineIsZero) {
	      nbrAutoFix++;
	      zeroLineFound=true;
	      rhand[index]=autoFixValue
	        -field::getNodalVariableValue(globalPointers[index]);
	      for (j=0;j<nbrActMax;j++)
	        mat[index][j]=0.0;
              mat[index][index]=1.0;
	      frontaleIteration_eliminateUnknown(index,variableIndex);
	      elimList.erase(LIit);
              break;
	    }	      
	  } // for (LIit=elimList.begin();LIit!=elimList.end();LIit++)
	} // while (zeroLineFound&&elimList.size()>0)
      } // if (autoFix&&elimList.size()>0)
      
      nbrDelayed=max(nbrDelayed,elimList.size());
      
      // Diagnostic at the end of current step:
      if (elimList.size()>static_cast<unsigned int>(maxNbrDelayed)) {
          errWrite("solver::frontaleIteration",strFromInt(elimList.size())
		   +" uneliminable unknowns in Gaussian elimination.");
        cerr << "   This occurs at step " << i+1 << " of " 
          << slSize << " corresponding to" << endl ;
        cerr << "      problem named \""
          << currentStep->getEquation()->getName() << "\"." << endl ;
        
        cerr << "   The problematic nodal variables are :" << endl ;
        for (LIit=elimList.begin();LIit!=elimList.end();LIit++) {
          cerr << "    Global nodal variable index :  " 
	    << globalPointers[*LIit] << "," << endl ;
          field::writeInfoOnNodalVariable(globalPointers[*LIit]);
        }
      } // if (elimList.size()>static_cast<unsigned int>(maxNbrDelayed))
        
      if (!err::getOk()) {
        errThrow("solver::frontaleIteration",
        	 "bailing out due to preceding errors.");
      }
      
    } // for (i=0;i<slSize;i++) (End of forward elimination)
    
    // Diagnostic at the end forward elimination:
    if (elimList.size()>0) {
    	errWrite("solver::frontaleIteration",strFromInt(elimList.size())
        	 +" uneliminable unknowns at the end of forward elimination.");
      
      cerr << "   The problematic nodal variables are :" << endl ;
      for (LIit=elimList.begin();LIit!=elimList.end();LIit++) {
    	cerr << "    Global nodal variable index :  " 
          << globalPointers[*LIit] << "," << endl ;
    	field::writeInfoOnNodalVariable(globalPointers[*LIit]);
      }
    } // if (elimList.size()>0)
      
    if (!err::getOk()) {
      errThrow("solver::frontaleIteration",
    	       "bailing out due to preceding errors.");
    }
    else {
      // Backward elimination :
      for (i=nbrVariables-1;i>=0;i--) {
        ligne=upperMat+i;
        counter=ligne->size;
        delta=ligne->independentValue;
        for (j=0;j<counter;j++)
    	  delta+=ligne->factors[j]*NVDelta[ligne->indices[j]];
        NVDelta[varToNVPointers[i]]=delta;
      }
      for (i=EC.size-1;i>=0;i--) {
        ligne=EC.exprData+i;
        counter=ligne->size;
        delta=ligne->independentValue;
        for (j=0;j<counter;j++)
    	  delta+=ligne->factors[j]*NVDelta[ligne->indices[j]];
        NVDelta[ligne->index]=delta;
      }
      
      // Freeing temporary containers
      LI.clear();
      LIB.clear();
      MII.clear();
      VILig=vector<Int>(0);
      VICol=vector<Int>(0);
      LE.clear();
      
      for (i=0;i<EC.size;i++) {
        ligne=EC.exprData+i;
        ligne->clear();
        ligne->setCapacity(0);
      }
      
      lastState=0; // still in solver
    }
  }
  catchAll("void solver::frontaleIteration(void)","")
}



void solver::frontaleIteration_eliminateUnknown(Int index,Int &variableIndex) {
  register Double factor;
  register Double *index1,*index2,*lastIndex;
  static Int firstNonZero,lastNonZero;
  static expression *ligne;
  static Int counter,j;

  // One tries to find the first and last non zero elements 
  //    in the line of mat:
  counter=0;
  firstNonZero=lastNonZero=-1;
  for (j=0;j<nbrActMax;j++) {
    if (active[j]&&mat[index][j]!=0.0) {
      if (firstNonZero==-1)
	firstNonZero=j;
      lastNonZero=j;
      counter++;
    }
  }
  
  // Elimination in the active matrix :
  for (j=0;j<nbrActMax;j++) {
    if (active[j]&&mat[j][index]!=0.0&&j!=index) {
      factor=-mat[j][index]/mat[index][index];
      index1=mat[index]+firstNonZero;
      index2=mat[j]+firstNonZero;
      lastIndex=mat[index]+lastNonZero+1;
      while (index1!=lastIndex)
	(*(index2++))+=factor*(*(index1)++);
      rhand[j]+=factor*rhand[index];
      mat[j][index]=0.0;
    }
  }
  // Here, one saves matrix U:
  factor=-1.0/mat[index][index];
  ligne=upperMat+variableIndex;
  ligne->setCapacity(max(counter-1,0));
  ligne->setIndex(variableIndex);
  ligne->setIndependentValue(rhand[index]/mat[index][index]);
  for (j=0;j<nbrActMax;j++) {
    if (active[j]&&j!=index&&mat[index][j]!=0.0)
      ligne->addTerm(globalPointers[j],factor*mat[index][j]);
    mat[index][j]=0.0;
  }
  rhand[index]=0.0;
  varToNVPointers[variableIndex]=globalPointers[index];
  globalPointers[index]=-1;
  active[index]=false;
  variableIndex++; 
}



void solver::frontaleIteration_permutLines(Int index1,Int index2) {
  register Double tmpDouble;
  register Double *pos1,*pos2,*lastPos;
  
  nbrPermut++;
  
  //cerr << "permutLines:" << setw(8) << index1
  //                       << setw(8) << index2
  //		         << setw(20) << mat[index1][index2]<< endl ;
  
  tmpDouble=rhand[index2];
  rhand[index2]=rhand[index1];
  rhand[index1]=tmpDouble;
  
  for (pos1=mat[index1],pos2=mat[index2],lastPos=pos1+nbrActMax;
       pos1<lastPos;
       pos1++,pos2++) {
     tmpDouble=*pos2;
     *pos2=*pos1;
     *pos1=tmpDouble;
  }
}



void solver::frontaleClear(void) {
  try {
    set<arrayFilter*>::iterator Sit;
    
    // constraints :

    EC.setCapacity(0);
    
    // steps :
    
    SL.clear();
    slSize=0;
    
    // filters :
    
    for (Sit=SAF.begin();Sit!=SAF.end();Sit++)
      (*Sit)->cleanFilters();
    SAF.clear();
    
    // other data :
    
    unknownsCapacity=0;
    delete[] bUnknowns; 
    bUnknowns=NULL;
    delete[] lUnknowns; 
    lUnknowns=NULL;
    nvSize=0;
    delete[] NVBackup;
    NVBackup=NULL;
    delete[] NVDelta;
    NVDelta=NULL;
    delete[] NVConstrs;
    NVConstrs=NULL;
    
    delete[] appearance; 
    appearance=NULL;
    delete[] disappearance; 
    disappearance=NULL;
    
    nbrActMax=0;
    if (mat!=NULL) {
      delete[] mat[0];
      delete[] mat; 
      mat=NULL;
    }
    delete[] rhand; 
    rhand=NULL;
    delete[] active; 
    active=NULL;
    delete[] matPointers; 
    matPointers=NULL; 
    delete[] globalPointers;
    globalPointers=NULL;
    elimList.resize(0);
    
    map2Max=0;
    delete[] ligMap2;
    ligMap2=NULL;
    delete[] colMap2;
    colMap2=NULL;
    
    elimList.resize(0);
    
    nbrVariables=0;
    delete[] varToNVPointers;
    varToNVPointers=NULL;
    delete[] upperMat;
    upperMat=NULL;
    
    tmpSize=0;
    tmpCapacity=0;
    delete[] tmpBool;
    tmpBool=NULL;
    delete[] tmpCol;
    tmpCol=NULL;
    delete[] tmpMat;
    tmpMat=NULL;
    delete tmpMatB;
    tmpMatB=NULL;
  }
  catchAll("void solver::frontaleClear(void)","")
}



bool solver::NewtonRaphson(void) {
  try {
    Int iterNumber=0;
    Double relativeVariation=(converge+diverge)/2;
    field::updateAllFieldsIndices();
    frontalePrepare();
    field::backupAllFieldsNodalValues(NVBackup);
    
    while (iterNumber<nitMax&&relativeVariation<diverge) {
      frontaleIteration();
      relativeVariation=field::updateAllFieldsNodalValues(NVDelta,bUnknowns,!silent);
      iterNumber++;
      if (!silent) {
	cerr << "(nbrDelayed=" << nbrDelayed 
	     << ",nbrAutoFix=" << nbrAutoFix
	     << ",nbrPermut=" << nbrPermut
	     << ")" << endl ;
        cerr << "Total relative variation at step " << iterNumber << " :  "
	  << relativeVariation << endl ;
      }
      if (relativeVariation<converge||nitMax==1) {
	if (!silent)
	  cerr << "Convergence assumed at step " << iterNumber << "." << endl ;
	frontaleClear();
	lastState=1; // converged
	return(true);
      }
    }
    
    nbrIterations=iterNumber;
    
    if (relativeVariation<diverge) {
      lastState=2; // not diverged but not converged
      if (!silent)
        cerr << "Maximum number of iterations has been reached, but without convergence!" 
	     << endl ;
    }
    else {
      lastState=3; // diverged
      if (!silent)
        cerr << "divergence has occured!" 
	     << endl ;
    }
    field::restoreAllFieldsNodalValues(NVBackup);
    frontaleClear();
    return(false);
  }
  catchAll("bool solver::NewtonRaphson(void)","")
}



void solver::setLastState(Int nbr) {
  lastState=nbr;
}



Int solver::getLastState(void) const {
  return(lastState);
}



void solver::setNewPoint(bool b) {
  for (Int i=0;i<problemsSize;i++)
    if (problems[i]!=NULL) 
      problems[i]->setNewPoint(b);
  for (Int i=0;i<constraintsSize;i++)
    if (constraints[i]!=NULL) 
      constraints[i]->setNewPoint(b);
  newPoint::setNewPoint(b);
}



void solver::accessNbrInit(void) {
  for (Int i=0;i<problemsSize;i++)
    if (problems[i]!=NULL) 
      problems[i]->accessNbrInit();
  for (Int i=0;i<constraintsSize;i++)
    if (constraints[i]!=NULL) 
      constraints[i]->accessNbrInit();
  newPoint::accessNbrInit();
}



printTree solver::getPrintTreeData(void) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf;
    printTree pt;
    
    if (getName()=="")
      sprintf(cbuf,"   anonymous");
    else
      sprintf(cbuf,"   name = %s",getName().c_str());
    sbuf=cbuf;
    pt+=sbuf;
    
    sprintf(cbuf,"   zero = %g",zero);
    pt+=cbuf;
    
    return(pt);
  }
  catchAll("solver::getPrintTreeData","")
}



printTree solver::getPrintTreeValues(void) const {
  try {
    printTree pt;
    return(pt);
  }
  catchAll("solver::getPrintTreeValues","")
}



printTree solver::getPrintTree(bool values) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf,typeName;
    printTree pt,ptArg;
    Int i;
    
    sprintf(cbuf,"solver %p",static_cast<const void*>(this));
    sbuf=cbuf;
    pt+=sbuf;
    
    pt+=getPrintTreeData();
    
    if (values)
      pt+=getPrintTreeValues();
    
    sbuf="   ";
    pt+=sbuf;
    
    pt.prepareForArguments();

    sprintf(cbuf," | problemsSize = %d",problemsSize);
    pt+=cbuf;
    sbuf=" | ";
    pt+=sbuf;
    for (i=0;i<problemsSize;i++) {
      sprintf(cbuf,"problem[%d]",i);
      ptArg=problems[i]->print::getPrintTreeSummary(cbuf);
      ptArg.prepareForAdd();
      pt+=ptArg;
    }
    
    sprintf(cbuf," | constraintsSize = %d",constraintsSize);
    pt+=cbuf;
    sbuf=" | ";
    pt+=sbuf;
    for (i=0;i<constraintsSize;i++) {
      sprintf(cbuf,"constraint[%d]",i);
      ptArg=constraints[i]->print::getPrintTreeSummary(cbuf);
      if (i<constraintsSize-1)
	ptArg.prepareForAdd();
      else
	ptArg.prepareForAddLast();
      pt+=ptArg;
    }
    
    return(pt);
  }
  catchAll("solver::getPrintTree","")
}



printTree solver::getPrintTreeSummary(void) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf,typeName;
    printTree pt;
    
    sprintf(cbuf,"solver %p",static_cast<const void*>(this));
    sbuf=cbuf;
    pt+=sbuf;
    
    if (getName()=="")
      sprintf(cbuf,"   anonymous");
    else
      sprintf(cbuf,"   name = %s",getName().c_str());
    sbuf=cbuf;
    pt+=sbuf;
    
    sbuf="   ";
    pt+=sbuf;
    
    return(pt);
  }
  catchAll("solver::getPrintTreeSummary","")
}

    
    
bool solver::check(ostream &os) {
  try {
    field::updateAllFieldsIndices();
    frontalePrepare();
    
    for (Int i=0;i<nvSize;i++) {
      if (bUnknowns[i]) {
        field::writeInfoOnNodalVariable(i);
        os << setw(3) << ""
	   << " globalIndex=" << i
	   << " layer=" << lUnknowns[i]
	   << " appearance=" << appearance[i]
	   << " disappearance=" << disappearance[i] 
	   << endl ;
	if (NVConstrs[i])
	  NVConstrs[i]->write(os); 

      }
    }
    
    frontaleClear();
    return(false);
  }
  catchAll("bool solver::check(ostream &os)","")
}


