
//
//  Copyright 2004 Renaud Sizaire
//
//  This file is part of keyFE2.
//
//  keyFE2 is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  keyFE2 is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with keyFE2; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <cmath>
#include <iomanip>
#include <iostream>
#include <stdio.h>

#include "timeDep_base.h"
#include "err/err.h"
#include "field/field_base.h"
#include "print/printTree.h"
#include "solver/solver_base.h"
#include "util/utils.h"

using namespace std ;
using namespace keyFE2 ;



timeDep::timeDep(void) : name(), parents() {
   try {
     timeScheme=0;
     
     maxStepsNbr=-1;
     
     epsilon=0.0;
     
     dtMin=-1.0;
     dtMax=-1.0;
     dtInit=-1.0;
     
     tInit=0.0;
     tFinal=0.0;
     
     influSolversNbr=0;
     influSolvers=NULL;
     
     activeFieldsNbr=0;
     activeFields=NULL;
     activeInfluence=NULL;
     previousFields=NULL;
     
     currentState=-1;
     workingFields=NULL;
     allFieldsNbr=0;
     allFieldsN=NULL;
   }
   catchAll("timeDep::timeDep(void)","")
}



timeDep::timeDep(const timeDep &in) : name(), parents() {
   try {
     timeScheme=in.timeScheme;
     
     maxStepsNbr=in.maxStepsNbr;
     
     epsilon=in.epsilon;
     
     dtMin=in.dtMin;
     dtMax=in.dtMax;
     dtInit=in.dtInit;
     
     tInit=in.tInit;
     tFinal=in.tFinal;
     
     influSolversNbr=0;
     influSolvers=NULL;
     
     activeFieldsNbr=0;
     activeFields=NULL;
     activeInfluence=NULL;
     previousFields=NULL;
     
     currentState=-1;
     workingFields=NULL;
     allFieldsNbr=0;
     allFieldsN=NULL;
   }
   catchAll("timeDep::timeDep(const timeDep &in)","")
}
  


timeDep::~timeDep(void) {
  try {
    clear(); // deletes working data.
    
    timeScheme=0;
    
    maxStepsNbr=-1;
    
    epsilon=0.0;
    
    dtMin=-1.0;
    dtMax=-1.0;
    dtInit=-1.0;
    
    tInit=0.0;
    tFinal=0.0;
    
    influSolversNbr=0;
    delete[] influSolvers;
    influSolvers=NULL;
    
    activeFieldsNbr=0;
    delete[] activeFields;
    activeFields=NULL;
    delete[] activeInfluence;
    activeInfluence=NULL;
    delete[] previousFields;
    previousFields=NULL;
  }
  catchAll("timeDep::~timeDep(void)","")
}



void timeDep::operator=(const timeDep &in) {
  if (this==&in)
    return;
  
  try {
    timeScheme=in.timeScheme;
    
    maxStepsNbr=in.maxStepsNbr;
    
    epsilon=in.epsilon;
    
    dtMin=in.dtMin;
    dtMax=in.dtMax;
    dtInit=in.dtInit;
    
    tInit=in.tInit;
    tFinal=in.tFinal;
    
    influSolversNbr=0;
    influSolvers=NULL;
    
    activeFieldsNbr=0;
    activeFields=NULL;
    activeInfluence=NULL;
    previousFields=NULL;
    
    currentState=-1;
    workingFields=NULL;
    allFieldsNbr=0;
    allFieldsN=NULL;
  }
  catchAll("void timeDep::operator=(const timeDep &in)","")
}



void timeDep::setTimeScheme(string timeSchemeName) {
  if (timeSchemeName=="Implicit Euler")
    timeScheme=1;
  else if (timeSchemeName=="Galerkin")
    timeScheme=2;
  else if (timeSchemeName=="Crank-Nicolson")
    timeScheme=3;
  else
    errThrow("timeDep::setTimeScheme",string("unknown scheme name :  ")
	     +timeSchemeName+".");
}



void timeDep::setTimeScheme(Int timeSchemeNbr) {
  timeScheme=timeSchemeNbr;
  if (timeSchemeNbr<1||timeSchemeNbr>3)
    errThrow("timeDep::setTimeScheme",string("unknown scheme number :  ")
	     +strFromInt(timeSchemeNbr)+".");
}



string timeDep::getTimeSchemeName(void) const {
  switch (timeScheme) {
  case 1 :
    return("Implicit Euler");
  case 2 :
    return("Galerkin");
  case 3 :
    return("Crank-Nicolson");
  default :
    errThrow("timeDep::getTimeSchemeName",string("unknown scheme number :  ")
	     +strFromInt(timeScheme)+".");
  }
}



Int timeDep::getTimeScheme(void) const {
  return(timeScheme);
}



void timeDep::setMaxStepsNbr(Int nbr) {
  maxStepsNbr=nbr;
}



void timeDep::setEpsilon(Double e) {
  epsilon=e;
}



void timeDep::setDtMin(Double dt) {
  dtMin=dt;
}



void timeDep::setDtMax(Double dt) {
  dtMax=dt;
}



void timeDep::setDtInit(Double dt) {
  dtInit=dt;
}



void timeDep::setTInit(Double t) {
  tInit=t;
}



void timeDep::setTFinal(Double t ) {
  tFinal=t;
}



void timeDep::addInfluField(field *F) {
  addActiveField(F,1);
}



field* timeDep::addActiveField(field *F,Int influ) {
  
  try {
    
    // One first checks whether the field is already in the list :
    
    for (Int i=0;i<activeFieldsNbr;i++) {
      if (activeFields[i]==F) {
	if (influ==1)
	  activeInfluence[i]=true;
	else if (influ==0)
	  activeInfluence[i]=false;    
	return(previousFields[i]);
      }
    }
    
    // One then adds the field to the list :
    
    field **tmpActiveFields=activeFields;
    bool *tmpActiveInfluence=activeInfluence;
    activeFields=new field* [activeFieldsNbr+1];
    activeInfluence=new bool[activeFieldsNbr+1];
    for (Int i=0;i<activeFieldsNbr;i++) {
      activeFields[i]=tmpActiveFields[i];
      activeFields[i]->addParent(activeFields+i);
      delinkAndDelete(tmpActiveFields[i],tmpActiveFields+i);
      tmpActiveFields[i]=NULL;
      activeInfluence[i]=tmpActiveInfluence[i];
    }
    delete[] tmpActiveFields;
    delete[] tmpActiveInfluence;
    activeFields[activeFieldsNbr]=F;
    activeFields[activeFieldsNbr]->addParent(activeFields+activeFieldsNbr);
    if (influ==1)
      activeInfluence[activeFieldsNbr]=true;
    else
      activeInfluence[activeFieldsNbr]=false;    
    
    field **tmpPreviousFields=previousFields;
    previousFields=new field* [activeFieldsNbr+1];
    for (Int i=0;i<activeFieldsNbr;i++) {
      previousFields[i]=tmpPreviousFields[i];
      previousFields[i]->addParent(previousFields+i);
      delinkAndDelete(tmpPreviousFields[i],tmpPreviousFields+i);
      tmpPreviousFields[i]=NULL;
    }
    delete[] tmpPreviousFields;
    field *previousF=new field(*F);
    previousFields[activeFieldsNbr]=previousF;
    previousFields[activeFieldsNbr]
      ->addParent(previousFields+activeFieldsNbr);
    activeFieldsNbr++;
    
    return(previousF);
  }
  catchAll("field* timeDep::addActiveField(field *F,Int influ)","")
}



void timeDep::addInfluSolver(solver *S) {
  try {
    // One first checks whether the solver is already in the list :
    for (Int i=0;i<influSolversNbr;i++)
      if (influSolvers[i]==S)
	return;
    
    // One then adds the solver to the list :
    solver **tmpInfluSolvers=influSolvers;
    influSolvers=new solver* [influSolversNbr+1];
    for (Int i=0;i<influSolversNbr;i++) {
      influSolvers[i]=tmpInfluSolvers[i];
      influSolvers[i]->addParent(influSolvers+i);
      delinkAndDelete(tmpInfluSolvers[i],tmpInfluSolvers+i);
      tmpInfluSolvers[i]=NULL;
    }
    delete[] tmpInfluSolvers;
    influSolvers[influSolversNbr]=S;
    influSolvers[influSolversNbr]->addParent(influSolvers+influSolversNbr);
    influSolversNbr++;  
  }
  catchAll("void timeDep::addInfluSolver(solver *S)","")
}



void timeDep::init(void) {
  try {
    Int i,j;
    field *tmpFields;
    
    currentState=0;
    currentStep=0;
    t=tInit;
    
    for (i=1;i<5;i++)
      lastDt[i]=0.0;
    lastDt[0]=dtInit;
    
    switch (timeScheme) {
    case 1 :
    case 2 :
      tmpFields=new field[2*activeFieldsNbr]; // Zp, dZdt[n]
      workingFields=new field* [activeFieldsNbr];
      for (i=0;i<activeFieldsNbr;i++) {
	workingFields[i]=tmpFields+2*i;
	for (j=0;j<2;j++) {
	  workingFields[i][j]=*activeFields[i];
	  workingFields[i][j].setAllNodalValues(0.0);
	}
      }
      break ;
    case 3 :
      tmpFields=new field[3*activeFieldsNbr]; // Zp, dZdt[n], dZdt[n-1]
      workingFields=new field* [activeFieldsNbr];
      for (i=0;i<activeFieldsNbr;i++) {
	workingFields[i]=tmpFields+3*i;
	for (j=0;j<3;j++) {
	  workingFields[i][j]=*activeFields[i];
	  workingFields[i][j].setAllNodalValues(0.0);
	}
      }
      break ;
    }
    
    allFieldsNbr=field::getAllFieldsSize();
    allFieldsN=new field[allFieldsNbr];
    for (i=0;i<allFieldsNbr;i++) {   
      allFieldsN[i]=*(field::getAllField(i));      
      allFieldsN[i].setAllNodalValues(0.0);    
    }
  }
  catchAll("void timeDep::init(void)","")
}



bool timeDep::solversCheck(void) const {
  try {
    for (Int i=0;i<influSolversNbr;i++)
      if (influSolvers[i]->getLastState()>1)
	return(false);
    return(true);  
  }
  catchAll("void timeDep::init(void)","")
}



void timeDep::solversReinit(void) const {
  try {
    for (Int i=0;i<influSolversNbr;i++)
      influSolvers[i]->setLastState(-1); // lastState = none
  }
  catchAll("void timeDep::solversReinit(void) const","")
}



void timeDep::nextStep(void) {
  try {
    Int i,j;
    Double newDt,tmpDt;
    Double normZ,normZn,normZ_Zp,test;
    Int NVnbr;
    Double *NVZ,*NVZn,*NVZp;
    
    // Evaluation of a new time step newDt :
    test=0.0;
    for (i=0;i<activeFieldsNbr;i++) {
      if (activeInfluence[i]) {
	normZ=0.0;
	normZn=0.0;
	normZ_Zp=0.0;
	NVnbr=activeFields[i]->getNodalValuesNbr();
	NVZ=activeFields[i]->getNodalValues();
	NVZp=workingFields[i][0].getNodalValues();
	NVZn=NULL;
	for (j=0;j<allFieldsNbr;j++)
	  if (activeFields[i]==field::getAllField(j))
	    NVZn=allFieldsN[j].getNodalValues();
	for (j=0;j<NVnbr;j++) {
	  normZ=max(normZ,abs(NVZ[j]));
	  normZn=max(normZn,abs(NVZn[j]));
	  normZ_Zp=max(normZ_Zp,abs(NVZ[j]-NVZp[j]));
	}
	if (max(normZ,normZn)==0.0) 
	  test=max(test,normZ_Zp);
	else
	  test=max(test,normZ_Zp/max(normZ,normZn));
      }
    }
    if (currentStep<=3)
      newDt=dtInit;
    else {
      if (timeScheme==1||timeScheme==2||currentStep<=3)
	tmpDt=lastDt[0]*pow(epsilon/test,0.5);
      else
	tmpDt=lastDt[0]*pow(epsilon/test,1.0/3.0);
      newDt=min(1.5*lastDt[0],tmpDt);
      newDt=min(newDt,dtMax);  
    }
    if (newDt<dtMin) {
      currentState=0;
      return;
    }
    
    // dZdt[n-1]=dZdt[n] :
    if (timeScheme==3)
      for (i=0;i<activeFieldsNbr;i++)
	workingFields[i][2].setAllNodalValues(1.0,workingFields[i][1]);
    
    // dZdt[n]=0 or dZdt[n]=(Z[n+1]-Zr)/thetaDt :
    if (currentStep<1)
      for (i=0;i<activeFieldsNbr;i++)
	workingFields[i][1].setAllNodalValues(0.0);
    else {
      for (i=0;i<activeFieldsNbr;i++) {
	workingFields[i][1].setAllNodalValues(1.0/getThetaDT(),
					      *activeFields[i],
					      -1.0/getThetaDT(),
					      *previousFields[i]);
      }
    }
    // Z[bak]=Z[n] :
    for (i=0;i<allFieldsNbr;i++) {
      field *tmpF=field::getAllField(i);
      allFieldsN[i].setAllNodalValues(1.0,*tmpF);
    }
    
    for (i=4;i>0;i--)
      lastDt[i]=lastDt[i-1];
    if (t+newDt>tFinal) {
      newDt=tFinal-t;
      lastDt[0]=newDt;
      t=tFinal;
    }
    else {
      lastDt[0]=newDt;
      t+=lastDt[0];
    }
    currentStep++;
    currentState=1;
    
    return;  
  } 
  catchAll("void timeDep::nextStep(void)","")
}



void timeDep::redoStep(void) {
  try {
    t-=lastDt[0];
    lastDt[0]=lastDt[0]/2.0;
    t+=lastDt[0];
    currentState=1;
    if (lastDt[0]<dtMin) {
      currentState=0;
      return;
    }  
  }
  catchAll("void timeDep::redoStep(void)","")
}



void timeDep::makePrediction(void) {
  try {
    Double thetaDt,fact1,fact2;
    Int i,index;
    
    if (timeScheme==1)
      theta=1.0;
    else if (timeScheme==2&&currentStep==1)
      theta=1.0;
    else if (timeScheme==2&&currentStep>1)
      theta=2.0/3.0;
    else if (timeScheme==3&&currentStep==1)
      theta=1.0;
    else if (timeScheme==3&&currentStep==2)
      theta=2.0/3.0;
    else if (timeScheme==3&&currentStep>=3)
      theta=1.0/2.0;
    
    thetaDt=theta*lastDt[0];
    
    // Zr=Z[n]+theta*Dt*dZdt[n] :
    for (i=0;i<activeFieldsNbr;i++) {
      index=field::getAllFieldIndex(activeFields[i]);
      previousFields[i]->setAllNodalValues(1.0,allFieldsN[index],
					   lastDt[0]-thetaDt,
					   workingFields[i][1]);
    }
    
    // Zp=Z[n+1]=...
    for (i=0;i<activeFieldsNbr;i++) {
      index=field::getAllFieldIndex(activeFields[i]);
      if (timeScheme==3&&currentStep>=10) {
	fact1=lastDt[0]+0.5*lastDt[0]*lastDt[0]/lastDt[1];
	fact2=-0.5*lastDt[0]*lastDt[0]/lastDt[1];
	activeFields[i]->setAllNodalValues(1.0,allFieldsN[index],
					   fact1,workingFields[i][1],
					   fact2,workingFields[i][2]);
      }
      else {
	activeFields[i]->setAllNodalValues(1.0,allFieldsN[index],
					   lastDt[0],workingFields[i][1]);
      }
      workingFields[i][0].setAllNodalValues(1.0,*activeFields[i]); // Zp
    }
  }
  catchAll("void timeDep::makePrediction(void)","")
}



void timeDep::clear(void) {
  try {
    currentState=0;
    currentStep=0;
    if (workingFields!=NULL) {
      delete[] workingFields[0];
      delete[] workingFields;
      workingFields=NULL;
    }
    allFieldsNbr=0;
    delete[] allFieldsN;
    allFieldsN=NULL;
  }
  catchAll("void timeDep::clear(void)","")
}



bool timeDep::isToBeContinued(void) {
  try {
    bool okIncr;
    
    if (currentState==-1) {
      init();
      okIncr=true;
    }
    else {
      okIncr=solversCheck();
    }
    
    if (okIncr) {
      currentState=0;
      // One checks whether final step is reached :
      if (t>=tFinal||currentStep==maxStepsNbr) {
	clear();
	return(false);
      }
      // Next step :
      nextStep();    
      if (currentState==0) {
	clear();
	return(false);   
      }
    }
    else {
      redoStep();
      if (currentState==0) {
	clear();
	return(false);
      }
    }
    
    cerr << "Making Time Step  : " << endl ;
    cerr << "   current step index :  " << currentStep << endl ;
    cerr << "   current time       :  " << t << endl ;
    cerr << "   current time step  :  " << lastDt[0] << endl ;
    
    makePrediction();
    solversReinit();
    
    return(true);
  }
  catchAll("bool timeDep::isToBeContinued(void)","")
}



bool timeDep::isActive(void) const {
  return(currentState==1);
}



Double timeDep::getThetaDT(void) const {
  try {
    return(theta*lastDt[0]);
  }
  catchAll("Double timeDep::getThetaDT(void) const","")
}



Int timeDep::getCurrentStepIndex(void) const {
  return(currentStep);
}



Double timeDep::getCurrentTimeStep(void) const {
  try {
    return(lastDt[0]);
  }
  catchAll("Double timeDep::getCurrentTimeStep(void) const","")
}



Double timeDep::getCurrentTime(void) const {
  return(t);
}



Double timeDep::getPreviousTime(void) const {
  try {
    return(t-lastDt[0]);
  }
  catchAll("Double timeDep::getPreviousTime(void) const","")
}



void timeDep::setNewPoint(bool b) {
  for (Int i=0;i<influSolversNbr;i++)
    if (influSolvers[i]!=NULL) 
      influSolvers[i]->setNewPoint(b);
  for (Int i=0;i<activeFieldsNbr;i++) {
    if (activeFields!=NULL)
      if (activeFields[i]!=NULL)
	activeFields[i]->setNewPoint(b);
    if (previousFields!=NULL)
      if (previousFields[i]!=NULL)
	previousFields[i]->setNewPoint(b);
    if (workingFields!=NULL)
      if (workingFields[i]!=NULL)
	workingFields[i]->setNewPoint(b);
    if (allFieldsN!=NULL)
      for (Int i=0;i<allFieldsNbr;i++)
	(allFieldsN+i)->setNewPoint(b);
  }
  newPoint::setNewPoint(b);
}



void timeDep::accessNbrInit(void) {
  for (Int i=0;i<influSolversNbr;i++)
    if (influSolvers[i]!=NULL) 
      influSolvers[i]->accessNbrInit();
  for (Int i=0;i<activeFieldsNbr;i++) {
    if (activeFields!=NULL)
      if (activeFields[i]!=NULL)
	activeFields[i]->accessNbrInit();
    if (previousFields!=NULL)
      if (previousFields[i]!=NULL)
	previousFields[i]->accessNbrInit();
    if (workingFields!=NULL)
      if (workingFields[i]!=NULL)
	workingFields[i]->accessNbrInit();
    if (allFieldsN!=NULL)
      for (Int i=0;i<allFieldsNbr;i++)
	(allFieldsN+i)->accessNbrInit();
  }
  newPoint::accessNbrInit();
}



printTree timeDep::getPrintTreeData(void) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf;
    printTree pt;
    
    if (getName()=="")
      sprintf(cbuf,"   anonymous");
    else
      sprintf(cbuf,"   name = %s",getName().c_str());
    pt+=cbuf;
    
    sprintf(cbuf,"   timeScheme = %d",timeScheme);
    pt+=cbuf;
    
    sprintf(cbuf,"   maxStepsNbr = %d",maxStepsNbr);
    pt+=cbuf;
    
    sprintf(cbuf,"   epsilon = %g",epsilon);
    pt+=cbuf;
    
    sprintf(cbuf,"   dtMin = %g",dtMin);
    pt+=cbuf;
    sprintf(cbuf,"   dtMax = %g",dtMax);
    pt+=cbuf;
    sprintf(cbuf,"   dtInit = %g",dtInit);
    pt+=cbuf;
    sprintf(cbuf,"   tInit = %g",tInit);
    pt+=cbuf;
    sprintf(cbuf,"   tFinal = %g",tFinal);
    pt+=cbuf;
    
    return(pt);
  }
  catchAll("timeDep::getPrintTreeData","")
}



printTree timeDep::getPrintTreeValues(void) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf;
    printTree pt;
    
    if (currentState==-1)
      pt+="   currentState = none";
    else if (currentState==0)
      pt+="   currentState = failed";
    else if (currentState==1)
      pt+="   currentState = continue";
    sprintf(cbuf,"   currentStep = %d",currentStep);
    pt+=cbuf;
    sprintf(cbuf,"   t = %g",t);
    pt+=cbuf;
    sprintf(cbuf,"   theta = %g",theta);
    pt+=cbuf;
    sprintf(cbuf,"   lastDt = %g",lastDt[0]);
    pt+=cbuf;
    
    return(pt);
  }
  catchAll("timeDep::getPrintTreeValues","")
}






printTree timeDep::getPrintTree(bool values) const {
  try {
    char cbuf[STRING_LENGTH];
    string sbuf,typeName;
    printTree pt,ptArg;
    Int i;
    
    sprintf(cbuf,"timeDep %p",static_cast<const void*>(this));
    sbuf=cbuf;
    pt+=sbuf;
    
    pt+=getPrintTreeData();
    
    if (values)
      pt+=getPrintTreeValues();
    
    sbuf="   ";
    pt+=sbuf;
    
    pt.prepareForArguments();

    sprintf(cbuf," | influSolversNbr = %d",influSolversNbr);
    pt+=cbuf;
    sbuf=" | ";
    pt+=sbuf;
    for (i=0;i<influSolversNbr;i++) {
      sprintf(cbuf,"influSolver[%d]",i+1);
      ptArg=influSolvers[i]->print::getPrintTreeSummary(cbuf);
      ptArg.prepareForAdd();
      pt+=ptArg;
    }
    
    sprintf(cbuf," | activeFieldsNbr = %d",activeFieldsNbr);
    pt+=cbuf;
    sbuf=" | ";
    pt+=sbuf;
    for (i=0;i<activeFieldsNbr;i++) {
      sprintf(cbuf,"activeField[%d]",i+1);
      ptArg=activeFields[i]->print::getPrintTreeSummary(cbuf);
      if (i<activeFieldsNbr-1)
	ptArg.prepareForAdd();
      else
	ptArg.prepareForAddLast();
      pt+=ptArg;
    }
    
    return(pt);
  }
  catchAll("timeDep::getPrintTree","")
}



